#include "config.h"
#include "genlib.h"
#include "dosfs.h"
#include "cli.h"
#include "sd.h"
#include "sd_lld.h"


/*------------------------------------
 * Note: this works with Altera's SPI
 * core.  It does not provide for
 * card detect or write protect.
 * ----------------------------------*/

// from system.h
#define ctl_base (int*)SD_0_BASE

#define	CRC			0x95		// all purpose CRC

// the configuration data
SD_CONF sdConf;

/*----------------------------------------------------------------------
   SpiTxdRxd

   Transmit a byte out on the Spi MOSI and return the data
   read on the MISO pin.
  ----------------------------------------------------------------------*/
unsigned char SpiTxdRxd( unsigned int base, unsigned char TxdData)
{
   unsigned char RxdData;
   unsigned int  Status;

   // wait until the shift register is empty
   do
   {
     Status = IORD_ALTERA_AVALON_SPI_STATUS(base);
   }
   while ((Status & ALTERA_AVALON_SPI_STATUS_TMT_MSK) == 0);

   // clear the rxd buffer
   RxdData = IORD_ALTERA_AVALON_SPI_RXDATA(base);

   // write the value to transmit
   IOWR_ALTERA_AVALON_SPI_TXDATA( base, TxdData);

   // wait until data is ready to be read
   do
   {
     Status = IORD_ALTERA_AVALON_SPI_STATUS(base);
   }
   while ((Status & ALTERA_AVALON_SPI_STATUS_RRDY_MSK) == 0);

   // read the value clocked in from the MISO pin
   RxdData = IORD_ALTERA_AVALON_SPI_RXDATA(base);

//   printf(" 0x%x", RxdData);

   return RxdData;
}

int SpiCommand( unsigned int base, PSD_CMD pCmd)
{
   unsigned int Status, i;

   Status = IORD_ALTERA_AVALON_SPI_STATUS(base);
//   printf("Status: %x\n", Status);

   // clear he status register
   IOWR_ALTERA_AVALON_SPI_STATUS( base, Status);

   // Set the SSO bit (force chipselect)
   IOWR_ALTERA_AVALON_SPI_SLAVE_SEL( base, 1);
   IOWR_ALTERA_AVALON_SPI_CONTROL(base, ALTERA_AVALON_SPI_CONTROL_SSO_MSK);

   // determine the number of lead-in clocks
   if( pCmd->Flags & SD_80_CLK_INIT)
      i = 10;  // 80 clocks - init
   else
      i =  1;  // just one byte - 8

   // output the clocks - ignore rxd
   while( i--)
   {
      SpiTxdRxd( base, 0xff);
   }

   // write the command - ignore rxd
   SpiTxdRxd( base, pCmd->Cmd | 0x40);
   SpiTxdRxd( base, pCmd->Arg >> 24);
   SpiTxdRxd( base, pCmd->Arg >> 16);
   SpiTxdRxd( base, pCmd->Arg >>  8);
   SpiTxdRxd( base, pCmd->Arg >>  0);
   SpiTxdRxd( base, pCmd->Crc      );

   // wait for a response
   i = 32;

   // msb of responce is 0, but wait only 32 trys
   while((( pCmd->pResp[0] = SpiTxdRxd( base, 0xff)) & 0x80) && ( i-- > 0));

   if ( i == 0)
   {
      // we failed
      printf("Timeout\n");
      return 0;
   }
   else
   {
      // get the rest of the responce
      for ( i = 1; i < pCmd->RespLength; i++)
      {
         pCmd->pResp[i] = SpiTxdRxd( base, 0xff);
      }
   }

   // the response is done - is this a block read or write?
   switch( pCmd->Cmd)
   {
      // it would be better to have flags in the command
      case SD_SEND_CSD:
      case SD_SEND_CID:
      case SD_READ_SINGLE_BLOCK:
      {
         // wait for the data token (0xfe)
        while( SpiTxdRxd( base, 0xff) != 0xfe);

        /* Keep clocking until all the data has been processed. */
        for ( i = 0; i < pCmd->DataLength; i++)
        {
          pCmd->pData[i] = SpiTxdRxd( base, 0xff);
        }

        pCmd->BlkCrc  = 0;
        pCmd->BlkCrc  = SpiTxdRxd( base, 0xff);
        pCmd->BlkCrc |= SpiTxdRxd( base, 0xff) << 8;

        break;
      }

      case SD_WRITE_SINGLE_BLOCK:
      {
         // write the data token (0xfe)
         SpiTxdRxd( base, 0xfe);

         // write all of the data
         for ( i = 0; i < pCmd->DataLength; i++)
         {
            SpiTxdRxd( base, pCmd->pData[i]);
         }

         SpiTxdRxd( base, (unsigned char)pCmd->BlkCrc);
         SpiTxdRxd( base, (unsigned char)pCmd->BlkCrc >> 8);

         break;
      }
   }

  // wait until not busy and command complete
  while( SpiTxdRxd( base, 0xff) != 0xff);

  // reset slave select
  IOWR_ALTERA_AVALON_SPI_CONTROL(base, 0);

  return 0;
}


static int sdSend( unsigned char cmd, unsigned long arg, int rl, unsigned char * pResp, int dl, unsigned char * pData)
{
   SD_CMD Cmd;

 //  printf("Cmd: %d\n", cmd);

   memset( (void *)&Cmd, 0, sizeof( SD_CMD));

   Cmd.Cmd        = cmd;
   Cmd.Arg        = arg;
   Cmd.Crc        = CRC;
   Cmd.pResp      = pResp;
   Cmd.RespLength = rl;
   Cmd.pData      = pData;
   Cmd.DataLength = dl;

   if ( cmd == SD_GO_IDLE_STATE)
   {
      Cmd.Flags |= SD_80_CLK_INIT;
   }


  SpiCommand( SD_0_BASE, &Cmd);

  return 0;
}

// no implementation yet
int sdPresent(void)
{
    return 0;
}

int sdInit( int interface, int verbosity)
{
   unsigned char  Resp;
   int            i = 0;


   sdSend( SD_GO_IDLE_STATE, 0, 1, (unsigned char *)&Resp, 0, NULL);

   // loop until the card is out of intialization
   do
   {
      sdSend( SD_APP_CMD,      0, 1, &Resp, 0, NULL);
      sdSend( SD_SEND_OP_COND, 0, 1, &Resp, 0, NULL);

   } while (( Resp & 0x1) == 1 && ( i++ < 100));

   // if card never got out of initialization - fail
   if ( i == 0)
   {
      printf("SD Initialization Failed - Resp = 0x%x\n", Resp);
      return 1;
   }

   // get the  operating conditions register
   sdSend( SD_READ_OCR, 0, sizeof(sdConf.sdOcr), (unsigned char *)sdConf.sdOcr, 0, NULL);

   // get the card identification data
   sdSend( SD_SEND_CID, 0, 1, &Resp, sizeof(sdConf.sdCid), (unsigned char *)sdConf.sdCid);

   // get the card specific data
   sdSend( SD_SEND_CSD, 0, 1, &Resp, sizeof(sdConf.sdCsd), (unsigned char *)sdConf.sdCsd);

   if ( verbosity)
   {
      i = (sdConf.sdCid[9] << 24) | (sdConf.sdCid[10] << 16) | (sdConf.sdCid[11] << 8) | sdConf.sdCid[12];

      printf("\tMfg. Code   : %c%c\n", sdConf.sdCid[1], sdConf.sdCid[2]);
      printf("\tProduct Name: %c%c%c%c%c\n",  sdConf.sdCid[3], sdConf.sdCid[4], sdConf.sdCid[5], sdConf.sdCid[6], sdConf.sdCid[7]);
      printf("\tRevision    : %1d%1d\n", ( sdConf.sdCid[8] >> 4) & 0xf, sdConf.sdCid[8] & 0xf);
      printf("\tSerial No.  : %d\n", i);
    }

   return 0;
}

int  sdRead(int interface, char *buf, int blknum, int blkcount)
{
  unsigned char Resp[2];

  sdSend( SD_SET_BLOCKLEN     ,  SD_BLKSIZE, 1, (unsigned char *)&Resp,          0, NULL);
  sdSend( SD_READ_SINGLE_BLOCK, blknum << 9, 1, (unsigned char *)&Resp, SD_BLKSIZE, buf);
  sdSend( SD_STOP_TRANSMISSION,           0, 2, (unsigned char *)&Resp,          0, NULL);

  return 0;
}

int  sdWrite(int interface, char *buf, int blknum, int blkcount)
{
  unsigned char Resp[2];

  sdSend( SD_SET_BLOCKLEN      , SD_BLKSIZE,  1, (unsigned char *)&Resp, 0, NULL);
  sdSend( SD_WRITE_SINGLE_BLOCK, blknum << 9, 1, (unsigned char *)&Resp, SD_BLKSIZE, buf);
  sdSend( SD_STOP_TRANSMISSION ,           0, 2, (unsigned char *)&Resp, 0, NULL);

  return 0;
}

