//=============================================================================
//
//      cpuio.c
//
//      CPU/Board Specific IO
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Date:         05-26-2002
// Description:  This file contains the IO functions required by Micro Monitor
//				 that are unique to each CPU/Board combination
//
//
//=============================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "ether.h"
#include "stddefs.h"
#include "uart16550.h"
#include "cpu_pci.h"
#include "ps2_kbd.h"
#include "arch_ppc.h"

extern void sed135x_on();
extern int i2c_init(void);
extern int fs6377_init(int mode);
extern int ps2_kbd_init(void);
extern int ps2_gotachar(void);
extern uchar ps2_getchar(void); 
extern void sed_pwr_dn(void);
extern int sed135x_init(void);
extern void sed_lcd_bkl(uchar bright);
extern void sed_putchar(char c);

extern void pci_init();
extern int pci_probe_bus(int bus);

// globals
extern int sed_color_depth;
extern int sed135x_ok;
extern int next_bus;
extern int mem_base;
extern int io_base;
extern int pci_dev;



// Delay for some usecs. - Not accurate, assumes ROM mode
// and no Cache
void udelay(int delay)
{
  int i;
  for ( i = LOOPS_PER_USEC * delay; i ; i--)
	;
}

#if INCLUDE_SED
// We use the SED GPIO1 to control the LCD backlight
void lcd_bkl(uchar bright)
{
	sed_lcd_bkl(bright);
}
#endif

int
getUartDivisor(int baud, unsigned char *hi, unsigned char *lo)
{
	*lo = ((MEMCLK/16)/baud) & 0x00ff;
	*hi = (((MEMCLK/16)/baud) & 0xff00) >> 8;
	return(0);
}

int
devInit(int baud)
{
	// turn on LED0 and LED1
	BCSR &= ~(LED0 | LED1);

	// Initialize the UART
    InitUART(baud);

	// Initialize the FS6377 Clock Generator
	// before we release reset to the SED
	if (fs6377_init(SED_DISP_MODE_CRT))
		printf("FS6377 Init Failed!\n");

	// Release PCI, PS2 and SED Reset
	BCSR |= P_RST;

	// Initialize the I2C Port
	i2c_init();

#if INCLUDE_SED
	// Initialize the display controller
	sed_color_depth = 4;
	sed135x_init();
	if (sed135x_ok == 0)
		printf("sed135x_init() failed!\n");
	sed135x_on();
#endif

#if INCLUDE_PS2KBD
	// Initialize the PS2 controller and keyboard - if attached
	ps2_kbd_init();
#endif

	// Probe the pci bus for devices and fill in pci_dev_info
	pci_init();
	next_bus = 0;
	mem_base = PCI_MEM_BASE;
	io_base = PCI_IO_BASE;
	pci_dev = 0;
	pci_probe_bus(0);

	return 0;
}

void
target_reset(void)
{
	coldstart();
}

void
initCPUio(void)
{
	/* Set the UART for two-port mode */
	UARTA(UART_DCR) = DCR_SDM;
	return;
}

void
intsrestore(ulong msr)
{
    ppcMtmsr(msr);
}

ulong
intsoff()
{
    ulong  msr;

    msr = ppcMfmsr();
    ppcMtmsr(msr & ~0x8000);
    return(msr);
}

/* cacheInitForTarget():
    Enable instruction cache only...
*/
void
cacheInitForTarget()
{
    asm(".equ   r3,3");
    asm(".equ   IC_CST,560");

    asm("   mfspr   r3,IC_CST");    
    asm("   lis     r3,0x0200");
    asm("   mtspr   IC_CST,r3");
    asm("   isync");
}
