//=============================================================================
//
//      cpuio.c
//
//      CPU/Board Specific IO
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Date:         05-26-2002
// Description:  This file contains the IO functions required by Micro Monitor
//				 that are unique to each CPU/Board combination
//
//
//=============================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "ether.h"
#include "stddefs.h"
#include "at91rm9200.h"
#include "at91rm9200_dbgu.h"
#include "at91rm9200_gpio.h"
#include "csb337_gpio.h"
#include "cpuio.h"
#include "warmstart.h"

extern void GPIO_set(int);
extern void GPIO_clr(int);
extern void lcd_putchar(char c);
extern void lcd_init(void);
extern void lcd_on(void);

extern  void putpsr(ulong);
extern  ulong getpsr();

/* InitUART():
 * This is the top-level UART initialization for the monitor. 
 */
void
InitUART(int baud)
{
	// clear error bits and reset tx and rx
	DBGU_REG(DBGU_CR) = DBGU_CR_RSTSTA
					   | DBGU_CR_RSTTX
					   | DBGU_CR_RSTRX;

	// clear any pending interrupts and mask all further interrupts
	DBGU_REG(DBGU_IDR) = DBGU_INT_ALL;
	DBGU_REG(DBGU_IMR) = 0x00000000;

	// Baud divisor = MCK / 16 / baud
	DBGU_REG(DBGU_BRGR) = ((MCK/16)/baud);	// integer only

	// set the port to no parity, no loopback
	DBGU_REG(DBGU_MR) = DBGU_MR_PAR_NONE | DBGU_MR_CHMODE_NORM;

	// Enable
	DBGU_REG(DBGU_CR) = DBGU_CR_TXEN | DBGU_CR_RXEN;
}

int
devInit(int baud)
{
	// Initialize the UART
    InitUART(baud);

	// Initialize the I2C Port
//	i2c_init();

	// Initialize the LCD controller
//	lcd_init();
//	lcd_on();

	// turn off LED2 (set on hw_inti2.c) and turn on
	// LED0 to indicate completion of init code
//	GPIO_set(USR_LED2);
//	GPIO_clr(USR_LED0);

	return 0;
}

int
ConsoleBaudSet(int baud)
{
	// Baud divisor = MCK / 16 / baud
	DBGU_REG(DBGU_BRGR) = ((MCK/16)/baud);	// integer only
	return(0);
}

/* target_console_empty():
 * Target-specific portion of flush_console() in chario.c.
 * This function returns 1 if there are no characters waiting to
 * be put out on the UART; else return 0 indicating that the UART
 * is still busy outputting characters from its FIFO.
 */
int
target_console_empty(void)
{
	if ((DBGU_REG(DBGU_CSR) & DBGU_INT_TXBUFE))
		return(1);
	return(0);
}

/* target_putchar():
 */
int
target_putchar(char c)
{
    /* Wait for transmit ready bit
	 */
    while(1) {
        if (DBGU_REG(DBGU_CSR) & DBGU_INT_TXRDY)
            break;
    }
	DBGU_REG(DBGU_THR) = c;
    return((int)c);
}

/* target_gotachar():
 * Return 0 if no char is avaialable at UART rcv fifo; else 1.
 * Do NOT pull character out of fifo, just return status. 
 */
int
target_gotachar()
{
    if (!(DBGU_REG(DBGU_CSR) & DBGU_INT_RXRDY))
        return(0);
    return(1);
}

/* target_getchar():
 */
int 
target_getchar(void)
{
    char    c;

    /* Retrieve character */
    c = DBGU_REG(DBGU_RHR);
    return((int)c);
}

void
initCPUio()
{
}

void
intsrestore(psr)
ulong     psr;
{
    putpsr(psr);
}

/*
 * Read the program status register (CPSR)
 * and set the FIQ and IRQ bits.
 */
ulong
intsoff()
{
    ulong  psr;

    psr = getpsr();

    /*
     * Set bit 6, bit 7 to disable interrupts.
     */
    putpsr(psr | 0x000000c0);
    return(psr);
}

/* cacheInitForTarget():
    Enable instruction cache only...
*/
void
cacheInitForTarget()
{
    asm("   MRC p15, 0, r0, c1, c0, 0");
    asm("   ORR r0, r0, #0x1000");  /* bit 12 is ICACHE enable*/
    asm("   MCR p15, 0, r0, c1, c0, 0");

    /* Flush instruction cache */
    asm("   MCR p15, 0, r0, c7, c5, 0");
}

void
target_reset(void)
{
	coldstart();
}
