//==========================================================================
//
// csb437_gpio.h
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Contributors: 
// Date:         6/20/2003
// Description:  This file contains the board level to vitual
//				 GPIO bit assignments
//

// The AT91RM9200 GPIO's are spread across four 32-bit ports A-D.
// For bits that have an alternate function assigned, we use the bit 
// position within the respective port as we need that to setup the
// registers.  But for bits that are used as GPIOs we assign a 
// virtual GPIO number which is used by the GPIO functions
//
// This section also defines the bit setting for the alternate
// functions of the pins.  If we want alternate function A, we prepend
// an A, for function B, a B.  This insures we will get an error, or
// at least it will be very visible if we try to assign a function
// to the wrong side.  Note the A or B is not related to the Port A
// or Port B, just the alternate function.
//
// Finally we build the default values to be written into the GPIO registers
// in gpio_init (in csb437_gpio.c).


#ifndef CSB437_GPIO_H
#define CSB437_GPIO_H

#include "bits.h"

// PORT A							   BIT	A/B	DIR	SENSE	DESCRIPTION
#define PA0				BIT0		// PA0  -	-	-		Unused on CSB437
#define	PA1				BIT1		// PA1  -	-	-		Unused on CSB437
#define PA2				BIT2		// PA2  -	-	-		Unused on CSB437
#define PA3				BIT3		// PA3  -	-	-		Unused on CSB437
#define PA4				BIT4		// PA4  -	-	-		Unused on CSB437
#define PA5				BIT5		// PA5  -	-	-		Unused on CSB437
#define PA6				BIT6		// PA6  - 	-	-		Unused on CSB437
#define A_ETXCK			BIT7		// PA7  A	-	-		Ethernet Transmit Clock
#define A_ETXEN			BIT8		// PA8  A	-	-		Ethernet Transmit Enable
#define A_ETXD0			BIT9		// PA9  A	-	-		Ethernet Transmit Data 0
#define A_ETXD1			BIT10		// PA10 A	-	-		Ethernet Transmit Data 1
#define A_ECRS			BIT11		// PA11 A	-	-		Ethernet Carrier Sense 
#define A_ERXD0			BIT12		// PA12 A	-	-		Ethernet Receive Data 0
#define A_ERXD1			BIT13		// PA13 A	-	-		Ethernet Receive Data 1
#define A_ERXER			BIT14		// PA14 A	-	-		Ethernet Receive Error
#define A_EMDC			BIT15		// PA15 A	-	-		Ethernet MD Clock
#define A_EMDIO			BIT16		// PA16 A	-	-		Ethernet MD Data
#define PA17			BIT17		// PA17 - 	-	-		Unused on CSB437
#define PA18			BIT18		// PA18 - 	-	-		Unused on CSB437
#define CF_IRQ			19			// PA19 - 	In	-		Expansion GPIO
#define USB_CD			20			// PA20 - 	In	High	USB Device Cable Detect
#define USB_PWR2		21 			// PA21 - 	Out	Low		USB Host 1 Power Enable
#define USB_PWR1		22 			// PA22 - 	Out	Low		USB Host 2 Power Enable
#define USB_OC			23 			// PA23 - 	In	Low		USB Overcurrent Detect
#define USB_SCON		24			// PA24 - 	In	High	USB Soft Connect
#define SDA				25			// PA25 - 	I/O	-		I2C Data - GPIO mode
#define SCL				26 			// PA26 - 	I/O	-		I2C Clock - GPIO mode
#define A_SDCK			BIT27		// PA27 A 	- 	- 		SD Card Clock 
#define A_SDCMD			BIT28		// PA28 A 	- 	- 		SD Card Command
#define A_SDD0			BIT29		// PA29 A 	- 	- 		SD Card Data 0
#define A_DBG_RXD		BIT30 		// PA30 A 	- 	- 		Debug Receive
#define A_DBG_TXD		BIT31 		// PA31 A 	- 	-		Debug Transmit

// default settings for PORT A
// Peripheral A select - SD Card 0, Ethernet and Debug UART
#define DEF_PIOA_ASR 	(A_ETXCK | A_ETXEN | A_ETXD0 | A_ETXD1 | A_ECRS | A_ERXD0 | A_ERXD1 | A_ERXER | A_EMDC | A_EMDIO | A_SDCK | A_SDCMD | A_SDD0 | A_DBG_RXD | A_DBG_TXD)

// Peripheral B select - None
#define DEF_PIOA_BSR	0	

 // enable pullups where desired - usually just unused GPIO
#define DEF_PIOA_PUER	(PA0 | PA1 | PA2 | PA3 | PA4 | PA5 | PA6 | PA17 | PA18)

// enable default state of GPIO outputs that need to start high
#define DEF_PIOA_SODR	(USB_PWR1 | USB_PWR2)

// enable default state of GPIO outputs that need to start low
#define DEF_PIOA_CODR	0

// enable all GPIO pins that start as outputs
#define DEF_PIOA_OER	(USB_PWR1 | USB_PWR2)

// PORT B							   BIT	A/B	DIR	SENSE	DESCRIPTION
#define USR_LED 		32			// PB0  -	Out Low		Alert LED
#define PB1		 		BIT1		// PB1  -	- 	-		Unused on CSB437
#define PB2		 		BIT2		// PB2  -	- 	-		Unused on CSB437
#define B_SDD1	 		BIT3		// PB3  B	-	-		SD Card Data 1 
#define B_SDD2	 		BIT4		// PB4  B	-	-		SD Card Data 2
#define B_SDD3	 		BIT5		// PB5  B	-	-		SD Card Data 3
#define A_TF1 			BIT6		// PB6  A 	-	-		SSC1 Transmit Frame
#define A_TK1			BIT7		// PB7  A 	-	-		SSC1 Transmit Clock
#define A_TD1			BIT8		// PB8  A 	-	-		SSC1 Transmit Data
#define A_RD1			BIT9		// PB9  A 	-	-		SSC1 Receive Data
#define PB10 			BIT10		// PB10 - 	-	-		Unused on CSB437
#define PB11	 		BIT11		// PB11 - 	-	-		Unused on CSB437
#define B_ETXD2	 		BIT12		// PB12 B	-	-		Ethernet Transmit Data 2
#define B_ETXD3	 		BIT13		// PB13 B	-	-		Ethernet Transmit Data 3
#define B_ETXER	 		BIT14		// PB14 B	-	-		Ethernet Transmit Error
#define B_ERXD2	 		BIT15		// PB15 B	-	-		Ethernet Receive Data  
#define B_ERXD3	 		BIT16		// PB16 B	-	-		Ethernet Receive Data 
#define B_ERXDV	 		BIT17		// PB17 B	-	-		Ethernet Receive Data Valid
#define B_ECOL	 		BIT18		// PB18 B	-	-		Ethernet Collision
#define B_ERXCK	 		BIT19		// PB19 B	-	-		Ethernet Receive Clock
#define PB20	 		BIT20		// PB20 -	-	-		Unused on CSB437
#define PB21	 		BIT21		// PB21 -	-	-		Unused on CSB437
#define PB22	 		BIT22		// PB22 -	-	-		Unused on CSB437
#define PB23	 		BIT23		// PB23 -	-	-		Unused on CSB437
#define PB24	 		BIT24		// PB24 -	-	-		Unused on CSB437
#define PB25	 		BIT25		// PB25 -	-	-		Unused on CSB437
#define PB26	 		BIT26		// PB26 -	-	-		Unused on CSB437
#define A_PCK0			BIT27		// PB27 A	-	-		Peripheral Clock 0 - Bus Clock
#define PB28			BIT28		// PB28 -	-	-		Unused on CSB437
#define PWR_ON			61			// PB29 -	In	Low		Power On Switch
// PB30 - Does not exist							
// PB31 - Does not exist							

// default settings for PORT B
// Peripheral A select - SSC1 and Peripheral Clock 0
#define DEF_PIOB_ASR	(A_TF1 | A_TK1 | A_TD1 | A_RD1 | A_PCK0)

// Peripheral B select - SD Card 0 and Ethernet
#define DEF_PIOB_BSR	(B_SDD1 | B_SDD2 | B_SDD3 | B_ETXD2 | B_ETXD3 | B_ETXER | B_ERXD2 | B_ERXD3 | B_ERXDV | B_ECOL | B_ERXCK)

 // enable pullups where desired - usually just unused GPIO
#define DEF_PIOB_PUER	(PB1 | PB2 | PB10 | PB11 | PB20 | PB21 | PB22 | PB23 | PB24 | PB25 | PB26 | PB28)

// enable default state of GPIO outputs that need to start high
#define DEF_PIOB_SODR	(USR_LED)

// enable default state of GPIO outputs that need to start low
#define DEF_PIOB_CODR	0	

// enable all GPIO pins that start as outputs
#define DEF_PIOB_OER	(USR_LED)

// PORT C							   BIT	A/B	DIR	SENSE	DESCRIPTION
#define PS2_INT	 		64			// PC0	-	In	Low		Expansion Interrupt 0
#define SMI_INT	 		65			// PC1	-	In	Low		Expansion Interrupt 1
#define PHY_INT	 		66			// PC2	-	In	Low		LT971A PHY Interrupt
#define CF_CD 			67			// PC3	-	In	Low		Compact Flash Detect
#define PC4		 		BIT4		// PC4	-	-	-		Unused on CSB437
#define PC5		 		BIT5		// PC5	-	-	-		Unused on CSB437
#define A_WAIT	 		BIT6		// PC6	A	-	-		WAIT Input	
#define A_A23	 		BIT7		// PC7	A	-	-		Address Bus 23
#define A_A24 			BIT8		// PC8  A 	-	-		Address Bus 24
#define A_CF_RW 		BIT9		// PC9  A 	-	-		Compact Flash R/W
#define A_CF_CS	 		BIT10		// PC10 A 	-	-		Compact Flash Chip Select
#define A_CF_CE1 		BIT11		// PC11 A 	-	-		Compact Flash Card Enable 1
#define A_CF_CE2	 	BIT12		// PC12 A 	-	-		Compact Flash Card Enable 2
#define PC13	 		BIT13		// PC13	-	-	-		Unused on CSB437
#define PC14		 	BIT14		// PC14	-	-	-		Unused on CSB437
#define PC15	 		BIT15		// PC15	-	-	-		Unused on CSB437
#define A_D16	 		BIT16		// PC16	A	-	-		Data Bus 16
#define A_D17	 		BIT17		// PC17	A	-	-		Data Bus 17
#define A_D18	 		BIT18		// PC18	A	-	-		Data Bus 18
#define A_D19	 		BIT19		// PC19	A	-	-		Data Bus 19
#define A_D20	 		BIT20		// PC20	A	-	-		Data Bus 20
#define A_D21	 		BIT21		// PC21	A	-	-		Data Bus 21
#define A_D22	 		BIT22		// PC22	A	-	-		Data Bus 22
#define A_D23	 		BIT23		// PC23	A	-	-		Data Bus 23
#define A_D24	 		BIT24		// PC24	A	-	-		Data Bus 24
#define A_D25	 		BIT25		// PC25	A	-	-		Data Bus 25
#define A_D26	 		BIT26		// PC26	A	-	-		Data Bus 26
#define A_D27	 		BIT27		// PC27	A	-	-		Data Bus 27
#define A_D28	 		BIT28		// PC28	A	-	-		Data Bus 28
#define A_D29	 		BIT29		// PC29	A	-	-		Data Bus 29
#define A_D30	 		BIT30		// PC30	A	-	-		Data Bus 30
#define A_D31	 		BIT31 		// PC31	A	-	-		Data Bus 31
// default settings for PORT C
// Peripheral A select - Wait, Upper Address, Compact Flash and Databus
#define DEF_PIOC_ASR	(A_WAIT | A_A23 | A_A24 | A_CF_RW | A_CF_CS | A_CF_CE1 | A_CF_CE2 | 0xffff0000)

// Peripheral B select - None
#define DEF_PIOC_BSR	0

 // enable pullups where desired - usually just expansion GPIO
#define DEF_PIOC_PUER	(PC4 | PC5 | PC13 | PC14 | PC15)

// enable default state of GPIO outputs that need to start high
#define DEF_PIOC_SODR	0	// no GPIO outputs assigned in PortC

// enable default state of GPIO outputs that need to start low
#define DEF_PIOC_CODR	0	// no GPIO outputs assigned in PortC

// enable all GPIO pins that start as outputs
#define DEF_PIOC_OER	0	// no GPIO outputs assigned in PortC

// PORT D					   	   BIT	A/B	DIR	SENSE	DESCRIPTION
#define CF_PWR 			96			// PD0	-	Out	High	Compact Flash Power Enable
#define CF_EN 			97			// PD1	-	Out	Low		Compact Flash Data Bus Enable
#define CF_RST 			98			// PD2	-	Out	Low		Compact Flash Reset
#define SD_PWR 			99			// PD3	-	Out	High	SD Card Power On
#define PIRQ 			100			// PD4	-	In	Low		Touch Interrupt
#define SD_CD			101			// PD5	-	In	Low		SD Card Detect
#define SD_WP			102			// PD6	-	In	Low		SD Card Write Protect
#define PD7				BIT7		// PD7	-	-	-		Unused on CSB437
#define PD8				BIT8		// PD8	-	-	-		Unused on CSB437
#define PD9				BIT9		// PD9	-	-	-		Unused on CSB437
#define PD10			BIT10		// PD10	-	-	-		Unused on CSB437
#define PD11			BIT11		// PD11	-	-	-		Unused on CSB437
#define PWR_HLD			108			// PD12	-	Out	High	Power Hold
#define PD13			BIT13		// PD13	-	-	-		Unused on CSB437
#define CH_PRES			110			// PD14	-	In	High	Charger Present
#define CH_STAT			111			// PD15	-	In	-		Charger Status - 0 = Charging
#define PD16			BIT16		// PD16	-	-	-		Unused on CSB437
#define PD17			BIT17		// PD17	-	-	-		Unused on CSB437
#define PD18			BIT18		// PD18	-	-	-		Unused on CSB437
#define PD19			BIT19		// PD19	-	-	-		Unused on CSB437
#define PD20			BIT20		// PD20	-	-	-		Unused on CSB437
#define PD21			BIT21		// PD21	-	-	-		Unused on CSB437
#define PD22			BIT22		// PD22	-	-	-		Unused on CSB437
#define PD23			BIT23		// PD23	-	-	-		Unused on CSB437
#define PD24			BIT24		// PD24	-	-	-		Unused on CSB437
#define PD25			BIT25		// PD25	-	-	-		Unused on CSB437
#define PD26			BIT26		// PD26	-	-	-		Unused on CSB437
#define PD27			BIT27		// PD27	-	-	-		Unused on CSB437
// PD28 Does Not Exist
// PD29 Does Not Exist
// PD30 Does Not Exist
// PD31 Does Not Exist

// default settings for PORT D
// Peripheral A select - none
#define DEF_PIOD_ASR	0

// Peripheral B select - None
#define DEF_PIOD_BSR	0

 // enable pullups on all unused GPIO
#define DEF_PIOD_PUER	(PD7 | PD8 | PD9 | PD10 | PD11 | PD13 | 0x0000ffff)

// enable default state of GPIO outputs that need to start high
#define DEF_PIOD_SODR	(CF_RST | CF_EN | PWR_HLD)

// enable default state of GPIO outputs that need to start low
#define DEF_PIOD_CODR	(CF_PWR | SD_PWR)

// enable all GPIO pins that start as outputs
#define DEF_PIOD_OER	(CF_RST | CF_EN | PWR_HLD | CF_PWR | SD_PWR)

#endif // #ifndef CSB437_GPIO_H
