//=============================================================================
//
//      cpuio.c
//
//      CPU/Board Specific IO
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Date:         04/08/2005
// Description:  This file contains the IO functions required by Micro Monitor
//				 that are unique to the CSB536FS
//
//
//=============================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "ether.h"
#include "stddefs.h"
#include "warmstart.h"
#include "mc9328mxl_uart.h"
#include "mc9328mxl.h"
#include "mc9328mxl_lcd.h"
#include "mc9328mxl_mem.h"
#include "cpu_gpio.h"
#include "fb_draw.h"

extern ulong i2c_init(void);
extern ulong getpsr(void);
extern void putpsr(ulong);

extern void lcd_putchar(char c);
extern void lcd_init(void);
extern void lcd_on(void);
extern int lcd_color_depth;		// 4, 8 or 16

// globals to hold background and foreground colors
uchar bcolor=0;		// vga black
uchar color=15;		// vga white

// Delay for some usecs. - Not accurate, assumes ROM mode
// and no Cache
void udelay(int delay)
{
  volatile int i;
  for ( i = LOOPS_PER_USEC * delay; i ; i--)
	;
}

/* InitUART():
 * This is the top-level UART initialization for the monitor. 
 */
void
InitUART(int baud)
{
    if (baud == 0)
        baud = ConsoleBaudRate;

    /* Keep track of the last baud rate, so that it can be used if the */
    /* incoming baudrate is NULL. */
    ConsoleBaudRate = baud;

	// set direction for UART1_TX
//	GPIOC_REG(GPIO_DDIR) = 0x00000800;
	
	// reset tx and rx
	UART1_REG(UART_CR2) = 0;

	// clear any pending interrupts and mask all further interrupts
	UART1_REG(UART_SR1) = 0xffffffff;
	UART1_REG(UART_SR2) = 0xffffffff;
	UART1_REG(UART_CR1) = 0x00000000;
	UART1_REG(UART_CR4) = 0x00000000;

	// enable the uart and it's clocks - no auto baud, no interrupts
	UART1_REG(UART_CR1) |= UART_CR1_UARTEN
					| UART_CR1_UARTCLKEN;

	// set the word parameters
	UART1_REG(UART_CR2) = UART_CR2_WS 		// 8-bit, 1 stop, no parity
					| UART_CR2_IRTS		// ignore RTS 
					| UART_CR2_SRST;	// no reset

	// set the fifo's to legal interrupt threshold values - unused by us
	// set input clock divisor to 1
	UART1_REG(UART_FCR) = UART_FCR_TXTL(2)
					| UART_FCR_RFDIV1
					| UART_FCR_RXTL(1);

	// set the baud rate denominator to 10,000.  Then divide
	// the desired baud by 100 and put that into the numerator.
	// This works when using a reference clock of 16Mhz with
	// RFDIV set to 1.  It also only works down to 1200 baud,
	// but who cares!
	UART1_REG(UART_BIR) = (baud/100) - 1;	// numerator
	UART1_REG(UART_BMR) = 10000 - 1;		// denominator

	// Enable TX and RX, release SRST
	UART1_REG(UART_CR2) |= UART_CR2_TXEN
					| UART_CR2_RXEN
					| UART_CR2_SRST;

    InitRemoteIO();
}

int
ConsoleBaudSet(int baud)
{
	// set the baud rate denominator to 10,000.  Then divide
	// the desired baud by 100 and put that into the numerator.
	// This works when using a reference clock of 16Mhz with
	// RFDIV set to 1.  It also only works down to 1200 baud,
	// but who cares!
	UART1_REG(UART_BIR) = (baud/100) - 1;	// numerator
	UART1_REG(UART_BMR) = 10000 - 1;		// denominator
	return(0);
}

int
devInit(int baud)
{
	// Disable all GPIO for now.  We will allow the gpio_init()
	// code to do it later.
	GPIOA_REG(GPIO_GIUS) = 0;	// no GPIO on Port A for now
	GPIOA_REG(GPIO_GPR) = 0;
	GPIOB_REG(GPIO_GIUS) = 0;	// no GPIO on Port B for now
	GPIOB_REG(GPIO_GPR) = 0;
	GPIOC_REG(GPIO_GIUS) = 0;	// no GPIO on Port C for now
	GPIOC_REG(GPIO_GPR) = 0;
	GPIOD_REG(GPIO_GIUS) = 0;	// no GPIO on Port D for now
	GPIOD_REG(GPIO_GPR) = 0;

	// Initialize GPIO
    GPIO_init();	

    // Initialize the UART
    InitUART(baud);

	// turn on LED0
	GPIO_clr(USR_LED0);

 	// Setup CS0 for 150ns StrataFlash
    EIM_REG(EIM_CS0U) = EIM_CSU_CNC_1		// hold CS negated one clock at the end of all accesses
					  | EIM_CSU_WSC(4)		// wait = 180ns (8+1 x 20ns, assumes BCLK = 48Mhz)
					  | EIM_CSU_WWS(1)		// additional wait states for writes
					  | EIM_CSU_EDC(1);		// 1 clock bus turn around time

	EIM_REG(EIM_CS0L) = EIM_CSL_OEA(2)		// CS asserted # 1/2 clocks before OE
					  | EIM_CSL_OEN(2)		// CS held # 1/2 clocks after OE
					  | EIM_CSL_WEA(2)		// CS asserted # 1/2 clocks before WE
					  | EIM_CSL_WEN(2)		// CS held # 1/2 clocks after WE
					  | EIM_CSL_CSA(0)		// setup and hold address 2 clocks around CS
					  | EIM_CSL_EBC			// assert EB0-3 for writes only
					  | EIM_CSL_DSZ_HALF_0	// 16-bits on D0-15
					  | EIM_CSL_CSEN;		// enable it
 
	// Setup CS1 for CS890a
    EIM_REG(EIM_CS1U) = EIM_CSU_CNC_1		// hold CS negated one clock at the end of all accesses
					  | EIM_CSU_WSC(10)		// wait = 180ns (8+1 x 20ns, assumes BCLK = 48Mhz)
					  | EIM_CSU_WWS(1)		// additional wait states for writes
					  | EIM_CSU_EDC(1);		// 1 clock bus turn around time

	EIM_REG(EIM_CS1L) = EIM_CSL_OEA(4)		// CS asserted # 1/2 clocks before OE
					  | EIM_CSL_OEN(4)		// CS held # 1/2 clocks after OE
					  | EIM_CSL_WEA(4)		// CS asserted # 1/2 clocks before WE
					  | EIM_CSL_WEN(4)		// CS held # 1/2 clocks after WE
					  | EIM_CSL_CSA(0)		// setup and hold address 2 clocks around CS
					  | EIM_CSL_EBC			// assert EB0-3 for writes only
					  | EIM_CSL_DSZ_HALF_0	// 16-bits on D0-15
					  | EIM_CSL_CSEN;		// enable it

    // check for the presence of a card.  If CF_CD is low, then the card is there.
#ifdef CF_DEBUG
	printf("CF Card Detect... ");
#endif
	if ((GPIO_tst(CF_CD) == 0))
	{
#ifdef CF_DEBUG
		printf("Compact Flash Card detected, enable power.\n");
#endif
		GPIO_set(CF_RST);		// reset
		udelay(100);
		GPIO_set(CF_VEN);		//enable power
		udelay(100);
		GPIO_clr(CF_RST);		// release reset
		udelay(100);
#ifdef CF_DEBUG
		printf("Configure CS5 - Compact Flash I/O Space              @ 0x16000000.\n");
		printf("                Reserved Space                       @ 0x16040000.\n");
		printf("                Compact Flash Attribute/Config Space @ 0x16080000.\n");
		printf("                Compact Flash Common Space           @ 0x160C0000.\n");
#endif
		EIM_REG(EIM_CS5U) = EIM_CSU_DTACK_SEL		 // 1 = PCMCIA wait function
   	                      | EIM_CSU_CNC_3      		 // delay clocks after each chip select
//   	                      | EIM_CSU_WSC(31)  		 // use DTACK
   	                      | EIM_CSU_WSC_DTACK  		 // use DTACK
   	                      | EIM_CSU_EDC(1);	   		 // Dead Cycles after reads

		EIM_REG(EIM_CS5L) = EIM_CSL_OEA(4)      	// # of 1/2 cycles after CS asserts before OE asserts
   	                      | EIM_CSL_OEN(4)      	// # of 1/2 cycles OE negates before CS negates
   	                   	  | EIM_CSL_WEA(4)	    	// # of 1/2 cycles EB0-3 assert before WE asserts
   	                   	  | EIM_CSL_WEN(4)	    	// # of 1/2 cycles EB0-3 negate before WE negates 
   	                   	  | EIM_CSL_CSA(0)	    	// # of clocks address is asserted before CS and held after CS
   	                   	  | EIM_CSL_DSZ_HALF_0  	// device is 16-bits wide, located on half word 0 (D15-0)
   	                   	  | EIM_CSL_CSEN;       	// 1 = Enable Chip Select

	}
	else
	{
#ifdef CF_DEBUG
		printf("Compact Flash Card not detected.\n");
#endif
	} // if CF_CD == 0

	// Initialize the I2C Port
	i2c_init();

	// Initialize the LCD controller
// 	lcd_color_depth = 16;		// 4, 8 or 16
//	lcd_init();
//	lcd_on();

	// turn on LED1
	GPIO_clr(USR_LED1);

	return 0;
}

int
target_console_empty(void)
{
	if (UART1_REG(UART_SR2) & UART_SR2_TXFE)
		return(1);
	return(0);
}

int
target_putchar(char c)
{
    /* Wait for transmit ready bit, then send character.
	 */
    while(!(UART1_REG(UART_SR2) & UART_SR2_TXFE)){}
	UART1_REG(UART_TXD) = (ulong)c;
    return((int)c);
}

int
target_gotachar(void)
{
    if (UART1_REG(UART_SR2) & UART_SR2_RDR)
        return(1);
    return(0);
}

int 
target_getchar(void)
{
    char    c;

    c = (uchar)UART1_REG(UART_RXD) & 0xff;
    return((int)c);
}

void 
target_reset(void)
{
	warmstart(INITIALIZE);
}

void
initCPUio()
{
}

void
intsrestore(psr)
ulong     psr;
{
    putpsr(psr);
}

/*
 * Read the program status register (CPSR)
 * and set the FIQ and IRQ bits.
 */
ulong
intsoff(void)
{
    ulong  psr;

    psr = getpsr();

    /*
     * Set bit 6, bit 7 to disable interrupts.
     */
    putpsr(psr | 0x000000c0);
    return(psr);
}

/* cacheInitForTarget():
    Enable instruction cache only...
*/
void
cacheInitForTarget()
{
    asm("   MRC p15, 0, r0, c1, c0, 0");
    asm("   ORR r0, r0, #0x1000");  /* bit 12 is ICACHE enable*/
    asm("   MCR p15, 0, r0, c1, c0, 0");

    /* Flush instruction cache */
    asm("   MCR p15, 0, r0, c7, c5, 0");
}
