/*
 *  pxa250_i2c.c
 *
 *  I2C algorithm for the PXA I2C bus access.
 *
 *  Copyright (C) 2003 Technology Rescue.
 *	Author : David de la Cruz.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *	History:
 *	15.04.2003 : CSB226 adaptation.
 *	24.11.2003 : Translate 7-bit address to 8-bit address and fixed RTC RAM read/write.
 */

#include "config.h"
#include "cli.h"
#include "i2c.h"
#include "genlib.h"
#include "stddefs.h"

#include "pxa250.h"
#include "pxa250_i2c.h"


//--------------------------------------------------------------------------
// ulong i2c_send_address(uchar dev_addr)
//
// This function writes an 8-bit value to the I2C device to start a transaction.
//
// Returns 0 if everything is OK.
// Error codes :
//	-10 : general purpose error code.
//	-11 : timeout = bus is busy.
//	-12 : ITE is not set.
//
ulong i2c_send_address(uchar dev_addr)
{
#ifdef I2C_DEBUG
	printf("i2c_send_address begin.\n");
#endif

	// First of all, clear I2C status register.
	I2C_REG(I2C_ISR) = I2C_ISR_INIT;

	// write the 7-bit device address with write = 0
	if(i2c_wr_byte(dev_addr)){
		return -10;
	}

	// Be sure bus is not busy.
	if(i2c_wait_bus_not_busy())
	{
		printf("bus is busy\n");
		return -11;
	}

	// issue a start command to initiate the write.
	i2c_start(ICR_TB, ICR_ALDIE);

	if(i2c_check_status(ISR_ITE))
		return -12;

#ifdef I2C_DEBUG
	printf("i2c_send_address end.\n");
#endif

	return 0;
}


//--------------------------------------------------------------------------
// ulong i2c_transmit_to_device(uchar addr_cmd, uchar *data, uchar length)
//
// This function transmit some bytes over i2c bus.
//
// Returns 0 if everything is OK.
// Error codes :
//	-1 : general purpose error code.
//	-2..-3 : timeout = bus is busy.
//	-4 : ITE not set.
//	Others : error code from called functions.
//
ulong i2c_transmit_to_device(uchar addr_cmd, uchar *data, uchar length)
{
	ulong error;			// used to store function return values.

#ifdef I2C_DEBUG
	printf("i2c_transmit_to_device begin.\n");
#endif

	// First of all, clear I2C status register.
	I2C_REG(I2C_ISR) = I2C_ISR_INIT;

	// send addr + cmd + R/W bit to i2c bus.
	error = i2c_send_address(addr_cmd & 0xfe);
	if(error)
		return error;

	// Write the (8-bit * length) data value. Assume length is > 1.
	while(length > 0)	// write bytes until the end of data buffer.
	{
#ifdef I2C_DEBUG
		printf("remain : %d\n", length);
#endif

		if(i2c_wr_byte(*data)){
			return -1;
		}

		// Be sure bus is not busy.
		if(i2c_wait_bus_not_busy())
		{
			printf("bus is busy\n");
			return -2;
		}

		if(length == 1)	// Is it the last byte.
		{	// if yes.
			i2c_transfer( 1, I2C_TRANSMIT, 0);
		}
		else
		{	// if not.
			i2c_transfer( 0, I2C_TRANSMIT, 1);
		}

		if(i2c_check_status(ISR_ITE))
			return -4;

		data++;			// next byte to transfer.
		length--;		// decrement length.
	}

	// Be sure bus is not busy.
	if(i2c_wait_bus_not_busy())
	{
		printf("bus is busy\n");
		return -3;
	}

	// Clear STOP bit.
	i2c_midbyte(0, 0);

#ifdef I2C_DEBUG
	printf("i2c_transmit_to_device end.\n");
#endif

	return 0;
}

//--------------------------------------------------------------------------
// ulong i2c_receive_from_device(uchar addr_cmd, uchar *data, uchar length)
//
// This function receive some bytes from i2c bus.
//
// Returns 0 if everything is OK.
// Error codes :
//	-1 : general purpose error code.
//	-2..-5 : timeout = bus is busy.
//	-6 : IRF not set.
//	Others : error code from called functions.
//
ulong i2c_receive_from_device(uchar addr_cmd, uchar *data, uchar length)
{
	ulong error;			// used to store function return values.

#ifdef I2C_DEBUG
	printf("i2c_receive_from_device begin.\n");
#endif

	// First of all, clear I2C status register.
	I2C_REG(I2C_ISR) = I2C_ISR_INIT;

	// send addr + cmd + R/W bit to i2c bus.
	error = i2c_send_address(addr_cmd | 0x01);
	if(error)
		return error;

	// Write the (8-bit * length) data value. Assume length is > 1.
	while(length > 0)	// write bytes until the end of data buffer.
	{
#ifdef I2C_DEBUG
		printf("remain : %d\n", length);
#endif

		// Be sure bus is not busy.
		if(i2c_wait_bus_not_busy())
		{
			printf("bus is busy\n");
			return -3;
		}

		if(length == 1)	// Is it the last byte.
		{	// if yes.
			i2c_transfer( 1, I2C_RECEIVE, 0);
		}
		else
		{	// if not.
			i2c_transfer( 0, I2C_RECEIVE, 1);
		}

		if(i2c_check_status(ISR_IRF))
			return -6;

		// Read data.
		*data = i2c_rd_byte();

		// Be sure bus is not busy.
		if(i2c_wait_bus_not_busy())
		{
			printf("bus is busy\n");
			return -2;
		}

		// clear ICR stop and ACKNAK bits.
		i2c_midbyte(0, ICR_ACKNAK);
	
		data++;			// next byte to transfer.
		length--;		// decrement length.
	}

	// Be sure bus is not busy.
	if(i2c_wait_bus_not_busy())
	{
		printf("bus is busy\n");
		return -4;
	}

	// Master abort.
	i2c_abort();

	// Be sure bus is not busy.
	if(i2c_wait_bus_not_busy())
	{
		printf("bus is busy\n");
		return -5;
	}

	// Clear Master Abort bit.
	i2c_midbyte(0, ICR_MA);

#ifdef I2C_DEBUG
	printf("i2c_receive_from_device end.\n");
#endif

	return 0;
}
//--------------------------------------------------------------------------
// i2c_wr_device(uchar dev_addr, uchar reg, uchar data)
//
// This function writes an 8-bit value to the I2C device at the requested
// register.
//
// Returns 0 if everything is OK.
// Error codes :
//	-1 : general purpose error code.
//	-2 : first ITE is not set.
//	-3 : second ITE is not set.
//	-7..-9 : timeout = bus is busy.
//	Others : error code from called functions.
//
ulong i2c_wr_device(uchar dev_addr, uchar reg, uchar data)
{
	ulong error;			// used to store function return values.

#ifdef I2C_DEBUG
	printf("i2c_wr_device begin.\n");
#endif

	// Translate 7-bit address to 8-bit address.
	dev_addr = dev_addr << 1; 

	// First of all, clear I2C status register.
	I2C_REG(I2C_ISR) = I2C_ISR_INIT;

	// send addr + cmd + R/W bit to i2c bus.
	error = i2c_send_address(dev_addr & 0xfe);
	if(error)
		return error;

	// write the 8-bit register address.
	if(i2c_wr_byte(reg)){
		return -1;
	}

	// Be sure bus is not busy.
	if(i2c_wait_bus_not_busy())
	{
		printf("bus is busy\n");
		return -7;
	}

	// initiate the write.
	i2c_midbyte(ICR_TB, 0x0);

	if(i2c_check_status(ISR_ITE))
		return -2;

	// write the 8-bit data value
	if(i2c_wr_byte(data)){
		return -1;
	}

	// Be sure bus is not busy.
	if(i2c_wait_bus_not_busy())
	{
		printf("bus is busy\n");
		return -8;
	}

	// initiate the last write.
	i2c_stop(ICR_TB, 0x0);

	if(i2c_check_status(ISR_ITE))
		return -3;

	// Be sure bus is not busy.
	if(i2c_wait_bus_not_busy())
	{
		printf("bus is busy\n");
		return -9;
	}

	i2c_midbyte(0, 0);		// Clear STOP bit.
	
#ifdef I2C_DEBUG
	printf("i2c_wr_device end.\n");
#endif

	return 0;
}

//--------------------------------------------------------------------------
// i2c_rd_device()
//
// This function reads an 8-bit value from the I2C device at the requested
// register.
//
// Returns 0 if everything is OK.
// Error codes :
//	-1 : general purpose error code.
//	-2 : second ITE is not set.
//	-3 : third ITE is not set.
//	-4 : IRF is not set.
//	Others : error code from called functions.
//
ulong i2c_rd_device(uchar dev_addr, uchar reg, uchar *data)
{
	ulong error;			// used to store function return values.

#ifdef I2C_DEBUG
	printf("i2c_rd_device begin.\n");
#endif

	// Translate 7-bit address to 8-bit address.
	dev_addr = dev_addr << 1; 

	// First of all, clear I2C status register.
	I2C_REG(I2C_ISR) = I2C_ISR_INIT;

	// send addr + cmd + R/W bit to i2c bus.
	error = i2c_send_address(dev_addr & 0xfe);
	if(error)
		return error;

	// write the 8-bit register address.
	if(i2c_wr_byte(reg)){
		return -1;
	}

	// initiate the write.
	i2c_midbyte(ICR_ALDIE | ICR_TB, 0x0);

	if(i2c_check_status(ISR_ITE))
		return -2;

	// write the 7-bit device address again plus data direction (read = 1)
	if(i2c_wr_byte(dev_addr | 0x01)){
		return -1;
	}

	// issue a start command to initiate the write.
	i2c_start(ICR_TB, ICR_ALDIE);

	if(i2c_check_status(ISR_ITE))
		return -3;

	// issue a stop to initiate the read.
	i2c_stop(ICR_ALDIE | ICR_ACKNAK | ICR_TB,0);

	if(i2c_check_status(ISR_IRF))
		return -4;

	*data = i2c_rd_byte();

	i2c_midbyte(0, ICR_ACKNAK);		// Clear STOP and ACKNAK bits.

#ifdef I2C_DEBUG
	printf("i2c_rd_device end.\n");
#endif

	return 0;
}


int
i2cCtrl(int interface, int cmd, unsigned long arg1, unsigned long arg2)
{
	switch(cmd)
	{
		case I2CCTRL_INIT:
			i2c_init();
			return 0;
		case I2CCTRL_SHOW:
			i2c_show_reg();
			return 0;
		default:
		  break;
	}
	return(-1);
}

int
i2cWrite(int interface, int bigaddr, uchar *data, int len)
{
	ulong error, index;
	uchar *buffer;

	switch(interface)
	{
	case 0x68:		// RTC.
		buffer = (uchar *)malloc((len + 1) * sizeof(uchar));							// allocate write buffer.
		buffer[0] = (uchar)(bigaddr & 0xFF);								// first write : register addr.
		for(index = 0; index < len; index++)
			buffer[index+1] = data[index];
		error = i2c_transmit_to_device((uchar)(interface <<1 ), buffer, len+1);	// Write "len" bytes at address 0x10.
		free((char *)buffer);
		if(error)
		{
		#ifdef I2C_DEBUG
			printf("i2cWrite error: %d\n", error);
		#endif
			return(-2);
		}
		break;
	default:
		#ifdef I2C_DEBUG
		printf("i2cWrite error: interface not supported\n");
		#endif
		return(-1);	// Not supported.
	}
	return(1);	// return 1 byte written.
}

int
i2cRead(int interface, int bigaddr, uchar *data, int len)
{
	ulong error;
	switch(interface)
	{
	case 0x68:		// RTC.
		error = i2c_transmit_to_device((uchar)(interface <<1 ), (uchar *)&bigaddr, 1);
		error = i2c_receive_from_device((uchar)(interface <<1 ), data, len);	// Read "len" bytes at address interface | bigaddr.
		if(error)
		{
		#ifdef I2C_DEBUG
			printf("i2cRead error: %d\n", error);
		#endif
			return(-2);
		}
		break;
	default:
		#ifdef I2C_DEBUG
		printf("i2cRead error: interface not supported\n");
		#endif
		return(-1);	// Not supported.
	}

	return(1);	// return 1 byte read.
}
