//=============================================================================
//
//      cpuio.c
//
//      CPU/Board Specific IO
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Date:         05-26-2002
// Description:  This file contains the IO functions required by Micro Monitor
//				 that are unique to each CPU/Board combination
//
//
//=============================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "ether.h"
#include "stddefs.h"
#include "at91rm9200.h"
#include "at91rm9200_mem.h"
#include "at91rm9200_dbgu.h"
#include "at91rm9200_gpio.h"
#include "csb637_gpio.h"
#include "cpuio.h"
#include "sed1356_16bit.h"
#include "warmstart.h"

//#define CF_DEBUG

extern void	i2c_init(void);
extern void lcd_putchar(char c);
extern void putpsr(ulong);
extern ulong getpsr(void);

#if INCLUDE_SED
extern int sed135x_init(void);
extern void sed_putchar(char c);
extern void sed135x_on(void);
extern int sed_color_depth;
extern int sed135x_ok;
extern int sed_fg_color;		// 0 to 15, used as lookup into VGA color table
extern int sed_bg_color;		// 0 to 15, used as lookup into VGA color table
extern int sed_disp_mode_crt;
extern int mode900lq;
uchar bcolor;
#endif


// Delay for some usecs. - Not accurate, assumes ROM mode
// and no Cache.
void udelay(int delay)
{
  int i;
  for ( i = LOOPS_PER_USEC * delay; i ; i--)
	;
}

void lcd_bkl(uchar bright){

//	printf("Turning Backlight On.\n");

	udelay(1000);
	GPIO_out(LCD_BKL);
	GPIO_set(LCD_BKL);
}

/* InitUART():
 * This is the top-level UART initialization for the monitor. 
 */
void
InitUART(int baud)
{
	// clear error bits and reset tx and rx
	DBGU_REG(DBGU_CR) = DBGU_CR_RSTSTA
					   | DBGU_CR_RSTTX
					   | DBGU_CR_RSTRX;

	// clear any pending interrupts and mask all further interrupts
	DBGU_REG(DBGU_IDR) = DBGU_INT_ALL;
	DBGU_REG(DBGU_IMR) = 0x00000000;

	// Baud divisor = MCK / 16 / baud
	DBGU_REG(DBGU_BRGR) = ((MCK/16)/baud);	// integer only

	// set the port to no parity, no loopback
	DBGU_REG(DBGU_MR) = DBGU_MR_PAR_NONE | DBGU_MR_CHMODE_NORM;

	// Enable
	DBGU_REG(DBGU_CR) = DBGU_CR_TXEN | DBGU_CR_RXEN;
}

int
ConsoleBaudSet(int baud)
{
	// Baud divisor = MCK / 16 / baud
	DBGU_REG(DBGU_BRGR) = ((MCK/16)/baud);	// integer only
	return(0);
}

void
sed_init(void)
{
	// CS2 - S1D13506, 16-bit, 10 wait states
//	printf("Configure CS2.\n");
	SMC_REG(SMC_CSR2) = SMC_CSR_RWHOLD(2)
					  | SMC_CSR_RWSETUP(4)
					  | SMC_CSR_ACSS_0
					  | SMC_CSR_DRP_NORMAL
					  | SMC_CSR_DBW_16
					  | SMC_CSR_BAT_16_1
					  | SMC_CSR_TDF(0)
					  | SMC_CSR_WSEN
					  | SMC_CSR_NWS(10);

#if INCLUDE_SED
	// Initialize the display controller
	if (getenv("CSB900LQ")) {
		mode900lq = 1;
		PIXELS_PER_ROW = 240;
		PIXELS_PER_COL = 320;
		COLS_PER_SCREEN = 30;
		ROWS_PER_SCREEN = 20;
		SED_HOR_PULSE_WIDTH_LCD = 0x07;
		SED_HOR_NONDISP_LCD = 0x11;
		SED_VER_PULSE_WIDTH_LCD = 0x00;
		SED_VER_PULSE_START_LCD = 0x00;
		SED_VER_NONDISP_LCD = 0x03;
	}
	else { // broke...
		mode900lq = 0;
		PIXELS_PER_ROW = 640;
		PIXELS_PER_COL = 480;
		COLS_PER_SCREEN = 80;
		ROWS_PER_SCREEN = 30;
		SED_HOR_PULSE_WIDTH_LCD = 0x0b;
		SED_HOR_NONDISP_LCD = 0x13;
		SED_VER_PULSE_WIDTH_LCD = 0x01;
		SED_VER_PULSE_START_LCD = 0x09;
		SED_VER_NONDISP_LCD = 0x2c;
	}
	bcolor = 0;				// black
	sed_color_depth = 16;
	sed_fg_color = 14;			// Bright Yellow
	sed_bg_color = 1;	// Blue
	sed_disp_mode_crt = 0;		// default to LCD
	sed135x_init();
	if (sed135x_ok == 0)
		printf("sed135x_init() failed!\n");
	sed135x_on();
#endif

}

void
cf_init(void)
{
	// read the CD signal to see if we have a card
	if (GPIO_tst(CF_CD)){
#ifdef CF_DEBUG
		printf("No Compact FLASH Card Detected.\n");
#endif
		return;
	}


	// assign cs4 to CF
	EBI_REG(EBI_CSA) |= EBI_CSA_CS4_CF;

	// make sure the SMC controls CF_CE1/2 and we enable WAIT
//	PIOC_REG(PIO_PDR) = CF_CE1 | CF_CE2 | A_WAIT;

	// set CS4 for 10 wait states, 16-bits
	SMC_REG(SMC_CSR4) = SMC_CSR_RWHOLD(2)
					  | SMC_CSR_RWSETUP(4)
					  | SMC_CSR_ACSS_3
					  | SMC_CSR_DRP_NORMAL
					  | SMC_CSR_DBW_16
					  | SMC_CSR_BAT_16_2
					  | SMC_CSR_TDF(4)
					  | SMC_CSR_WSEN
					  | SMC_CSR_NWS(20);

	// enable power to the card
	GPIO_out(CF_RST );	// make it an output
	GPIO_set(CF_RST);	// hold card in reset

	GPIO_out(CF_PWR);	// make it an output
	GPIO_set(CF_PWR);	// turn card power on

	udelay (100);

	GPIO_clr(CF_RST);	// release card reset

#ifdef CF_DEBUG
	printf("Compact FLASH Card Detected.\n");
#endif

	udelay (10000);
}

int
devInit(int baud)
{
	// Initialize the GPIO's
	GPIO_init();

	// Initialize the UART
    InitUART(baud);

	// Initialize the CF Socket
	cf_init();

	// Initialize the I2C Port
	i2c_init();

	GPIO_out(VCC5_EN);
	GPIO_set(VCC5_EN);

	GPIO_out(USR_LED);
	GPIO_clr(USR_LED);
	udelay (10000);

	return 0;
}

void
initCPUio(void)
{
}

int
target_console_empty(void)
{
	if ((DBGU_REG(DBGU_CSR) & DBGU_INT_TXBUFE))
		return(1);
	return(0);
}

int
target_putchar(char c)
{
    /* Wait for transmit ready bit
	 */
    while(1) {
        if (DBGU_REG(DBGU_CSR) & DBGU_INT_TXRDY)
            break;
    }
	DBGU_REG(DBGU_THR) = c;
#if INCLUDE_SED
	sed_putchar(c);
#endif
    return((int)c);
}

int
target_gotachar()
{
    if (!(DBGU_REG(DBGU_CSR) & DBGU_INT_RXRDY))
        return(0);
    return(1);
}

/* target_getchar():
 */
int 
target_getchar(void)
{
    char    c;

    /* Retrieve character */
    c = DBGU_REG(DBGU_RHR);
    return((int)c);
}

void
target_reset(void)
{
	monrestart(INITIALIZE);
}

void
cacheInitForTarget()
{
    asm("   MRC p15, 0, r0, c1, c0, 0");
    asm("   ORR r0, r0, #0x1000");  /* bit 12 is ICACHE enable*/
    asm("   MCR p15, 0, r0, c1, c0, 0");

    /* Flush instruction cache */
    asm("   MCR p15, 0, r0, c7, c5, 0");
}

void
intsrestore(psr)
ulong     psr;
{
    putpsr(psr);
}

/*
 * Read the program status register (CPSR)
 * and set the FIQ and IRQ bits.
 */
ulong
intsoff()
{
    ulong  psr;

    psr = getpsr();

    /*
     * Set bit 6, bit 7 to disable interrupts.
     */
    putpsr(psr | 0x000000c0);
    return(psr);
}



