//=============================================================================
//
//      cpuio.c
//
//      CPU/Board Specific IO
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Date:         05-26-2002
// Description:  This file contains the IO functions required by Micro Monitor
//				 that are unique to each CPU/Board combination
//
//
//=============================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "ether.h"
#include "stddefs.h"
#include "uart16550.h"
#include "au1100_gpio.h"
#include "warmstart.h"

extern void GPIO_init();
extern int GPIO1_clr(int);
extern int GPIO1_set(int);
extern void tlbwrite(int, ulong);
extern int lcd_color_depth;

// structure to setup the MMU TLB
typedef struct
{
    unsigned long   EntryLo0;
    unsigned long   PageMask;
    unsigned long   EntryLo1;
    unsigned long   EntryHi;
} TLBEntry;

// Delay for some usecs. - Not accurate, assumes ROM mode
// and no Cache
void udelay(int delay)
{
  int i;
  for ( i = LOOPS_PER_USEC * delay; i ; i--)
	;
}

/* InitUART():
 * This is the top-level UART initialization for the monitor. 
 */
void
InitUART(int baud)
{
	/* set the failsafe serial port to a sane state */
	// enable the DUART
	UARTA(UART_ER) = UART_ER_CE;	// enable the clocks first
	UARTA(UART_ER) = UART_ER_CE | UART_ER_ME;	// now enable the module

	// Au1500 uses a dedicated clock divsor register.
	// clocks are derived from MEMCLK/2.
	UARTA(UART_CDR) = (((MEMCLK/2)/16)/baud);

	UARTA(UART_LCTL) = 0x03;	/* DLAB off, 8-1-N */
	UARTA(UART_MCTL) = 0x01;	/* drop all lines - raise DTR */
	UARTA(UART_IEN) = 0x00;		/* disable all interrupts */
	UARTA(UART_FCTL) = 0x01;	/* enable FIFO mode */
}

int
ConsoleBaudSet(int baud)
{
	UARTA(UART_CDR) = (((MEMCLK/2)/16)/baud);
	return(0);
}

int
devInit(int baud)
{

	// initialize the GPIO ports
	GPIO_init();

	// Initialize the UART first so we can print messages
    InitUART(baud);

	return 0;
}

/* target_console_empty():
 * Target-specific portion of flush_console() in chario.c.
 * This function returns 1 if there are no characters waiting to
 * be put out on the UART; else return 0 indicating that the UART
 * is still busy outputting characters from its FIFO.
 */
int
target_console_empty(void)
{
	if (UARTA(UART_LSTAT) & SIO_LSTAT_TRDY)
		return(0);
	return(1);
}

/* target_putchar():
 * Target-specific portion of character output to console used by
 * putchar() in chario.c.
 */
int
target_putchar(char c)
{
	ulong temp32;

    /* Wait for transmit ready bit
	 */
    while(1) {
        if (UARTA(UART_LSTAT) & SIO_LSTAT_TRDY)
            break;
    }

	temp32 = c & 0xff;
	UARTA(UART_TXD) = temp32;
    return((int)c);
}

/* target_gotachar():
 * Return 0 if no char is avaialable at UART rcv fifo; else 1.
 * Do NOT pull character out of fifo, just return status. 
 */
int
target_gotachar(void)
{
    if (UARTA(UART_LSTAT) & SIO_LSTAT_RRDY)
        return(1);
    return(0);
}

/* target_getchar():
 * Target-specific portion of character input used by getchar()
 * in chario.c.  This function should assume that when it is called,
 * the caller has already polled on gotachar() waiting for a character
 * presence; hence, this function only needs to return the character.
 */
int 
target_getchar(void)
{
    return((int)UARTA(UART_RXD));
}

void
initCPUio(void)
{
}

void
intsrestore(ulong msr)
{
}

ulong
intsoff()
{
	return(0);
}

void
target_reset(void)
{
	warmstart(INITIALIZE);
}

/* cacheInitForTarget():
 *	Called out of the generic CacheInit() function.
 */
void
cacheInitForTarget()
{
}

/* Stubs until we establish more exception handling...
 */
void
exc080(void)
{
	puts("EXCEPTION 080!\n");
	while(1);
}
void
exc100(void)
{
	puts("EXCEPTION 100!\n");
	while(1);
}
void
exc200(void)
{
	puts("EXCEPTION 200!\n");
	while(1);
}
void
exc280(void)
{
	puts("EXCEPTION 280!\n");
	while(1);
}
void
exc300(void)
{
	puts("EXCEPTION 300!\n");
	while(1);
}
void
exc380(void)
{
	puts("EXCEPTION 380!\n");
	while(1);
}
void
exc400(void)
{
	puts("EXCEPTION 400!\n");
	while(1);
}
void
exc480(void)
{
	puts("EXCEPTION 480!\n");
	while(1);
}
