/*
 *  pxa270_i2c_lowlevel.c
 *
 *  I2C adapter for the PXA I2C bus access.
 *
 *  Copyright (C) 2003 Technology Rescue.
 *	Author : David de la Cruz.
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *	History:
 *	15.04.2003 : CSB226 adaptation.
 */

#include "config.h"
#include "cli.h"
#include "i2c.h"
#include "genlib.h"
#include "stddefs.h"

#include "pxa270.h"
#include "pxa270_i2c.h"

extern void udelay(int);

/* place a byte in the transmit register */
unsigned long i2c_wr_byte(unsigned char value) 
{
        I2C_REG(I2C_IDBR) = value;
		return 0;
}

/* read byte in the receive register */
unsigned char i2c_rd_byte(void) 
{
        return (unsigned char) (0xff & I2C_REG(I2C_IDBR));
}

/*
 * void i2c_start(unsigned long icr_set, unsigned long icr_reset)
 *
 * icr_set : mask to set bits with start condition.
 * icr_reset : mask to reset bits with start condition.
 */
void i2c_start(unsigned long icr_set, unsigned long icr_reset)
{
	ulong icr;

	// read icr register.
	icr = I2C_REG(I2C_ICR);

	// set icr_set bits into icr register.
	icr |= ICR_START | icr_set;

	// reset icr_reset bits into icr register.
	icr	&= ~(ICR_STOP | icr_reset);

	// write back icr register.
	I2C_REG(I2C_ICR) = icr;
}

void i2c_repeat_start(unsigned char icr_set, unsigned char icr_reset)
{
	i2c_start(icr_set, icr_reset);
}

/*
 * void i2c_stop(unsigned long icr_set, unsigned long icr_reset)
 *
 * icr_set : mask to set bits with stop condition.
 * icr_reset : mask to reset bits with stop condition.
 */
void i2c_stop(unsigned long icr_set, unsigned long icr_reset)
{
	ulong icr;

	// read icr register.
	icr = I2C_REG(I2C_ICR);

	// set icr_set bits into icr register.
	icr |= ICR_STOP | icr_set;

	// reset icr_reset bits into icr register.
	icr &= ~(ICR_START | icr_reset);

	// write back icr register.
	I2C_REG(I2C_ICR) = icr;
}

/*
 * void i2c_midbyte(unsigned long icr_set, unsigned long icr_reset)
 *
 * icr_set : mask to set bits with stop condition.
 * icr_reset : mask to reset bits with stop condition.
 */
void i2c_midbyte(unsigned long icr_set, unsigned long icr_reset)
{
	ulong icr;

	// read icr register.
	icr = I2C_REG(I2C_ICR);

	// set icr_set bits into icr register.
	icr |= icr_set;

	// reset icr_reset bits into icr register.
	icr &= ~(ICR_START | ICR_STOP | icr_reset);

	// write back icr register.
	I2C_REG(I2C_ICR) = icr;
}

void i2c_abort(void)
{
	I2C_REG(I2C_ICR) |= ICR_MA;
}

int i2c_wait_bus_not_busy( void)
{
        int timeout = DEF_TIMEOUT;

        while (timeout-- && (I2C_REG(I2C_ISR) & ISR_IBB)) {
                udelay(100); /* wait for 100 us */
        }

        return (timeout<=0);
}

int i2c_wait_unit_not_busy( void)
{
        int timeout = DEF_TIMEOUT;

        while (timeout-- && (I2C_REG(I2C_ISR) & ISR_UB)) {
                udelay(100); /* wait for 100 us */
        }

        return (timeout<=0);
}


void i2c_transfer( int lastbyte, int receive, int midbyte)
{
	if( lastbyte)
	{
		if( receive==I2C_RECEIVE)
			i2c_stop(ICR_TB | ICR_ACKNAK,0);
		else 
			i2c_stop(ICR_TB | ICR_ALDIE,0);
	}
	else if( midbyte)
	{
		if( receive==I2C_RECEIVE)
			i2c_midbyte(ICR_TB ,ICR_ACKNAK);
		else 
			i2c_midbyte(ICR_TB | ICR_ALDIE,0);
	}
}

int i2c_check_status(ulong status_to_check)
{
	// poll ICPR register until I2C interrupt bit is 1.
	while(!(INT_REG(INT_ICPR) & 0x00040000))
	{	// wait.
	}

	// read I2C status register and check if ITE bit is set.
	if(I2C_REG(I2C_ISR) & status_to_check)
	{	// in this case, clear ISR_ITE in status reg.
		I2C_REG(I2C_ISR) = status_to_check;
	}
	else
	{	// else, return failure code.
		return -1;
	}
	return 0;
}

void i2c_init(void)
{
#ifdef I2C_DEBUG
	printf("Resetting I2C Controller Unit\n");
#endif

	/* reset the unit */
	I2C_REG(I2C_ICR) = ICR_UR;
        udelay(100);

	/* clear interrupt bits */
	I2C_REG(I2C_ISR) = I2C_ISR_INIT;

	/* restore reset bit to default state. */
	I2C_REG(I2C_ICR) &= ~ICR_UR;
        udelay(100);

	/* set the global I2C clock on */
	CLK_REG(CLK_CKEN) |= CKEN14_I2C;

	/* set our slave address */
	I2C_REG(I2C_ISAR) = I2C_PXA_SLAVE_ADDR;

	/* set control register values */
	I2C_REG(I2C_ICR) = I2C_ICR_INIT;

	/* clear interrupt bits */
	I2C_REG(I2C_ISR) = I2C_ISR_INIT;

	/* enable unit */
	I2C_REG(I2C_ICR) |= ICR_IUE;
        udelay(100);

#ifdef I2C_DEBUG
	i2c_show_reg();
#endif
}

void i2c_show_reg( void)
{
	printf("I2C Controller Unit regs\n");
	printf("Status Register  %lx\n",I2C_REG(I2C_ISR));
	printf("Control Register %lx\n",I2C_REG(I2C_ICR));
	printf("Data Buffer      %lx\n",I2C_REG(I2C_IDBR));
	printf("Bus Monitor Reg  %lx\n",I2C_REG(I2C_IBMR));
}
