//==========================================================================
//
//      cpu_lcd.c
//
// Author(s):   Luis Torrico - Cogent Computer Systems, Inc.
// Date:        12-10-2008
// Description:	Init Code for Freescale MX31 LCD Controller
// NOTE 		Only 16-bit mode has been tested!
//
//==========================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "stddefs.h"
#include "cli.h"
#include "cpu_gpio.h"
#include "umongpio.h"
#include "cpu_lcd.h"
#include "cpu_ipu.h"
#include "mx31_lcd.h"
#include "font8x16.h"
#include "fb_draw.h"
#include "warmstart.h"

//--------------------------------------------------------------------------
// function prototypes and externs
//
void fbdev_init(void);

extern void udelay(int);
extern int GPIO_clr(int);
extern int GPIO_set(int);
extern int GPIO_tst(int);
extern int GPIO_out(int);
extern int GPIO_in(int);

// ADDED FOR WRITING CHARACTERS TO THE DISPLAY
void lcd_putchar(char c);
void lcd_writechar(uchar c);
void lcd_clr_row(int char_row);
void lcd_clr_scr(void);
void lcd_switch_buffer(void);

// globals to keep track of foreground, background colors and x,y position
int lcd_color_depth;		// 4, 8 or 16
int lcd_fg_color;			// 0 to 15, used as lookup into VGA color table
int lcd_bg_color;			// 0 to 15, used as lookup into VGA color table
int lcd_col;				// current column, 0 to COLS_PER_SCREEN - 1
int lcd_row;				// current row, 0 to (ROWS_PER_SCREEN * 2) - 1
int lcd_tst_mode = 0;
ulong lcd_fb_offset;	   	// current offset into frame buffer for lcd_putchar

static ulong bpp_to_pixfmt(int bpp)
{
        ulong pixfmt = 0;
        switch (bpp) {
        case 24:
                pixfmt = IPU_PIX_FMT_BGR24;
                break;
        case 32:
                pixfmt = IPU_PIX_FMT_BGR32;
                break;
        case 16:
                pixfmt = IPU_PIX_FMT_RGB565;
                break;
        }
        return pixfmt;
}

//#define LCD_DBG
//--------------------------------------------------------------------------
// fbdev_init
//
// This function sets up the i.MX31 LCD Controller, to be used
// as uMon's frame buffer device.
//
void
fbdev_init(void)
{
    int delay_value;
    delay_value = 100;
	ipu_di_signal_cfg_t sig_pol;
	void lcd_on(void);

	if (StateOfMonitor != INITIALIZE)
		return;

	lcd_color_depth = 16;
	lcd_fg_color = vga_lookup[LCD_FG_DEF];
	lcd_bg_color = vga_lookup[LCD_BG_DEF];

	sig_pol.datamask_en = 0;
	sig_pol.clksel_en = 0;
	sig_pol.clkidle_en = 0;
	sig_pol.data_pol = 1;
	sig_pol.clk_pol = 1;
	sig_pol.enable_pol = 1;
	sig_pol.Hsync_pol = 1;
	sig_pol.Vsync_pol = 1;

	// Enable IPU and CSI clocks
	IPU_REG(DI_HSP_CLK_PER) = 0x00100010L;
	// Set SDC refresh channels as high priority
	IPU_REG(IDMAC_CHA_PRI) = 0x0000C000L;
	// Set to max back to back burst requests
	IPU_REG(IDMAC_CONF) = 0x00000070L;
	
/*	
 * @param       panel           The type of panel.
 *
 * @param       refreshRate     Desired refresh rate for the panel in Hz. 
 *
 * @param       width           The width of panel in pixels.
 *
 * @param       height          The height of panel in pixels.
 *
 * @param       pixel_fmt       Input parameter for pixel format of buffer. Pixel 
 *                              format is a FOURCC ASCII code.
 *
 * @param       hStartWidth     The number of pixel clocks between the HSYNC
 *                              signal pulse and the start of valid data.
 *
 * @param       hSyncWidth      The width of the HSYNC signal in units of pixel 
 *                              clocks.
 *
 * @param       hEndWidth       The number of pixel clocks between the end of
 *                              valid data and the HSYNC signal for next line.
 *
 * @param       vStartWidth     The number of lines between the VSYNC
 *                              signal pulse and the start of valid data.
 *
 * @param       vSyncWidth      The width of the VSYNC signal in units of lines 
 *
 * @param       vEndWidth       The number of lines between the end of valid
 *                              data and the VSYNC signal for next frame.
 *
 * @param       sig             Bitfield of signal polarities for LCD interface. */
	
	ipu_sdc_init_panel(IPU_PANEL_TFT, 60, 
                           480, 272,
//                           IPU_PIX_FMT_RGB565,
						   IPU_PIX_FMT_RGB666,
						   43,
                           41,
                           2,
                           12,
                           10,
                           2,
						    sig_pol);
//	ipu_sdc_init_panel(IPU_PANEL_TFT, 60, 
//                           800, 480,
////                           IPU_PIX_FMT_RGB565,
//						   IPU_PIX_FMT_RGB666,
//						   90,
//                           30,
//                           60,
//                           12,
//                           10,
//                           28,
//						    sig_pol);
	ipu_sdc_set_window_pos(MEM_SDC_BG, 0, 0);
	ipu_sdc_set_global_alpha(0, 0);
	ipu_sdc_set_color_key(MEM_SDC_BG, 0, 0);
	ipu_init_channel(MEM_SDC_BG, 0);
	ipu_init_channel_buffer(MEM_SDC_BG, IPU_INPUT_BUFFER, 
						bpp_to_pixfmt(16),
	                    480, 272,
	                    480,
	                    IPU_ROTATE_NONE,
	                    (void*)LCD_BUF_ADD, 
	                    (void*)LCD_BUF_ADD);
//	ipu_init_channel_buffer(MEM_SDC_BG, IPU_INPUT_BUFFER, 
//						bpp_to_pixfmt(16),
//	                    800, 480,
//	                    800,
//	                    IPU_ROTATE_NONE,
//	                    (void*)LCD_BUF_ADD, 
//	                    (void*)LCD_BUF_ADD);
	ipu_select_buffer(MEM_SDC_BG, IPU_INPUT_BUFFER, 0);
	ipu_enable_channel(MEM_SDC_BG);
    
	lcd_on();

	printf("MX31 LCD Initialization Complete.\n");

	return;
}

/* fbdev_setstart():
 * Used by uMon's FBI interface to establish the starting address of
 * the frame buffer memory.
 */
void
fbdev_setstart(long offset)
//fbdev_setstart()
{
	// Enable IPU and CSI clocks
	IPU_REG(DI_HSP_CLK_PER) = 0x00100010L;

	ipu_sdc_set_window_pos(MEM_SDC_BG, 0, 0);
	ipu_init_channel(MEM_SDC_BG, 0);
	ipu_init_channel_buffer(MEM_SDC_BG, IPU_INPUT_BUFFER, 
						bpp_to_pixfmt(16),
	                    480, 272,
	                    480,
	                    IPU_ROTATE_NONE,
	                    (void*)offset, 
	                    (void*)offset);
//	ipu_init_channel_buffer(MEM_SDC_BG, IPU_INPUT_BUFFER, 
//						bpp_to_pixfmt(16),
//	                    800, 480,
//	                    800,
//	                    IPU_ROTATE_NONE,
//	                    (void*)LCD_BUF_ADD, 
//	                    (void*)LCD_BUF_ADD);
	ipu_select_buffer(MEM_SDC_BG, IPU_INPUT_BUFFER, 0);
	ipu_enable_channel(MEM_SDC_BG);
    
	return;
}

//--------------------------------------------------------------------------
// lcd_clear_screen()
//
ulong lcd_clear_screen()
{
	volatile unsigned short *buffer = LCD_BUF_ADD;
	int i;
	
	for (i = 0; i < PIXELS_PER_ROW * PIXELS_PER_COL; i++)
	{
        //buffer[i] = 0xEE31;
        buffer[i] = 0x0;
	}
	
	return 0;
}

//--------------------------------------------------------------------------
// lcd_on()
void lcd_on()
{
	// Make sure it's off first
//	lcd_off();
    
	udelay(10000);
    GPIO_set(LCD_BKL);
}

//--------------------------------------------------------------------------
// lcd_tst()
//
//void lcd_tst(void)
int lcd_tst(int argc,char *argv[])
{
	volatile ushort wr16, rd16;
	int i, x, opt;
	int no_wait = 0;
	int init_only = 0;
	char c;
	lcd_color_depth = 16;

//	lcd_off();

    while ((opt=getopt(argc,argv,"clnsxd:4,8,16")) != -1) {
        switch(opt) {
			case 'd':	// set the color depth
                switch(*optarg) {
            	    case '4':
	        	        lcd_color_depth = 4;
						printf("Forcing 4bpp Mode!\n");
        	            break;
    	            case '8':
            	        lcd_color_depth = 8;
						printf("Forcing 8bpp Mode!\n");
                    	break;
	                default:	// test with 16bpp
	                    lcd_color_depth = 16;
						printf("Forcing 16bpp Mode!\n");
						break;
                }
	            break;
	        case 'n':   // no waiting for keypress - fastest operation
				no_wait = 1;
				printf("No Keypress Mode, Must Reset System to Stop!\n");
				break;
	        case 'x':   // init only 
				no_wait = 1;
				printf("Initializing LCD, Skipping testsp!\n");
				init_only = 1;
				break;
			default:	// test with current mode
				break;
		}
	}

	// get the new parameters into the LCD controller
	fbdev_init();
	//lcd_init();
	//lcd_on();

	if (init_only) return 0;

	printf("Frame Buffer R/W...");
	// do an address=data read/write test on the frame buffer
	// PIXELS_PER_COL * PIXELS_PER_ROW is the highest pixel.  
	// Multiply by bits_per_pixel (sed_color_depth), then
	// divide by 8 to get the actual byte count.
	for (i = LCD_ROW_SIZE(lcd_color_depth); i < LCD_FB_SIZE(lcd_color_depth) + LCD_ROW_SIZE(lcd_color_depth); i += 2){	
		LCD_BUF(i) = i & 0xffff;
		rd16 = LCD_BUF(i);
		if(rd16 != (i & 0xffff)){
			printf("Fail at 0x%08x, WR 0x%08x, RD 0x%04lx!\n",LCD_BUF_ADD + i, i, (ulong)rd16);
			return -1;
		}
	} 

	printf("OK!, Press key to continue.\n");

	c = getchar();

	printf("Frame Buffer Start: 0x%08x, End 0x%08x\n",LCD_BUF_ADD, 
			LCD_BUF_ADD + LCD_FB_SIZE(lcd_color_depth) + LCD_ROW_SIZE(lcd_color_depth));
	if (no_wait)
	{
		printf("Begin Full Screen Color Test.\n");
		while(1){
			// fill the frame buffer with incrementing color values
			for (x = 0; x < 16; x++){
				switch (lcd_color_depth){
					case 4:  wr16 = x | x << 4 | x << 8 | x << 12; break;
					case 8:  wr16 = x | x << 8; break;
					default: wr16 = vga_lookup[x]; break;	// 16-bits bypasses the lookup table
				}
				for (i = 0; i < LCD_FB_SIZE(lcd_color_depth); i += 2){	
					LCD_BUF(i) = wr16;
				}
			} // for x
		} // while
	} // no_wait
	else
	{
		printf("Begin Full Screen Color Test, Press any key to go to next color, \'x\' to end.\n");
		while(1){
			// fill the frame buffer with incrementing color values
			for (x = 0; x < 16; x++){
				switch (lcd_color_depth){
					case 4:  wr16 = x | x << 4 | x << 8 | x << 12; break;
					case 8:  wr16 = x | x << 8; break;
					default: wr16 = vga_lookup[x]; break;	// 16-bits bypasses the lookup table
				}
				for (i = LCD_ROW_SIZE(lcd_color_depth); i < LCD_FB_SIZE(lcd_color_depth) + LCD_ROW_SIZE(lcd_color_depth); i += 2){	
					LCD_BUF(i) = wr16;
				}
				c = getchar();
				if (c == 'x') goto lcd_tst_next;
			} // for x
		} // while
	} // else no keycheck test

	lcd_tst_next:

	// write a one pixel border around the screen
	lcd_fg_color = 15;	// VGA Bright White
	fb_draw_line2(LEFT, TOP, RIGHT, TOP, lcd_fg_color);
	fb_draw_line2(LEFT, TOP, LEFT, BOTTOM, lcd_fg_color);
	fb_draw_line2(LEFT, BOTTOM, RIGHT, BOTTOM, lcd_fg_color);	// bottom left to bottom right
	fb_draw_line2(RIGHT, TOP, RIGHT, BOTTOM, lcd_fg_color);	// bottom right to top right
	// draw an x
	fb_draw_line2(LEFT, TOP, RIGHT, BOTTOM, lcd_fg_color);
	fb_draw_line2(LEFT, BOTTOM, RIGHT, TOP, lcd_fg_color);
	// draw 3 circles at the center of the screen, one inside the other
	lcd_fg_color = 12;	// VGA Bright Red
	fb_draw_circle2(CENTER_X, CENTER_Y, 100, lcd_fg_color);
	lcd_fg_color = 10;	// VGA Bright Green
	fb_draw_circle2(CENTER_X, CENTER_Y, 66, lcd_fg_color);
	lcd_fg_color = 9;	// VGA Bright Blue
	fb_draw_circle2(CENTER_X, CENTER_Y, 33, lcd_fg_color);

	return 0;
}

// fb_set_pixel sets a pixel to the specified color.
// This is target specific and is called from the generic
// fb_draw.c functions
void fb_set_pixel(ulong X, ulong Y, uchar color)
{
   	// Make sure the specified pixel is valid.
#if 0
   	if((X < 0) || (X >= PIXELS_PER_ROW) || (Y < 0) || (Y >= PIXELS_PER_COL))
   	{   
		printf("fb_set_pixel() bad X (%ld) or Y (%ld)!\n", X, Y);
		return;
   	}
#else
	if (X < 0)
		X = 0;
	else {
		if (X >= PIXELS_PER_ROW)
			X = PIXELS_PER_ROW-1;
	}
	if (Y < 0)
		Y = 0;
	else {
		if (Y >= PIXELS_PER_COL)
			Y = PIXELS_PER_COL-1;
	}
#endif

	LCD_BUF(LCD_GET_PIXEL_ADD(X, Y)) = vga_lookup[color];
}

char *lcd_tstHelp[] = {
    "MCIMX31 LCD controller test",
    "-[n,x,d[4,8,16]]",
	"The user may set color depth to run the test at.",
	"The frame buffer R/W test will test all of the frame ",
	"buffer regardless of depth.",
    "Options...",
    "  -n    run test without keycheck - CAUTION: RESET SYSTEM TO STOP!",
	"  -d4   run test, force a depth of 4-bits/pixel",
	"  -d8   run test, force a depth of 8-bits/pixel",
	"  -d16  run test, force a depth of 16-bits/pixel",
	"  -x    init only, do not run frame buffer tests",
	"",
	"  No options, default to current mode and depth.",
	0
};

