//==========================================================================
//
// mc9328mx21_gpio.c
//
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Contributors: 
// Date:         09/18/2005
// Description:  This file contains code to intialize the MC9328MX21 GPIO
//               section as well as functions for manipulating the GPIO 
//               bits
//
//--------------------------------------------------------------------------

#include "config.h"
#include "cpuio.h"
#include "stddefs.h"
#include "genlib.h"
#include "mc9328mx21.h"
#include "cpu_gpio.h"    // pull in target board specific header

//#define GPIO_DBG

//--------------------------------------------------------
// GPIO_init()
//
// This function sets the startup state for the MC9328MX21 GPIO 
// registers as used on the target CPU.  Refer to cpu_gpio.h 
// for a description of the default values.  Here we just put
// them into the chip in the following order:
// 1. Port x DDIR  - Data Direction Register
// 2. Port x DR    - Data Register
// 3. Port x GPR   - Primary/Secondary Assignment Register
// 4. Port x GPIUS - GPIO Usage Register
// 4. Port x OCR1  - GPIO Output Configuration Register (bits 0-15)
// 4. Port x OCR2  - GPIO Output Configuration Register (bits 16-31)
// 4. Port x ICR1  - GPIO Input Configuration Register (bits 0-15)
// 4. Port x ICR   - GPIO Input Configuration Register (bits 16-31)

void GPIO_init()
{
	// Port A
	GPIOA_REG(GPIO_DDIR)  = PORTA_DDIR;
	GPIOA_REG(GPIO_DR)    = PORTA_DR;
	GPIOA_REG(GPIO_GPR)   = PORTA_GPR;
	GPIOA_REG(GPIO_GIUS)  = PORTA_GPIUS;
	GPIOA_REG(GPIO_OCR1)  = PORTA_OCR1;
	GPIOA_REG(GPIO_OCR2)  = PORTA_OCR2;
	GPIOA_REG(GPIO_ICR1)  = PORTA_ICR1;
	GPIOA_REG(GPIO_ICR2)  = PORTA_ICR2;
	GPIOA_REG(GPIO_ICONFA1)  = 0;
	GPIOA_REG(GPIO_ICONFA2)  = 0;

	// Port B
	GPIOB_REG(GPIO_DDIR)  = PORTB_DDIR;
	GPIOB_REG(GPIO_DR)    = PORTB_DR;
	GPIOB_REG(GPIO_GPR)   = PORTB_GPR;
	GPIOB_REG(GPIO_GIUS)  = PORTB_GPIUS;
	GPIOB_REG(GPIO_OCR1)  = PORTB_OCR1;
	GPIOB_REG(GPIO_OCR2)  = PORTB_OCR2;
	GPIOB_REG(GPIO_ICR1)  = PORTB_ICR1;
	GPIOB_REG(GPIO_ICR2)  = PORTB_ICR2;
	GPIOB_REG(GPIO_ICONFA1)  = 0;
	GPIOB_REG(GPIO_ICONFA2)  = 0;

	// Port C
	GPIOC_REG(GPIO_DDIR)  = PORTC_DDIR;
	GPIOC_REG(GPIO_DR)    = PORTC_DR;
	GPIOC_REG(GPIO_GPR)   = PORTC_GPR;
	GPIOC_REG(GPIO_GIUS)  = PORTC_GPIUS;
	GPIOC_REG(GPIO_OCR1)  = PORTC_OCR1;
	GPIOC_REG(GPIO_OCR2)  = PORTC_OCR2;
	GPIOC_REG(GPIO_ICR1)  = PORTC_ICR1;
	GPIOC_REG(GPIO_ICR2)  = PORTC_ICR2;
	GPIOC_REG(GPIO_ICONFA1)  = 0;
	GPIOC_REG(GPIO_ICONFA2)  = 0;

	// Port D
	GPIOD_REG(GPIO_DDIR)  = PORTD_DDIR;
	GPIOD_REG(GPIO_DR)    = PORTD_DR;
	GPIOD_REG(GPIO_GPR)   = PORTD_GPR;
	GPIOD_REG(GPIO_GIUS)  = PORTD_GPIUS;
	GPIOD_REG(GPIO_OCR1)  = PORTD_OCR1;
	GPIOD_REG(GPIO_OCR2)  = PORTD_OCR2;
	GPIOD_REG(GPIO_ICR1)  = PORTD_ICR1;
	GPIOD_REG(GPIO_ICR2)  = PORTD_ICR2;
	GPIOD_REG(GPIO_ICONFA1)  = 0;
	GPIOD_REG(GPIO_ICONFA2)  = 0;

	// Port E
	GPIOE_REG(GPIO_DDIR)  = PORTE_DDIR;
	GPIOE_REG(GPIO_DR)    = PORTE_DR;
	GPIOE_REG(GPIO_GPR)   = PORTE_GPR;
	GPIOE_REG(GPIO_GIUS)  = PORTE_GPIUS;
	GPIOE_REG(GPIO_OCR1)  = PORTE_OCR1;
	GPIOE_REG(GPIO_OCR2)  = PORTE_OCR2;
	GPIOE_REG(GPIO_ICR1)  = PORTE_ICR1;
	GPIOE_REG(GPIO_ICR2)  = PORTE_ICR2;
	GPIOE_REG(GPIO_ICONFA1)  = 0;
	GPIOE_REG(GPIO_ICONFA2)  = 0;

	// Port F
	GPIOF_REG(GPIO_DDIR)  = PORTF_DDIR;
	GPIOF_REG(GPIO_DR)    = PORTF_DR;
	GPIOF_REG(GPIO_GPR)   = PORTF_GPR;
	GPIOF_REG(GPIO_GIUS)  = PORTF_GPIUS;
	GPIOF_REG(GPIO_OCR1)  = PORTF_OCR1;
	GPIOF_REG(GPIO_OCR2)  = PORTF_OCR2;
	GPIOF_REG(GPIO_ICR1)  = PORTF_ICR1;
	GPIOF_REG(GPIO_ICR2)  = PORTF_ICR2;
	GPIOF_REG(GPIO_ICONFA1)  = 0;
	GPIOF_REG(GPIO_ICONFA2)  = 0;

	// Function Multiplexing Control Register
	SYS_REG(SYS_FMCR) = FMCR_REG;
}

//--------------------------------------------------------
// GPIO_set()
//
// This function sets the desired bit passed in.
// NOTE: We do not test to see if setting the bit
// would screw up any alternate functions.  Use 
// this function with caution!
//

int GPIO_set(int gpio_bit)
{
	// quick sanity test
#ifdef GPIO_DBG
	printf("GPIO_set %d.\n", gpio_bit);
#endif
	if (gpio_bit > 192) return -1;

	if (gpio_bit < 32)
	{
		// Port A
		GPIOA_REG(GPIO_DR) |= (1 << (gpio_bit - 0));
	}
	else if (gpio_bit < 64)
	{
		// Port B
		GPIOB_REG(GPIO_DR) |= (1 << (gpio_bit - 32));
	}
	else if (gpio_bit < 96)
	{
		// Port C
		GPIOC_REG(GPIO_DR) |= (1 << (gpio_bit - 64));
	}
	else if (gpio_bit < 128)
	{
		// Port D
		GPIOD_REG(GPIO_DR) |= (1 << (gpio_bit - 96));
	}
	else if (gpio_bit < 160)
	{
		// Port E
		GPIOE_REG(GPIO_DR) |= (1 << (gpio_bit - 128));
	}
	else
	{
		// Port F
		GPIOF_REG(GPIO_DR) |= (1 << (gpio_bit - 160));
	}
	return 0;
}

//--------------------------------------------------------
// GPIO_clr()
//
// This function clears the desired bit passed in.
//

int GPIO_clr(int gpio_bit)
{
#ifdef GPIO_DBG
	printf("GPIO_clr %d.\n", gpio_bit);
#endif
 	// quick sanity test
	if (gpio_bit > 192) return -1;

	if (gpio_bit < 32)
	{
		// Port A
		GPIOA_REG(GPIO_DR) &= ~(1 << (gpio_bit - 0));
	}
	else if (gpio_bit < 64)
	{
		// Port B
		GPIOB_REG(GPIO_DR) &= ~(1 << (gpio_bit - 32));
	}
	else if (gpio_bit < 96)
	{
		// Port C
		GPIOC_REG(GPIO_DR) &= ~(1 << (gpio_bit - 64));
	}
	else if (gpio_bit < 128)
	{
		// Port D
		GPIOD_REG(GPIO_DR) &= ~(1 << (gpio_bit - 96));
	}
	else if (gpio_bit < 160)
	{
		// Port E
		GPIOE_REG(GPIO_DR) &= ~(1 << (gpio_bit - 128));
	}
	else
	{
		// Port F
		GPIOF_REG(GPIO_DR) &= ~(1 << (gpio_bit - 160));
	}
    return 0;
}
//--------------------------------------------------------
// GPIO_tst()
//
// This function returns the state of desired bit passed in.
// It does not test to see if it's an input or output and thus
// can be used to verify if an output set/clr has taken place
// as well as for testing an input state.
//

int GPIO_tst(int gpio_bit)
{
#ifdef GPIO_DBG
	printf("GPIO_tst %d.\n", gpio_bit);
#endif
 	// quick sanity test
	if (gpio_bit > 192) return -1;

	if (gpio_bit < 32)
	{
		// Port A
		if (GPIOA_REG(GPIO_SSR) & (1 << (gpio_bit - 0))) return 1;
	}
	else if (gpio_bit < 64)
	{
		// Port B
		if (GPIOB_REG(GPIO_SSR) & (1 << (gpio_bit - 32))) return 1;
	}
	else if (gpio_bit < 96)
	{
		// Port C
		if (GPIOC_REG(GPIO_SSR) & (1 << (gpio_bit - 64))) return 1;
	}
	else if (gpio_bit < 128)
	{
		// Port D
		if (GPIOD_REG(GPIO_SSR) & (1 << (gpio_bit - 96))) return 1;
	}
	else if (gpio_bit < 160)
	{
		// Port E
		if (GPIOE_REG(GPIO_SSR) & (1 << (gpio_bit - 128))) return 1;
	}
	else
	{
		// Port F
		if (GPIOF_REG(GPIO_SSR) & (1 << (gpio_bit - 160))) return 1;
	}
	return 0; // bit was not set
}

//--------------------------------------------------------
// GPIO_in()
//
// This function changes the direction of the desired bit 
// to input.  NOTE: We do not test to see if changing the
// direction of the bit would screw up anything.  Use this
// function with caution!
//
// This only worlks if the GPIO has been defined as a GPIO
// during init.  It iwll not override the init setting, only
// change the direction bit

int GPIO_in(int gpio_bit)
{
#ifdef GPIO_DBG
	printf("GPIO_in %d.\n", gpio_bit);
#endif
	// quick sanity test
	if (gpio_bit > 192) return -1;

	if (gpio_bit < 32)
	{
		// Port A
		GPIOA_REG(GPIO_DDIR) &= ~(1 << (gpio_bit - 0));
	}
	else if (gpio_bit < 64)
	{
		// Port B
		GPIOB_REG(GPIO_DDIR) &= ~(1 << (gpio_bit - 32));
	}
	else if (gpio_bit < 96)
	{
		// Port C
		GPIOC_REG(GPIO_DDIR) &= ~(1 << (gpio_bit - 64));
	}
	else if (gpio_bit < 128)
	{
		// Port D
		GPIOD_REG(GPIO_DDIR) &= ~(1 << (gpio_bit - 96));
	}
	else if (gpio_bit < 160)
	{
		// Port E
		GPIOE_REG(GPIO_DDIR) &= ~(1 << (gpio_bit - 128));
	}
	else
	{
		// Port F
		GPIOF_REG(GPIO_DDIR) &= ~(1 << (gpio_bit - 160));
	}
    return 0;
}

//--------------------------------------------------------
// GPIO_out()
//
// This function changes the direction of the desired bit 
// to output.  NOTE: We do not test to see if changing the
// direction of the bit would screw up anything.  Use this
// function with caution!
//
// This only worlks if the GPIO has been defined as a GPIO
// during init.  It iwll not override the init setting, only
// change the direction bit
int GPIO_out(int gpio_bit)
{
#ifdef GPIO_DBG
	printf("GPIO_out %d.\n", gpio_bit);
#endif
	// quick sanity test
	if (gpio_bit > 192) return -1;

	if (gpio_bit < 32)
	{
		// Port A
		GPIOA_REG(GPIO_DDIR) |= (1 << (gpio_bit - 0));
	}
	else if (gpio_bit < 64)
	{
		// Port B
		GPIOB_REG(GPIO_DDIR) |= (1 << (gpio_bit - 32));
	}
	else if (gpio_bit < 96)
	{
		// Port C
		GPIOC_REG(GPIO_DDIR) |= (1 << (gpio_bit - 64));
	}
	else if (gpio_bit < 128)
	{
		// Port D
		GPIOD_REG(GPIO_DDIR) |= (1 << (gpio_bit - 96));
	}
	else if (gpio_bit < 160)
	{
		// Port E
		GPIOE_REG(GPIO_DDIR) |= (1 << (gpio_bit - 128));
	}
	else
	{
		// Port F
		GPIOF_REG(GPIO_DDIR) |= (1 << (gpio_bit - 160));
	}
	return 0;
}

