//==========================================================================
//
//      cpu_i2c.c
//
// Author(s):   Michael Kelly - Cogent Computer Systems, Inc.
// Date:        03/26/2003
// Description:	CSB335 - LH79520 Single Board IIC Routines 
//
//==========================================================================

#include "config.h"
#include "cpuio.h"
#include "cpu_gpio.h"
#include "mpc5121e.h"
#include "genlib.h"
#include "stddefs.h"
#include "cli.h"
#include "umongpio.h"

//--------------------------------------------------------------------------
// I2C Macros
//

#define I2C_SCL_CLR				GPIO_out(GPIO_7)
#define I2C_SCL_SET				GPIO_in(GPIO_7) 	

#define I2C_SDA_CLR				GPIO_out(GPIO_1)		
#define I2C_SDA_SET				GPIO_in(GPIO_1) 		

//#define I2C_SCL_CLR				GPIO_clr(GPIO_7)
//#define I2C_SCL_SET				GPIO_set(GPIO_7) 	

//#define I2C_SDA_CLR				GPIO_clr(GPIO_1)		
//#define I2C_SDA_SET				GPIO_set(GPIO_1) 		

#define I2C_SDA_RD				GPIO_tst(GPIO_1)
#define I2C_SCL_RD				GPIO_tst(GPIO_7)

#define I2C_DELAY 				udelay(1)

#define CDCE925_I2C_ADD	0x64	// I2C Address of CDCE925

//--------------------------------------------------------------------------
// function prototypes
//
extern int GPIO_set(int);
extern int GPIO_clr(int);
extern int GPIO_tst(int);
extern int GPIO_in(int);
extern int GPIO_out(int);
ulong i2c_init(void);
ulong i2c_wr_device(uchar dev, uchar reg, uchar data);
ulong i2c_rd_device(uchar dev, uchar reg, uchar *data);
ulong i2c_wr_byte(uchar data);
uchar i2c_rd_byte(void);
void i2c_start(void);
void i2c_stop(void);
int i2c(int argc,char *argv[]);

extern void udelay(int delay);

int cdce925_init(int);

//--------------------------------------------------------------------------
// Programmed values
//
// The following array data is derived from the
// the CDCE7925 Configuration Utility from TI.

// Used for Setting clocks for the CSB781
// PSC_CLK = 24.576MHz, PHY_CLK = 25MHz,
// USB_CLK = 25MHz, CPU_CLK = 24.576MHz
const uchar cdce925_data[] = {
0x81,	// Byte 00 - 1000 0001
0x00,	// Byte 01 - 0000 0000
0x34,	// Byte 02 - 0011 0100
0x01,	// Byte 03 - 0000 0001
0x02,	// Byte 04 - 0000 0010
0x50,	// Byte 05 - 0101 0000
0x60,	// Byte 06 - 0110 0000
0x00,	// Byte 07 - 0000 0000
0x00,	// Byte 08 - 0000 0000
0x00,	// Byte 09 - 0000 0000
0x00,	// Byte 10 - 0000 0000
0x00,	// Byte 11 - 0000 0000
0x00,	// Byte 12 - 0000 0000
0x00,	// Byte 13 - 0000 0000
0x00,	// Byte 14 - 0000 0000
0x00,	// Byte 15 - 0000 0000
0x00,	// Byte 16 - 0000 0000
0x00,	// Byte 17 - 0000 0000
0x00,	// Byte 18 - 0000 0000
0x00,	// Byte 19 - 0000 0000
0x5D,	// Byte 20 - 0101 1101
0x02,	// Byte 21 - 0000 0010
0x08,	// Byte 22 - 0000 1000
0x00,	// Byte 23 - 0000 0000
0xC3,	// Byte 24 - 1100 0011
0x53,	// Byte 25 - 0101 0011
0x52,	// Byte 26 - 0101 0010
0x07,	// Byte 27 - 0000 0111
0x00,	// Byte 28 - 0000 0000
0x40,	// Byte 29 - 0100 0000
0x02,	// Byte 30 - 0000 0010
0x08,	// Byte 31 - 0000 1000
0x00,	// Byte 32 - 0000 0000
0x00,	// Byte 33 - 0000 0000
0x00,	// Byte 34 - 0000 0000
0x00,	// Byte 35 - 0000 0000
0xDD,	// Byte 36 - 1101 1101
0x02,	// Byte 37 - 0000 0010
0x01,	// Byte 38 - 0000 0001
0x00,	// Byte 39 - 0000 0000
0x00,	// Byte 40 - 0000 0000
0x40,	// Byte 41 - 0100 0000
0x02,	// Byte 42 - 0000 0010
0x08,	// Byte 43 - 0000 1000
0x00,	// Byte 44 - 0000 0000
0x40,	// Byte 45 - 0100 0000
0x02,	// Byte 46 - 0000 0010
0x08	// Byte 47 - 0000 1000
};

//--------------------------------------------------------------------------
// cdce925_init
//
// This function copies the data from cdce925_data into the cdce925
// using low level I2C primitives.  This code assumes the I2C hardware
// has been initialized already.

int cdce925_init(int mode)
{
	
	int i;
	uchar temp = 0xff;
		
	// copy the data into the CDCE925
	for (i = 0; i < 48; i++){

		// configure CDCE925 - upper bit of register must be 1 for byte read/writes
		  if(i2c_wr_device(CDCE925_I2C_ADD, (i | 0x80), cdce925_data[i]))
	      return -1;
	}

// wait for CDCE925 to settle to new values before we use them
	udelay(10);

	//printf("CDCE925 Initialization Complete\n");

	return 0;
}

//--------------------------------------------------------------------------
// i2c_init()
//
// Nothing to do but set both bits 0, but input
ulong i2c_init()
{
    GPIO_out(GPIO_7);
    GPIO_clr(GPIO_7);
    GPIO_in(GPIO_7);
	
    GPIO_out(GPIO_1);
    GPIO_clr(GPIO_1);
    GPIO_in(GPIO_1);
	return 0;

}

//--------------------------------------------------------------------------
// i2c_wr_device()
//
// This function writes an 8-bit value to the I2C device at the requested
// register.
//
ulong i2c_wr_device(uchar dev, uchar reg, uchar data)
{

	// issue a start command
	i2c_start();

	// write the 7-bit device address with write = 0
	if(i2c_wr_byte((dev << 1) & 0xfe)){
		return -1;
	}
	// Write the 8-bit register address
	if(i2c_wr_byte(reg)){
		return -1;
	}
	// Write the 8-bit data value
	if(i2c_wr_byte(data)){
		return -1;
	}

	// issue a stop
	i2c_stop();
	
	return 0;
}

//--------------------------------------------------------------------------
// i2c_rd_device()
//
// This function reads an 8-bit value from the I2C device at the requested
// register.
//
ulong i2c_rd_device(uchar dev, uchar reg, uchar *data)
{

	// issue a start command
	i2c_start();

	// write the 7-bit device address with write = 0
	if(i2c_wr_byte((dev << 1) & 0xfe)){
		return -1;
	}
	// Write the 8-bit register address
	if(i2c_wr_byte(reg)){
		return -1;
	}
	// repeat the start command
	i2c_start();
	// write the 7-bit device address again plus data direction (read = 1)
	if(i2c_wr_byte((dev << 1) | 0x01)){
		return -1;
	}
	*data = i2c_rd_byte();

	// issue a stop
	i2c_stop();

	return 0;
}

//--------------------------------------------------------------------------
// i2c_wr_byte()
//
// This function writes an 8-bit value to the I2C bus, MSB first.
// Data is written by changing SDA during SCL low, then bringing
// SCL high.  SCL is returned low to setup for the next transition.
//
ulong i2c_wr_byte(uchar data)
{

	int i;

	for (i = 0; i < 8; i++){
		if (data & 0x80) {
			// write a 1 bit
			I2C_SDA_SET;
			I2C_DELAY;
			I2C_SCL_SET;
			I2C_DELAY;
			I2C_SCL_CLR;
			I2C_DELAY;
		}
		else {
			// write a 0 bit
			I2C_SDA_CLR;
			I2C_DELAY;
			I2C_SCL_SET;
			I2C_DELAY;
			I2C_SCL_CLR;
			I2C_DELAY;
		}
		data = data << 1;
	}
	// Release SDA, bring SCL high, then read SDA.
	// A low indicates an acknowledge.
	I2C_SDA_SET;
	I2C_DELAY;
   	I2C_SCL_SET;
	I2C_DELAY;
	if(I2C_SDA_RD){	// a high means no ack
		// re-enable SDA for output
	   	I2C_SCL_CLR;
		I2C_DELAY;
		return -1;
	}

   	I2C_SCL_CLR;
	I2C_DELAY;

	return 0;
}

//--------------------------------------------------------------------------
// i2c_rd_byte()
//
// This function reads an 8-bit data value from the I2C bus, MSB first. 
// Data is read from SDA after each low to high SCL transition.
//
uchar i2c_rd_byte()
{

	int i;
	uchar volatile data;

	data = 0;

	for (i = 0; i < 8; i++){
		data = data << 1;
		data = data & 0xfe;
		// clock the data out of the slave
		I2C_SCL_SET;
		I2C_DELAY;
		// check it
		if (I2C_SDA_RD){
			data = data | 0x01;
		}
		I2C_SCL_CLR;
		I2C_DELAY;
	}
	// generate an extra SCL transition
	// The slave generates no acknowledge for reads.
	I2C_SCL_SET;
	I2C_DELAY;
	I2C_SCL_CLR;
	I2C_DELAY;

	return data;
}


//--------------------------------------------------------------------------
// i2c_start()
//
// This function issues an I2C start command which is a high to low
// transition on SDA while SCL is high.
//
void i2c_start()
{

	I2C_SDA_SET;
	I2C_DELAY;
   	I2C_SCL_SET;
	I2C_DELAY;
   	I2C_SDA_CLR;
	I2C_DELAY;
   	I2C_SCL_CLR;
	I2C_DELAY;
	I2C_SDA_SET;
	I2C_DELAY;
}

//--------------------------------------------------------------------------
// i2c_stop()
//
// This function issues an I2C stop command which is a low to high
// transition on SDA while SCL is high.
//
void i2c_stop()
{

	I2C_SDA_CLR;
	I2C_DELAY;
	I2C_SCL_SET;
	I2C_DELAY;
	I2C_SDA_SET;
	I2C_DELAY;
	I2C_SCL_CLR;
	I2C_DELAY;
}

char *i2cHelp[] = {
	" This command allows the user to read ",
	" and write devices on the i2c bus. \n ",
    " Usage:",
    " i2c -[w] {device} {register} {val/count}",
    " Options...",
    " -w write val to device/register",
	" no options, read from device/register up to count",
	0
};

int i2c(int argc,char *argv[])
{
	int i, opt;
	int write = 0;
	uchar dev, reg, data, count;

	// Initialize the I2C Port
	//i2c_init();

    while ((opt=getopt(argc,argv,"w")) != -1) {
    if (opt == 'w') write = 1;
	}

	// make sure we have the right number of paramters
	if (argc < (optind+3))
		return(CMD_PARAM_ERROR);

	dev = (uchar) strtoul(argv[optind],(char **)0,0);
	reg = (uchar) strtoul(argv[optind+1],(char **)0,0);

	// 3rd arg is the data value if it's a write, count if it's a read
	data = (uchar) strtoul(argv[optind+2],(char **)0,0);
	count = data;
	// do it
	if (write)
	{
		printf("Writing 0x%02x to Device 0x%02x @ Register 0x%02x.\n", data, dev, reg);
    	if(i2c_wr_device(dev, reg, data))
    	{
			printf("I2C Bus Failure - Check Paramters!\n");
    		return (CMD_FAILURE);
		}
	}
	else
	{
		for (i = 0; i < count; i++)
    	{
			printf("Read Device 0x%02x, Register 0x%02x = ", dev, reg + i);
			if(i2c_rd_device(dev, reg + i, &data))
			{
				printf("I2C Bus Failure - Check Paramters!\n");
    			return (CMD_FAILURE);
			}
			printf("0x%02x.\n", data);
		}
	}
	return(CMD_SUCCESS);
}

