//==========================================================================
//
//      mpc5121e_lcd.c
//
// Author(s):   Luis Torrico - Cogent Computer Systems, Inc.
// Date:        04-22-2009
// Description:	Init Code for Freescale MPC5121e LCD Controller
// NOTE 		Only 16-bit mode has been tested!
//
//==========================================================================

#include "config.h"
#include "cpuio.h"
#include "stddefs.h"
#include "genlib.h"
#include "mpc5121e.h"
#include "cpu_gpio.h"
#include "vga_lookup.h"
#include "font8x16.h"
#include "fb_draw.h"
#include "warmstart.h"

//--------------------------------------------------------------------------
// function prototypes and externs
//
void fbdev_init(void);

extern void udelay(int);
extern int GPIO_clr(int);
extern int GPIO_set(int);
extern int GPIO_tst(int);
extern int GPIO_out(int);
extern int GPIO_in(int);

// ADDED FOR WRITING CHARACTERS TO THE DISPLAY
void lcd_putchar(char c);
void lcd_writechar(uchar c);
void lcd_clr_row(int char_row);
void lcd_clr_scr(void);
void lcd_switch_buffer(void);

// globals to keep track of foreground, background colors and x,y position
int lcd_color_depth;		// 4, 8 or 16
int lcd_fg_color;			// 0 to 15, used as lookup into VGA color table
int lcd_bg_color;			// 0 to 15, used as lookup into VGA color table
int lcd_col;				// current column, 0 to COLS_PER_SCREEN - 1
int lcd_row;				// current row, 0 to (ROWS_PER_SCREEN * 2) - 1
int lcd_tst_mode = 0;
ulong lcd_fb_offset;	   	// current offset into frame buffer for lcd_putchar

//#define LCD_DBG
//--------------------------------------------------------------------------
// fbdev_init
//
// This function sets up the MPC5121e LCD Controller, to be used
// as uMon's frame buffer device.
//
void
fbdev_init(void)
{
	//ushort temp16;
	unsigned char *gamma_table_base;
	unsigned int i, j;

	//if (StateOfMonitor != INITIALIZE)
	//	return;

	lcd_color_depth = 16;
	lcd_fg_color = vga_lookup[LCD_FG_DEF];
	lcd_bg_color = vga_lookup[LCD_BG_DEF];

    // Set Display Size
    DIU_REG(DIU_DISP_SIZE) = DIU_DELTA_Y(PIXELS_PER_COL) | DIU_DELTA_X(PIXELS_PER_ROW);

	// Set HSYNC timing
    DIU_REG(DIU_HSYN_PARA) = DIU_BP_H(LCD_H_BACK) | DIU_PW_H(LCD_H_WIDTH) | DIU_FP_H(LCD_H_FRONT);

	// Set VSYNC timing
    DIU_REG(DIU_VSYN_PARA) = DIU_BP_V(LCD_V_BACK) | DIU_PW_V(LCD_V_WIDTH) | DIU_FP_V(LCD_V_FRONT);

	// Set control signal polarities, invert syncs
    DIU_REG(DIU_SYN_POL) = 0x00000000;
 
	// Set base address for gamma
    DIU_REG(DIU_GAMMA) = LCD_GAMMA_ADD;

	// Disable pallette
    DIU_REG(DIU_PALETTE) = 0x00000000;

	// Disable cursor
    DIU_REG(DIU_CURSOR) = 0x00000000;
    DIU_REG(DIU_CURS_POS) = DIU_CURSOR_Y(PIXELS_PER_COL) | DIU_CURSOR_X(PIXELS_PER_ROW);

	/* Prep for DIU init  - gamma table */
	gamma_table_base = LCD_GAMMA_ADD;
	for (i = 0; i <= 2; i++)
		for (j = 0; j <= 255; j++)
			*gamma_table_base++ = j;

	// Set area descriptor settings
	LCD_ADR(AREA1_BITMAP_ADDR)  = BE2LE(LCD_BUF_ADD);
	LCD_ADR(AREA1_PIX_FORMAT)  = BE2LE(0x18310565);
	LCD_ADR(AREA1_SS_GA)  = BE2LE(AREA1_SS_GA_ALPHA(0xff) | AREA1_SS_GA_DELTA_YS(PIXELS_PER_COL) | AREA1_SS_GA_DELTA_XS(PIXELS_PER_ROW));
	LCD_ADR(AREA1_AOI_SIZE)  = BE2LE(AREA1_AOIS_FLIP(0x00) | AREA1_AOIS_DELTA_YI(PIXELS_PER_COL) | AREA1_AOIS_DELTA_XI(PIXELS_PER_ROW));
	LCD_ADR(AREA1_AOI_OFFSET)  = BE2LE(AREA1_AOIO_DELTA_YI(0x00) | AREA1_AOIO_DELTA_XI(0x00));
	LCD_ADR(AREA1_DIS_OFFSET)  = BE2LE(AREA1_DISO_OFFSET_YD(0x00) | AREA1_DISO_OFFSET_YD(0x00));
	LCD_ADR(AREA1_CHROMA_MAX)  = 0x00000000;
	LCD_ADR(AREA1_CHROMA_MIN)  = 0xffffff00;
	LCD_ADR(AREA1_NEXT_AD)  = 0x00000000;

	// Set base address for plane 1
    DIU_REG(DIU_DESC_1) = LCD_ADR_ADD;
    DIU_REG(DIU_DESC_2) = LCD_ADR_ADD;
    DIU_REG(DIU_DESC_3) = LCD_ADR_ADD;

	// Select DIU mode 1
    DIU_REG(DIU_MODE) = 0x00000001;

	// Turn on backlight
	GPIO_set(GPIO_6);
	GPIO_out(GPIO_6);
		
	printf("MPC5121e LCD Initialization Complete.\n");

	return;
}

/* fbdev_setstart():
 * Used by uMon's FBI interface to establish the starting address of
 * the frame buffer memory.
 */
void
fbdev_setstart(long offset)
{
	// Set area descriptor settings
	LCD_ADR(AREA1_BITMAP_ADDR)  = BE2LE(offset);
	LCD_ADR(AREA1_PIX_FORMAT)  = BE2LE(0x18310565);
	LCD_ADR(AREA1_SS_GA)  = BE2LE(AREA1_SS_GA_ALPHA(0xff) | AREA1_SS_GA_DELTA_YS(PIXELS_PER_COL) | AREA1_SS_GA_DELTA_XS(PIXELS_PER_ROW));
	LCD_ADR(AREA1_AOI_SIZE)  = BE2LE(AREA1_AOIS_FLIP(0x00) | AREA1_AOIS_DELTA_YI(PIXELS_PER_COL) | AREA1_AOIS_DELTA_XI(PIXELS_PER_ROW));
	LCD_ADR(AREA1_AOI_OFFSET)  = BE2LE(AREA1_AOIO_DELTA_YI(0x00) | AREA1_AOIO_DELTA_XI(0x00));
	LCD_ADR(AREA1_DIS_OFFSET)  = BE2LE(AREA1_DISO_OFFSET_YD(0x00) | AREA1_DISO_OFFSET_YD(0x00));
	LCD_ADR(AREA1_CHROMA_MAX)  = 0x00000000;
	LCD_ADR(AREA1_CHROMA_MIN)  = 0xffffff00;
	LCD_ADR(AREA1_NEXT_AD)  = 0x00000000;

	// Set base address for plane 1
    DIU_REG(DIU_DESC_1) = LCD_ADR_ADD;
    DIU_REG(DIU_DESC_2) = LCD_ADR_ADD;
    DIU_REG(DIU_DESC_3) = LCD_ADR_ADD;

	// Select DIU mode 1
    DIU_REG(DIU_MODE) = 0x00000001;

	// Turn on backlight
	GPIO_set(GPIO_6);
	GPIO_out(GPIO_6);
		
	return;
}
//#if INCLUDE_LCD
char *lcd_tstHelp[] = {
    "MPC5121e LCD controller test",
    "-[n,x,d[4,8,16]]",
	"The user may set color depth to run the test at.",
	"The frame buffer R/W test will test all of the frame ",
	"buffer regardless of depth.",
    "Options...",
    "  -n    run test without keycheck - CAUTION: RESET SYSTEM TO STOP!",
	"  -d4   run test, force a depth of 4-bits/pixel",
	"  -d8   run test, force a depth of 8-bits/pixel",
	"  -d16  run test, force a depth of 16-bits/pixel",
	"  -x    init only, do not run frame buffer tests",
	"",
	"  No options, default to current mode and depth.",
	0
};

int lcd_tst(int argc,char *argv[])
{
	volatile ushort wr16, rd16;
	int i, x, opt;
	int no_wait = 0;
	int init_only = 0;
	char c;

	lcd_tst_mode = 1;

    while ((opt=getopt(argc,argv,"clnsxd:4,8,16")) != -1) {
        switch(opt) {
			case 'd':	// set the color depth
                switch(*optarg) {
            	    case '4':
	        	        lcd_color_depth = 4;
						printf("Forcing 4bpp Mode!\n");
        	            break;
    	            case '8':
            	        lcd_color_depth = 8;
						printf("Forcing 8bpp Mode!\n");
                    	break;
	                default:	// test with 16bpp
	                    lcd_color_depth = 16;
						printf("Forcing 16bpp Mode!\n");
						break;
                }
	            break;
	        case 'n':   // no waiting for keypress - fastest operation
				no_wait = 1;
				printf("No Keypress Mode, Must Reset System to Stop!\n");
				break;
	        case 'x':   // init only 
				no_wait = 1;
				printf("Initializing LCD, Skipping testsp!\n");
				init_only = 1;
				break;
			default:	// test with current mode
				break;
		}
	}

	// get the new parameters into the LCD controller
	fbdev_init();

	if (init_only) return 0;

	printf("Frame Buffer R/W...");
	// do an address=data read/write test on the frame buffer
	// PIXELS_PER_COL * PIXELS_PER_ROW is the highest pixel.  
	// Multiply by bits_per_pixel (sed_color_depth), then
	// divide by 8 to get the actual byte count.
	for (i = 0; i < LCD_FB_SIZE(lcd_color_depth) + LCD_ROW_SIZE(lcd_color_depth); i += 2){	
		LCD_BUF(i) = i & 0xffff;
		rd16 = LCD_BUF(i);
		if(rd16 != (i & 0xffff)){
			printf("Fail at 0x%08x, WR 0x%08x, RD 0x%04lx!\n",LCD_BUF_ADD + i, i, (ulong)rd16);
			return -1;
		}
	} 

	printf("OK!, Press key to continue.\n");

	c = getchar();

	printf("Frame Buffer Start: 0x%08x, End 0x%08x\n",LCD_BUF_ADD, 
			LCD_BUF_ADD + LCD_FB_SIZE(lcd_color_depth) + LCD_ROW_SIZE(lcd_color_depth));
	if (no_wait)
	{
		printf("Begin Full Screen Color Test.\n");
		while(1){
			// fill the frame buffer with incrementing color values
			for (x = 0; x < 16; x++){
				switch (lcd_color_depth){
					case 4:  wr16 = x | x << 4 | x << 8 | x << 12; break;
					case 8:  wr16 = x | x << 8; break;
					default: wr16 = vga_lookup[x]; break;	// 16-bits bypasses the lookup table
				}
				for (i = 0; i < LCD_FB_SIZE(lcd_color_depth); i += 2){	
					LCD_BUF(i) = wr16;
				}
			} // for x
		} // while
	} // no_wait
	else
	{
		printf("Begin Full Screen Color Test, Press any key to go to next color, \'x\' to end.\n");
		while(1){
			// fill the frame buffer with incrementing color values
			for (x = 0; x < 16; x++){
				switch (lcd_color_depth){
					case 4:  wr16 = x | x << 4 | x << 8 | x << 12; break;
					case 8:  wr16 = x | x << 8; break;
					default: wr16 = vga_lookup[x]; break;	// 16-bits bypasses the lookup table
				}
				for (i = 0; i < LCD_FB_SIZE(lcd_color_depth) + LCD_ROW_SIZE(lcd_color_depth); i += 2){	
					LCD_BUF(i) = wr16;
				}
				c = getchar();
				if (c == 'x') goto lcd_tst_next;
			} // for x
		} // while
	} // else no keycheck test

	lcd_tst_next:

	// write a one pixel border around the screen
	lcd_fg_color = 15;	// VGA Bright White
	fb_draw_line2(LEFT, TOP, RIGHT, TOP, lcd_fg_color);
	fb_draw_line2(LEFT, TOP, LEFT, BOTTOM, lcd_fg_color);
	fb_draw_line2(LEFT, BOTTOM, RIGHT, BOTTOM, lcd_fg_color);	// bottom left to bottom right
	fb_draw_line2(RIGHT, TOP, RIGHT, BOTTOM, lcd_fg_color);	// bottom right to top right
	// draw an x
	fb_draw_line2(LEFT, TOP, RIGHT, BOTTOM, lcd_fg_color);
	fb_draw_line2(LEFT, BOTTOM, RIGHT, TOP, lcd_fg_color);
	// draw 3 circles at the center of the screen, one inside the other
	lcd_fg_color = 12;	// VGA Bright Red
	fb_draw_circle2(CENTER_X, CENTER_Y, 100, lcd_fg_color);
	lcd_fg_color = 10;	// VGA Bright Green
	fb_draw_circle2(CENTER_X, CENTER_Y, 66, lcd_fg_color);
	lcd_fg_color = 9;	// VGA Bright Blue
	fb_draw_circle2(CENTER_X, CENTER_Y, 33, lcd_fg_color);

	return 0;
}

// fb_set_pixel sets a pixel to the specified color.
// This is target specific and is called from the generic
// fb_draw.c functions
void fb_set_pixel(ulong X, ulong Y, uchar color)
{
   	// Make sure the specified pixel is valid.
#if 0
   	if((X < 0) || (X >= PIXELS_PER_ROW) || (Y < 0) || (Y >= PIXELS_PER_COL))
   	{   
		printf("fb_set_pixel() bad X (%ld) or Y (%ld)!\n", X, Y);
		return;
   	}
#else
	if (X < 0)
		X = 0;
	else {
		if (X >= PIXELS_PER_ROW)
			X = PIXELS_PER_ROW - 1;
	}
	if (Y < 0)
		Y = 0;
	else {
		if (Y >= PIXELS_PER_COL)
			Y = PIXELS_PER_COL - 1;
	}
#endif

	LCD_BUF(LCD_GET_PIXEL_ADD(X, Y)) = vga_lookup[color];
}
//#endif

