/* etherdev.c:
 * Ethernet device layer using SCC4 of the MPC852T.
 *
 * General notice:
 * This code is part of a boot-monitor package developed as a generic base
 * platform for embedded system designs.  As such, it is likely to be
 * distributed to various projects beyond the control of the original
 * author.  Please notify the author of any enhancements made or bugs found
 * so that all may benefit from the changes.  In addition, notification back
 * to the author will allow the new user to pick up changes that may have
 * been made by other users after this version of the code was distributed.
 * 
 * Author:	Ed Sutter
 * email:	esutter@lucent.com
 * phone:	908-582-2351
 */
#include "config.h"
#if (INCLUDE_ETHERNET == 1) && (USE_FEC == 0)
#include "cpuio.h"
#include "ether.h"
#include "stddefs.h"
#include "genlib.h"
#include "arch_ppc.h"
#include "mpc860.h"
#include "masks860.h"

extern	char *processARP();
extern	EPPC *IMMRBase;

int EtherRLCnt, EtherDEFCnt, EtherLCCnt, EtherDRVRETRYCnt;
int eninit();

static uchar xbuf[XBUFCNT][1600];	/* Transmit & receive buffers pointed */
static uchar rbuf[RBUFCNT][1600];	/* to by the buffer descriptors. */

struct	enet_bd	*tbptr, *rbptr, *prevtbptr;

/* ShowEtherdevStats():
	Called by the command "ether stat" on the monitor's command line.
	This function and a common ShowEthernetStats() are used to dump
	generic as well as device-specific ethernet driver statistics.
*/
void
ShowEtherdevStats()
{
	printf("SCC4-based ethernet interface\n");
	printf("Driver retry count:      %d\n",EtherDRVRETRYCnt);
	printf("Retry_count_exceeded:    %d times\n",EtherRLCnt);
	printf("Late-collision count:    %d\n",EtherLCCnt);
	printf("Defer-indication count:  %d\n",EtherDEFCnt);
}

/* EtherdevStartup():
	Called by EthernetStartup().  Initialize all device-specific 
	counters, reset and initialized the ethernet device.
*/
int
EtherdevStartup(int verbose)
{
	EtherRLCnt = 0;
	EtherLCCnt = 0;
	EtherDEFCnt = 0;
	EtherDRVRETRYCnt = 0;

	/* Put ethernet controller in reset: */
	enreset();

	/* Initialize controller: */
	eninit();

	return(0);
}

/* DisableEtherdev():
	Called by DisableEthernet() to do the device-specific portion of 
	the turn-down.
*/
void
DisableEtherdev()
{
	enreset();
    IMMRBase->pio_pcpar &= ~(0x0008);    /* set port C -> RTS4 = *TENA */
    IMMRBase->cpmi_cimr &= ~CIMR_SCC4;
	EtherIsActive = 0;
	IPMonCmdActive = 0;
}

void
enreset()
{
	int	timeout;

	/* Issue CPM Command to init RX/TX params for SCC4: */
	/* (wait for completion) */
	IMMRBase->cp_cr = 0x0001;
	timeout = 500000;
	while(IMMRBase->cp_cr & 0x0001) {
		timeout--;
		if (timeout <= 0) {
			printf("enreset() timeout\n");
			break;
		}
	}
}

/* eninit():
	These steps are basically taken from the MPC860 User's Manual (rev1)
	section 28.22 "SCC Ethernet Programming Example".
*/
int
eninit()
{
	ushort	mac;
	uchar	*cp;
	int		i, timeout;
	struct	enet_bd	*tbd, *rbd;

	/* Set IMMR pointing to the MPC860's internal register memory map. */
   	IMMRBase = (EPPC *)(ppcMfimmr() & 0xFFFF0000);

	/* Following steps outlined in section 27.22 (SCC Ethernet Programming
	 * Example) of the MPC866 User Manual...
	 * Our hardware uses SCC4, with pins configured as follows: 
	 *		TXD4 is muxed with PA[8]
	 *		RXD4 is muxed with PA[9]
	 *		RTS4 is muxed with PC[12] 
	 *		CTS4 is muxed with PC[5]
	 *		CD4 is muxed with PC[4]
	 * 		TIN3/CLK5 is PA[3], TOUT3/CLK6 is PA[2]
	 */
	/* TXD4 & TXD4 config...  */
	IMMRBase->pio_papar |=  0x30c0;	
	IMMRBase->pio_padir |=  0x30c0;
	IMMRBase->pio_padat &= ~0x30c0;

	/* CTS4 & CD4 config (don't enable RTS4 yet)...  */
	IMMRBase->pio_pcpar |=  0x0c00;
	IMMRBase->pio_pcdir &=  ~0x0c00;
	IMMRBase->pio_pcdat &= ~0x0c00;

	/* Route clocks based on this hardware: */
	IMMRBase->si_sicr  &= 0x00ffffff;	/* Clear SCC4 info */
	IMMRBase->si_sicr  |= 0x25000000;	/* recv = CLK5, xmit = CLK6 */

	/* Initialize SDMA configu to 0x0001 (step 7): */
	IMMRBase->dma_sdcr = 0x0001;

	/* Configure General SCC Mode Register for ethernet: */
	IMMRBase->scc_regs[SCC4_REG].scc_gsmr_h = 0;
	IMMRBase->scc_regs[SCC4_REG].scc_gsmr_l = 0x1088000c;

	/* Initialize Buffer Descriptors (located in DPRAM): */
	/* Refer to mpc860.h for the layout of DPRAM as used by MicroMonitor. */
	tbd = tbptr = IMMRBase->qcp_or_ud.umd.e_bdtbl;
	rbd = rbptr = &IMMRBase->qcp_or_ud.umd.e_bdtbl[XBUFCNT];

	/* Set parameter ram to point to buffer descriptors in DPRAM: */
	/* Note that this will generate a compiler warning because we are */
	/* casting a pointer to a short. This can be ignored because these */
	/* two entries are offsets relative to the base of the IMMR, not the */
	/* actual address. */
	IMMRBase->PRAM[PAGE1].enet_scc.rbase = (ushort)rbd;
	IMMRBase->PRAM[PAGE1].enet_scc.tbase = (ushort)tbd;

	for (i=0;i<XBUFCNT;i++,tbd++) {
		tbd->status = 0;
		tbd->length = 0;
		tbd->baddr = xbuf[i];
		if (i == (XBUFCNT-1))
			tbd->status |= BD_WRAP;	/* Set WRAP bit on final BD */
	}
	for(i=0;i<RBUFCNT;i++,rbd++) {
		rbd->status = BD_READY;
		rbd->length = 0;
		rbd->baddr = rbuf[i];
		if (i == (RBUFCNT-1))
			rbd->status |= BD_WRAP;	/* Set WRAP bit on final BD */
	}

	/* Issue CPM Command to init RX/TX params for SCC4: */
	/* (wait for completion) */
	IMMRBase->cp_cr = 0x0001;
	timeout = 500000;
	while(IMMRBase->cp_cr & 0x0001) {
		timeout--;
		if (timeout <= 0)
			break;
	}

	/* Set RFCR,TFCR -- Rx,Tx Function Code */
	/* Normal Operation and Motorola byte ordering */
	IMMRBase->PRAM[PAGE1].enet_scc.rfcr = 0x10;  
	IMMRBase->PRAM[PAGE1].enet_scc.tfcr = 0x10;  

	/* MRBLR = MAX buffer length (first value larger than 1518 evenly */
	/* divisible by 4). */
	IMMRBase->PRAM[PAGE1].enet_scc.mrblr = 1520;    

	/* Initialize based on example in mpc860 manual. */
	IMMRBase->PRAM[PAGE1].enet_scc.c_pres = 0xffffffff;
	IMMRBase->PRAM[PAGE1].enet_scc.c_mask = 0xdebb20e3;
	IMMRBase->PRAM[PAGE1].enet_scc.crcec = 0;
	IMMRBase->PRAM[PAGE1].enet_scc.alec = 0;
	IMMRBase->PRAM[PAGE1].enet_scc.disfc = 0;
	IMMRBase->PRAM[PAGE1].enet_scc.pads = 0x8888;
	IMMRBase->PRAM[PAGE1].enet_scc.ret_lim = 15;
	IMMRBase->PRAM[PAGE1].enet_scc.mflr = 1518;
	IMMRBase->PRAM[PAGE1].enet_scc.minflr = 64;
	IMMRBase->PRAM[PAGE1].enet_scc.maxd1 = 1520;
	IMMRBase->PRAM[PAGE1].enet_scc.maxd2 = 1520;
	IMMRBase->PRAM[PAGE1].enet_scc.gaddr1 = 0;
	IMMRBase->PRAM[PAGE1].enet_scc.gaddr2 = 0;
	IMMRBase->PRAM[PAGE1].enet_scc.gaddr3 = 0;
	IMMRBase->PRAM[PAGE1].enet_scc.gaddr4 = 0;

	/* Load the PADDR registers with this board's ethernet address. */
	cp = (uchar *)&mac;
	cp[0] = BinEnetAddr[5];
	cp[1] = BinEnetAddr[4];
	IMMRBase->PRAM[PAGE1].enet_scc.paddr_h = mac;
	cp[0] = BinEnetAddr[3];
	cp[1] = BinEnetAddr[2];
	IMMRBase->PRAM[PAGE1].enet_scc.paddr_m = mac;
	cp[0] = BinEnetAddr[1];
	cp[1] = BinEnetAddr[0];
	IMMRBase->PRAM[PAGE1].enet_scc.paddr_l = mac;

	/* This is step 22... */
	IMMRBase->PRAM[PAGE1].enet_scc.p_per = 0;
	IMMRBase->PRAM[PAGE1].enet_scc.iaddr1 = 0;
	IMMRBase->PRAM[PAGE1].enet_scc.iaddr2 = 0;
	IMMRBase->PRAM[PAGE1].enet_scc.iaddr3 = 0;
	IMMRBase->PRAM[PAGE1].enet_scc.iaddr4 = 0;
	IMMRBase->PRAM[PAGE1].enet_scc.taddr_h = 0;
	IMMRBase->PRAM[PAGE1].enet_scc.taddr_m = 0;
	IMMRBase->PRAM[PAGE1].enet_scc.taddr_l = 0;

	/* Clear any events: */
	IMMRBase->scc_regs[SCC4_REG].scc_scce = 0xffff;

	/* Enable RXF events: */
	IMMRBase->scc_regs[SCC4_REG].scc_sccm = ENET_SCCM_RXF;

	IMMRBase->cpmi_cimr |= CIPR_SCC4;

	IMMRBase->scc_regs[SCC4_REG].scc_dsr = 0xd555;
	IMMRBase->scc_regs[SCC4_REG].scc_psmr = 0x080a;

	/* Enable RTS4...  */
	IMMRBase->pio_pcpar |=  0x0008;
	IMMRBase->pio_pcdir &=  ~0x0008;
	IMMRBase->pio_pcdat &= ~0x0008;

	/* 35. Enable transmit/receive: */
	IMMRBase->scc_regs[SCC4_REG].scc_gsmr_l = 0x1088003c;
	return(0);
}

/* enselftest():
	1. Enable ethernet interface.
	2. Put the PHY in loopback.
	3. Send a packet, and see if that packet is received.
	4. Disable the ethernet interface.
	Return 1 if passed, else -1.
*/
int
enselftest(int verbose)
{
	printf("Self-test not available\n");
	return(1);
}

/* geteinbuf():
	Get ethernet input buffer...
	Do whatever is necessary to pull the packet out of the ethernet device
	and pass that packet (along with the size of the packet) to the
	processPACKET() function.
*/
static int
geteinbuf()
{
	int	cnt;
	struct 	enet_bd *rp;

	cnt = 0;

	/* Stay in this loop while there is an active receive buffer. */
	/* If receive buffer is empty, just return. */
	while(!(rbptr->status & BD_READY)) {

		EtherRFRAMECnt++;
		processPACKET((struct ether_header *)rbptr->baddr,rbptr->length);

		/* Make the buffer ready. */
		rbptr->status |= BD_READY;

		/* If WRAP, then back to top; else, increment to next BD. */
		if (rbptr->status & BD_WRAP)
			rbptr = &IMMRBase->qcp_or_ud.umd.e_bdtbl[XBUFCNT];
		else
			rbptr++;
		cnt++;
	}
#if 1
	/* Make sure no other receive buffers are ready... */
	rp = rbptr;
	while(1) {
		if (rp->status & BD_WRAP)
			rp = &IMMRBase->qcp_or_ud.umd.e_bdtbl[XBUFCNT];
		else
			rp++;
		if (rp == rbptr)
			break;
		if (!(rp->status & BD_READY))
			printf("rbptr buffer ready out of order!\n");
	}
#endif
	return(cnt);
}

/* polletherdev():
	Called by pollethernet() to do the device-specific portion of the
	ethernet polling.  This code is the first step in processing incoming
	ethernet packets.  If a packet is available, the functino geteinbuf()
	(get ethernet input buffer) is called to pass the packet to the higher
	level processPACKET() function.
	IMPORTANT NOTE:
	This function MUST be re-entrant.  There is a possibility that the
	processPACKET() function will ultimately call polletherdev() again.
	This means that once the packet has been detected in the device, it 
	is important that this code do whatever is necessary to assure that
	the same packet will not be seen as available to the nested caller.
	In other words...
	clear the packet from the device prior to calling processPACKET().
*/
int
polletherdev()
{
	vushort	event;
	int		pcnt;

	pcnt = 0;
	event = IMMRBase->scc_regs[SCC4_REG].scc_scce;
	IMMRBase->scc_regs[SCC4_REG].scc_scce = event;
	if (event & ENET_SCCE_RXF) {	/* Received frame interrupt */
		geteinbuf();
		pcnt++;
	}
	return(pcnt);
}

/* getXmitBuffer():
	Called by various points in the IP/UDP/ICMP code for a buffer that
	will ultimately be the next output packet.
*/
uchar *
getXmitBuffer()
{
	/* Wait for the next buffer descriptor to become available... */
	while(tbptr->status & BD_READY);
	return(tbptr->baddr);
}

/* sendBuffer():
   Set the flag within the current transmit buffer descriptor indicating
   that it is ready to be transmitted.  Then increment the BD pointer to 
   the next buffer descriptor.
*/
int
sendBuffer(int length)
{
#if INCLUDE_ETHERVERBOSE
	if (EtherVerbose &  SHOW_OUTGOING)
		printPkt((struct ether_header *)tbptr->baddr,length,ETHER_OUTGOING);
#endif

	/* Keep stats: */
	if (EtherXFRAMECnt) {
		if (prevtbptr->status & BD_DEF)
			EtherDEFCnt++;
		if (prevtbptr->status & BD_LC)
			EtherLCCnt++;
		if (prevtbptr->status & BD_RL)
			EtherRLCnt++;
	}

	EtherXFRAMECnt++;

	/* Make the current transmit buffer ready and populate the length. */
	tbptr->length = (ushort)length;
#if 1
	tbptr->status |= (BD_PAD | BD_READY | BD_LAST | BD_TC);
#else
	tbptr->status |= (BD_READY | BD_LAST | BD_TC);
#endif

	/* Keep track of this BD so that the status can be looked at when the */
	/* next transmission is done. */
	prevtbptr = tbptr;

	/* If WRAP, then back to top; else, just increment to next BD. */
	if (tbptr->status & BD_WRAP)
		tbptr = IMMRBase->qcp_or_ud.umd.e_bdtbl;
	else
		tbptr++;
	return(0);
}

/* Disable/Enable either broadcast or promiscuous mode... */

void
disableBroadcastReception()
{
	IMMRBase->scc_regs[SCC4_REG].scc_psmr |= ENET_PSMR_BRO;
}

void
enableBroadcastReception()
{
	IMMRBase->scc_regs[SCC4_REG].scc_psmr &= ~ENET_PSMR_BRO;
}

void
enablePromiscuousReception()
{
	IMMRBase->scc_regs[SCC4_REG].scc_psmr |= ENET_PSMR_PRO;
}

void
disablePromiscuousReception()
{
	IMMRBase->scc_regs[SCC4_REG].scc_psmr &= ~ENET_PSMR_PRO;
}

void
enresetmmu()
{
}

/* extGetEtherAdd() & extGetIpAdd():
	These two functions return a char * to a string that represents an
	IP or MAC address.  If they return NULL, then it is assumed by the
	calling function that there is no target specific hardware that
	contains the IP and/or MAC address.
*/
char *
extGetEtherAdd()
{
	return((char *)0);
}
char *
extGetIpAdd()
{
	return((char *)0);
}
#endif
