/* etherdev.c:
 *	Ethernet Abstraction Layer For Micro Monitor.
 *
 *	This file contains the interface layer between Micro Monitor
 *	and the ethernet driver for the SMSC LAN911x.
 *
 *	General notice:
 *	This code is part of a boot-monitor package developed as a generic base
 *	platform for embedded system designs.  As such, it is likely to be
 *	distributed to various projects beyond the control of the original
 *	author.  Please notify the author of any enhancements made or bugs found
 *	so that all may benefit from the changes.  In addition, notification back
 *	to the author will allow the new user to pick up changes that may have
 *	been made by other users after this version of the code was distributed.
 *
 *	Author:	Torsten Tetz, Bosch Security Systems
 *	based on template/etherdev.c file by Ed Sutter
 */
#include "config.h"
#include "genlib.h"
#include "ether.h"
#include "etherdev.h"

#if INCLUDE_ETHERNET

/* Prepare global buffers for outgoing and incoming network packets. */
unsigned long tx_buf[TBUFSIZE/4];
unsigned long rx_buf[RBUFSIZE/4];

/* enreset():
 * Reset the PHY and MAC.
 *
 * Common code of MicroMonitor (../common/start.c "init2()") calls "enreset()" 
 * before the first call of "EtherdevStartup()".
 *
 * After this reset, the network will not work, because the controller is not 
 * configured and enabled here.
 */
void
enreset(void)
{
	/* Check if controller is present first. */
	if(smsc911x_verify_chip_ID()) {
	
		smsc911x_reset();
		smsc911x_PHY_reset();
	}
}

/* EtherdevStartup():
 * Start up the ethernet. First called in "../common/start.c" inside of _init3().
 *
 * It can be assumed at this point that the array uchar BinEnetAddr[6]
 * contains the 6-byte MAC address.
 *
 * Returns -1 on error and 0 when successful.
 */
int
EtherdevStartup(int verbose)
{
	/* Save the given verbosity for the ethernet driver. */
	EtherVerbose = verbose;

	/* Initialize the controller. */
	return(smsc911x_init());
}

/* disablePromiscuousReception():
 * Provide the code that disables promiscuous reception.
 */
void
disablePromiscuousReception(void)
{
	smsc911x_disablePromiscuousReception();
}

/* enablePromiscuousReception():
 * Provide the code that enables promiscuous reception.
 */
void
enablePromiscuousReception(void)
{
	smsc911x_enablePromiscuousReception();
}

/* disableBroadcastReception():
 * Provide the code that disables broadcast reception.
 */
void
disableBroadcastReception(void)
{
	/* ADD_CODE_HERE */
}

/* enableBroadcastReception():
 * Provide the code that enables broadcast reception.
 */
void
enableBroadcastReception(void)
{
	/* ADD_CODE_HERE */
}

/* enselftest():
 * Run a self test of the ethernet device(s).  This can be stubbed
 * with a return(1).
 * Return 1 if success; else -1 if failure.
 */
int
enselftest(int verbose)
{
	return(1);
}

/* ShowEtherdevStats():
 * This function is used to display device-specific stats (error counts
 * usually).
 */
void
ShowEtherdevStats(void)
{
	printf("statistics not yet implemented\n");
}

/* getXmitBuffer():
 * Return a pointer to the buffer that is to be used for transmission of
 * the next packet.  Since the monitor's driver is EXTREMELY basic,
 * there will only be one packet ever being transmitted.  No need to queue
 * up transmit packets.
 * To avoid alignment problems the buffer has to be created as an unsigned
 * long array in this uMon SH3 port. But the common prototype for all uMon 
 * ports uses an unsigned char pointer as return value of getXmitBuffer().
 */
unsigned char *
getXmitBuffer(void)
{
	return((unsigned char *)tx_buf);
}

/* sendBuffer():
 * Send out the packet assumed to be built in the buffer returned by the
 * previous call to getXmitBuffer() above.
 */
int
sendBuffer(int length)
{
	int temp;

#if INCLUDE_ETHERVERBOSE
	if(EtherVerbose & SHOW_OUTGOING) {
		printPkt((struct ether_header *)tx_buf,length,ETHER_OUTGOING);
	}
#endif

	/* Bump up the packet length to a minimum of 64 bytes. */
	if(length < 64) {
		length = 64;
	}

	/* Tell the smsc911x to send the tx buffer pointed to by tx_buf. */
	temp = smsc911x_send((unsigned long *)tx_buf,(int)length);

	/* Add 1 to the number of transmitted frames. */	
	EtherXFRAMECnt++;

	if (temp == 0) {
		return(0);
	}
	else {
		return(-1);
	}
}

/* DisableEtherdev():
 * Fine as it is...
 */
void
DisableEtherdev(void)
{
	enreset();
}

/* extGetIpAdd():
 * If there was some external mechanism (other than just using the
 * IPADD shell variable established in the monrc file) for retrieval of
 * the board's IP address, then do it here...
 */
char *
extGetIpAdd(void)
{
	return((char *)0);
}

/* extGetEtherAdd():
 * If there was some external mechanism (other than just using the
 * ETHERADD shell variable established in the monrc file) for retrieval of
 * the board's MAC address, then do it here...
 */
char *
extGetEtherAdd(void)
{
	return((char *)0);
}

/* polletherdev():
 * Called continuously by the monitor (ethernet.c) to determine if there
 * is any incoming ethernet packet.
 */
int
polletherdev(void)
{
	unsigned long	length = 0;

	length = smsc911x_package_arrived();

	if(length) {
		if(smsc911x_receive(rx_buf, length) == -1) {

			/* Receiving error has occured. */
			return(0);
		}
		else {
			/* Add 1 to the number of received frames, */
			EtherRFRAMECnt++;

			/* Process the received packet. */
			processPACKET((struct ether_header *)rx_buf, length);
			return(1);
		}
	}
	return(0);
}

#endif

