/* ad5314.c:
 * This code supports the BF537's SPI interface connection to 
 * the Analog Devices AD5314 4-channel 10-bit D-to-A converter (DAC).
 */
#include <math.h>
#include "config.h"
#if PLATFORM_TOSA
#include "cdefBF537.h"
#include "genlib.h"
#include "ctype.h"
#include "cli.h"
#include "cpu.h"
#include "cpuio.h"
#include "spi_bf537.h"
#include "ad5314.h"
#include "ad5384.h"
#include "ad7490.h"
#include "monlib.h"
#include "stddefs.h"

#define MAXVOLTS_FLOAT 	((float)2.5)
#define MAXVOLTS_HEX 	0x3ff		// AD5314 is 10-bit

extern double atof(char *);
extern double strtod(char *,char **);
extern char *sprnfloat(char *,int,float,int);

static int ad5314_baud;
static char ad5314_verbose, ad5314_initialized;
static unsigned short ad5314_vout[4];

int
ad5314_voltsTohex(float volts, unsigned short *hex)
{
	if ((volts > MAXVOLTS_FLOAT) || (volts < 0)) {
		printf("Voltage out of range\n");
		return(-1);
	}
	
	*hex = (unsigned short)((volts * MAXVOLTS_HEX)/MAXVOLTS_FLOAT);
	return(0);
}

int
ad5314_hexTovolts(unsigned short hex, float *volts)
{
	if (hex > MAXVOLTS_HEX) {
		printf("Voltage out of range\n");
		return(-1);
	}

	*volts = (float)((float)(MAXVOLTS_FLOAT * (float)hex)/(float)MAXVOLTS_HEX);
	return(0);
}

int
ad5314_write(int dev, unsigned short val)
{
	volatile unsigned short tmp;

	if (ad5314_verbose > 2)
		printf("ad5314_write(%04x)...\n",val);

	SPI_ENABLE();
	SPI_WRITE_MODE();
	*pSPI_FLG &= FLG6;		// Activate chip-select to AD5314 */

	*pSPI_TDBR = val;
	tmp = *pSPI_RDBR;		// Dummy read initiates the WRITE
	SPI_FINWAIT();			// Wait for SPI finish

	*pSPI_FLG |= ~FLG6;		// Deactivate chip-select to AD5314 */
	SPI_DISABLE();
	__builtin_bfin_ssync();

	return(0);
}

int
ad5314_read(int dev, unsigned short *val)
{
	volatile unsigned short tmp;

	if (ad5314_verbose > 2)
		printf("ad5314_read()...\n");

	SPI_ENABLE();
	SPI_READ_MODE();

	/* Activate chip-select to AD5314 */
	*pSPI_FLG &= FLG6;

	/* Read 16-bits of data...
	 */
	tmp = *pSPI_RDBR;

	SPI_FINWAIT();

	*pSPI_TDBR = 0;			/* Dummy write initiates the READ. */
	SPI_RXSWAIT();			/* Wait for receive buffer full */
	*val = *pSPI_RDBR;

	SPI_FINWAIT();

	*pSPI_FLG |= ~FLG6;

	SPI_DISABLE();
	__builtin_bfin_ssync();

	return(0);
}

int
ad5314_getvout(int dev, int chan, ushort *val)
{
	if (chan > 3) {
		printf("channel out of range\n");
		return(-1);
	}

	*val = ad5314_vout[chan];
	return(0);
}

int
ad5314_setvout(int dev, int chan, ushort val)
{
	int rc;

	if (ad5314_verbose > 1)
		printf("ad5314_setvout(dev %d, chan %d = 0x%x)...\n", dev, chan, val);

	if (chan > 3) {
		printf("channel out of range\n");
		return(-1);
	}

	rc = ad5314_write(dev, ((chan << 14) | (val << 2) | AD5314_PD));
	if (rc == 0)
		ad5314_vout[chan] = val;

	return(rc);
}

static void
ad5314_qinit(int dev)
{
	SPI_ENABLE();

	*pSPI_BAUD = (SYSCLKFREQ/(ad5314_baud*2));
	__builtin_bfin_ssync();

	/* Enable SPI interface as...
	 *  Master, MSB first, 16-bit interface
	 */
	*pSPI_CTL = (SPE | MSTR | CPHA | SIZE);
	__builtin_bfin_ssync();

	SPI_DISABLE();
}

int
ad5314_init(int dev)
{
	char *env;

	if (ad5314_verbose > 1)
		printf("ad5314_init(%d)...\n",dev);

	SPI_ENABLE();

	/* The AD5314 is on SPI CS6 line, so enable that line.
	 */
	*pSPI_FLG |= FLS6;
	__builtin_bfin_ssync();

	if ((env = getenv("DACBAUD")) != (char *)0)
		ad5314_baud = strtol(env,0,0);
	else
		ad5314_baud = DACBAUD;

	ad5314_qinit(dev);

	*pSPI_STAT = (MODF | TXE | RBSY | TXCOL);
	__builtin_bfin_ssync();

	/* Write to all four channels, then on the fourth (final)
	 * write, allow LDAC to be low so that ouputs are updated.
	 */
	ad5314_write(0,(0 << 14) | AD5314_PD | AD5314_LDAC);
	ad5314_write(0,(1 << 14) | AD5314_PD | AD5314_LDAC);
	ad5314_write(0,(2 << 14) | AD5314_PD | AD5314_LDAC);
	ad5314_write(0,(3 << 14) | AD5314_PD);

	/* This device is write-only, so we have to keep track of the
	 * last setting...
	 */
	ad5314_vout[0] = 0;
	ad5314_vout[1] = 0;
	ad5314_vout[2] = 0;
	ad5314_vout[3] = 0;

	SPI_DISABLE();

	ad5314_initialized = 1;
	return(0);
}

char *ad5314Help[] = {
	"Access the AD5314 digital-to-analog converter.",
	"[-d:v] {cmd} [cmd-specific args]",
	" Options:",
	" -d#        specify device number (default=0)",
	" -v         verbosity level (default=0)",
	"",
	" Commands:",
	" vout {chan} [volts]    read/write data reg (per channel)",
	0,
};

int
ad5314Cmd(int argc,char *argv[])
{
	int devno, opt, rc = 0;
	char *cmd;

	devno = 0;
	ad5314_verbose = 0;
	while((opt=getopt(argc,argv,"d:v")) != -1) {
		switch(opt) {
		case 'd':
			devno = atoi(optarg);
			break;
		case 'v':
			ad5314_verbose++;
			break;
		default:
			return(CMD_PARAM_ERROR);
		}
	}
	
	if (devno >= AD5314_DEVTOT) {
		printf("Devno %d too big\n",devno);
		return(CMD_FAILURE);
	}

	if (argc <= optind)
		return(CMD_PARAM_ERROR);

	cmd = argv[optind];

	if ((argc == optind+1) && (strcmp(cmd,"init") == 0)) {
		ad5314_init(devno);
		return(CMD_SUCCESS);
	}

	if (ad5314_initialized == 0) {
		printf("Run 'ad5314 init' first\n");
		return(CMD_FAILURE);
	}

	ad5314_qinit(devno);

	if (strcmp(cmd,"vout") == 0) {
		int chan;
		char *range, *env, varname[16];
		ushort vout;

		range = argv[optind+1];
		if (argc == optind+2) {
			char buf[32];
			float fpval;
			int tot = 0;

			// Retrieve vout setting for requested channel range:
			for(chan=0;chan<4;chan++) {
				if (inRange(range,chan)) {
					if (tot == 0)
						printf("VOUT [%s]: ",range);

					// If variable is set, then return fake value...
					sprintf(varname,"VOUT%d",chan);
					if ((env = getenv(varname)) != (char *)0) {
						printf("%d=%s,",chan,env);
					}
					else {
						rc = ad5314_getvout(devno,chan, &vout);
						ad5314_hexTovolts(vout & 0x3fff,&fpval);
						printf("%d=%s,",
							chan,sprnfloat(buf,sizeof(buf),fpval,3));
					}
					tot++;
				}
			}
			if (tot > 0)
				printf("\n");
		}
		else if (argc == optind+3) {
			if (ad5314_voltsTohex((float)strtod(argv[optind+2],0), &vout) == -1)
				return(CMD_PARAM_ERROR);
			for(chan=0;chan<40;chan++) {
				if (inRange(range,chan))
					rc = ad5314_setvout(devno,chan, vout);
			}
		}
		else
			return(CMD_PARAM_ERROR);
	}
	else
		return(CMD_PARAM_ERROR);

	if (rc != 0)
		printf("RC: %d\n",rc);

	return(CMD_SUCCESS);
}

int
getTEC(void)
{
	return(*pPORTGIO & TOSA_TEC_EN_BIT ? 1:0);
}

char *tecHelp[] = {
	"Turn on/off the TEC power supply.",
	"[on | off]",
	0,
};

int
tecCmd(int argc,char *argv[])
{
	if (argc == 1) {
		printf("TEC power supply: %s\n",getTEC() == 1 ? "off" : "on");
		return(CMD_SUCCESS);
	}

	if ((argc == 2) && (!strcmp(argv[1],"on"))) {
		TOSA_TEC_CLR();
	}
	else if ((argc == 2) && (!strcmp(argv[1],"off"))) {
		TOSA_TEC_SET();
	}
	else
		return(CMD_PARAM_ERROR);

	return(CMD_SUCCESS);
}

#endif
