/* ad5384.c:
 * This code supports the BF537's SPI interface connection to 
 * the Analog Devices AD5384 40-channel D-to-A converter (DAC).
 * 
 * The full swing voltage range is 0-2.5vdc, which is controlled
 * with 14 bits.
 */
#include <math.h>
#include "config.h"
#include "cdefBF537.h"
#include "genlib.h"
#include "ctype.h"
#include "cli.h"
#include "cpu.h"
#include "spi_bf537.h"
#include "ad5384.h"
#include "ad7490.h"
#include "monlib.h"
#include "stddefs.h"

extern double atof(char *);
extern double strtod(char *,char **);
extern char *sprnfloat(char *,int,float,int);

static int ad5384_incrTbl[40];
static int ad5384_verbose;
static int ad5384_initialized;
static long ad5384_baud;

#define MAXVOLTS_FLOAT	((float)2.5)
#define MAXVOLTS_HEX 	0x3fff		// AD5384 is 14-bit

/* There is some confusion with regard to the use of the term 
 * "channel".  Throughout this AD5384 code, "channel" simply refers to
 * the VOUTX line of the device.  Unfortunately, this does not directly
 * map to the "channels" as they relate to the functionality of the
 * hardware.  So, below is a mapping of VOUTX to label on the ROSA
 * schematic:
 *
 * VOUT0:  RX_LOS_THRESHOLD_CHAN0
 * VOUT1:  VAPD_TARGET_CHAN0
 * VOUT2:  RX_LOS_THRESHOLD_CHAN2
 * VOUT3:  VAPD_TARGET_CHAN2
 * VOUT4:  RX_LOS_THRESHOLD_CHAN4
 * VOUT5:  VAPD_TARGET_CHAN4
 * VOUT6:  RX_LOS_THRESHOLD_CHAN6
 * VOUT7:  VAPD_TARGET_CHAN6
 * VOUT8:  RX_LOS_THRESHOLD_CHAN8
 * VOUT9:  VAPD_TARGET_CHAN8
 * VOUT10: RX_LOS_THRESHOLD_CHAN1
 * VOUT11: VAPD_TARGET_CHAN1
 * VOUT12: RX_LOS_THRESHOLD_CHAN3
 * VOUT13: VAPD_TARGET_CHAN3
 * VOUT14: RX_LOS_THRESHOLD_CHAN5
 * VOUT15: VAPD_TARGET_CHAN5
 * VOUT16: RX_LOS_THRESHOLD_CHAN7
 * VOUT17: VAPD_TARGET_CHAN7
 * VOUT18: RX_LOS_THRESHOLD_CHAN9
 * VOUT19: VAPD_TARGET_CHAN9
 */

/* This function provides a translation from the ad5384's
 * VAPD_TARGET_CHANXX (5384) to APD_MON_CHANXX (7490).
 * Input is the VOUT number on the AD5384, and the return
 * (if successful) is the VIN number on the 7490.
 */
int
ad5384vout_to_ad7490vin(int vout)
{
	switch(vout) {
#if PLATFORM_ROSA
	case 1:		// VOUT01 of AD5384 corresponds to VIN9 of AD7490
		return(9);
	case 3:		// VOUT03 of AD5384 corresponds to VIN7 of AD7490
		return(7);
	case 5:		// VOUT05 of AD5384 corresponds to VIN5 of AD7490
		return(5);
	case 7:		// VOUT07 of AD5384 corresponds to VIN3 of AD7490
		return(3);
	case 9:		// VOUT09 of AD5384 corresponds to VIN1 of AD7490
		return(1);
	case 11:	// VOUT11 of AD5384 corresponds to VIN8 of AD7490
		return(8);
	case 13:	// VOUT13 of AD5384 corresponds to VIN6 of AD7490
		return(6);
	case 15:	// VOUT15 of AD5384 corresponds to VIN4 of AD7490
		return(4);
	case 17:	// VOUT17 of AD5384 corresponds to VIN2 of AD7490
		return(2);
	case 19:	// VOUT19 of AD5384 corresponds to VIN0 of AD7490
		return(0);
#else
#endif
	default:
		printf("AD5384vout-to-AD7490vin translation failure (%d)\n",vout);
		return(-1);
	}
}

int
ad5384_voltsTohex(float volts, unsigned short *hex)
{
	if ((volts > MAXVOLTS_FLOAT) || (volts < 0)) {
		printf("Voltage out of range\n");
		return(-1);
	}
	
	*hex = (unsigned short)((volts * MAXVOLTS_HEX)/MAXVOLTS_FLOAT);
	return(0);
}

int
ad5834_hexTovolts(unsigned short hex, float *volts)
{
	if (hex > MAXVOLTS_HEX) {
		printf("Voltage out of range\n");
		return(-1);
	}

	*volts = (float)((float)(MAXVOLTS_FLOAT * (float)hex)/(float)MAXVOLTS_HEX);
	return(0);
}

static int
ad5384_write(int dev, unsigned char b1, unsigned char b2, unsigned char b3)
{
	volatile ushort tmp;

	if (ad5384_verbose > 2)
		printf("ad5384_write(%02x,%02x,%02x)...\n",b1,b2,b3);

	// Note regarding the 'dev' paramter...
	// If there was more than one ad5384 in the system, then this is
	// where the code would be that would distinguish between the
	// different SPI addresses.  For now (Nov 2009), we only have
	// one, so the 'dev' parameter is ignored.

	SPI_ENABLE();
	SPI_WRITE_MODE();
	*pSPI_FLG &= FLG2;		// Activate chip-select to AD5384 */

	/* Write 24-bits of data...
	 */
	*pSPI_TDBR = (ushort)b1;
	tmp = *pSPI_RDBR;		// Dummy read initiates the WRITE
	SPI_FINWAIT();			// Wait for SPI finish
	*pSPI_TDBR = (ushort)b2;
	tmp = *pSPI_RDBR;
	SPI_FINWAIT();
	*pSPI_TDBR = (ushort)b3;
	tmp = *pSPI_RDBR;
	SPI_FINWAIT();

	*pSPI_FLG |= ~FLG2;		// Deactivate chip-select to AD5384 */
	SPI_DISABLE();
	__builtin_bfin_ssync();

	return(0);
}

static int
ad5384_read(int dev, unsigned char *buf)
{
	int i;
	volatile ushort tmp;

	if (ad5384_verbose > 2)
		printf("ad5384_read()...\n");

	// See note in ad5384_write() regarding the use (or the lack thereof)
	// of the 'dev' parameter.
	SPI_ENABLE();
	SPI_READ_MODE();

	/* Activate chip-select to AD5384 */
	*pSPI_FLG &= FLG2;

	/* Read 24-bits of data...
	 */
	tmp = *pSPI_RDBR;

	SPI_FINWAIT();

	for(i=0;i<3;i++) {			/* Retrieve 3 bytes from device. */
		*pSPI_TDBR = 0;			/* Dummy write initiates the READ. */
		SPI_RXSWAIT();			/* Wait for receive buffer full */
		buf[i] = (unsigned char)*pSPI_RDBR;
	}

	SPI_FINWAIT();

	*pSPI_FLG |= ~FLG2;

	SPI_DISABLE();
	__builtin_bfin_ssync();

	return(0);
}

/* Special Function Register access commands:
 */
int
ad5384_nop(int dev)
{
	if (ad5384_verbose > 1)
		printf("ad5384_nop(%d)...\n",dev);
	return(ad5384_write(dev,(AREG | AD5384_WRITE),SFR | 0,0));
}

int
ad5384_writeclrcode(int dev,ushort code)
{
	if (ad5384_verbose > 1)
		printf("ad5384_writeclrcode(%d)...\n",dev);
	return(ad5384_write(dev,(AREG | AD5384_WRITE | A0),
		SFR | ((code & 0x3f00)>>8),code & 0xff));
}

int
ad5384_softclear(int dev)
{
	if (ad5384_verbose > 1)
		printf("ad5384_softclear(%d)...\n",dev);
	return(ad5384_write(dev,(AREG | AD5384_WRITE | A1),SFR | 0,0));
}

int
ad5384_softpowerdown(int dev)
{
	if (ad5384_verbose > 1)
		printf("ad5384_softpowerdown(%d)...\n",dev);
	return(ad5384_write(dev,(AREG | AD5384_WRITE | A3),SFR | 0,0));
}

int
ad5384_softpowerup(int dev)
{
	if (ad5384_verbose > 1)
		printf("ad5384_softpowerup(%d)...\n",dev);
	return(ad5384_write(dev,(AREG | AD5384_WRITE | A3 | A0),SFR | 0,0));
}

int
ad5384_ctrlwrite(int dev, ushort reg)
{
	if (ad5384_verbose > 1)
		printf("ad5384_ctrlwrite(%d, 0x%04x)...\n",dev, reg);
	return(ad5384_write(dev,(AREG | AD5384_WRITE | A3 | A2),
		SFR | ((reg & 0x3f00)>>8),reg & 0xff));
}

int
ad5384_ctrlread(int dev, unsigned char *buf)
{
	int rc;

	rc = ad5384_write(dev, (AREG | READ | A3 | A2),SFR | 0,0);
	if (rc < 0)
		return(rc);
	return(ad5384_read(dev, buf));
}

int
ad5384_monitorchannel(int dev)
{
	if (ad5384_verbose > 1)
		printf("ad5384_monitorchannel(%d)...\n",dev);
	return(ad5384_write(dev,(AREG | AD5384_WRITE | A3 | A1),SFR | 0,0));
}

int
ad5384_softreset(int dev)
{
	if (ad5384_verbose > 1)
		printf("ad5384_softreset(%d)...\n",dev);
	return(ad5384_write(dev,
		(AREG | AD5384_WRITE | A3 | A2 | A1 | A0),SFR | 0,0));
}

int
ad5384_setvout(int dev, int chan, ushort val)
{
	if (ad5384_verbose > 1)
		printf("ad5384_setvout(dev %d, chan %d = 0x%x)...\n", dev, chan, val);
	if (chan > 39) {
		printf("channel out of range\n");
		return(-1);
	}
	return(ad5384_write(dev, (AREG | AD5384_WRITE | chan),
		INPUTDATA | ((val & 0x3f00)>>8),val & 0xff));
}

int
ad5384_getvout(int dev, int chan, ushort *val)
{
	int rc;
	unsigned char buf[3];

	rc = ad5384_write(dev, (AREG | READ | chan),INPUTDATA | 0,0);
	if (rc < 0)
		return(rc);
	rc = ad5384_read(dev, buf);
	*val = ((buf[1] << 8) | buf[0]);
	return(rc);
}

int
ad5384_setgain(int dev, int chan, ushort val)
{
	if (ad5384_verbose > 1)
		printf("ad5384_setgain(dev %d, chan %d = 0x%x)...\n", dev, chan, val);
	if (chan > 39) {
		printf("channel out of range\n");
		return(-1);
	}
	return(ad5384_write(dev, (AREG | AD5384_WRITE | chan),
		GAIN | ((val & 0x3f00)>>8),val & 0xff));
}

int
ad5384_getgain(int dev, int chan, ushort *val)
{
	int rc;
	unsigned char buf[3];

	rc = ad5384_write(dev, (AREG | READ | chan),GAIN | 0,0);
	if (rc < 0)
		return(rc);
	rc = ad5384_read(dev, buf);
	*val = ((buf[1] << 8) | buf[0]);
	return(rc);
}

int
ad5384_setoffset(int dev, int chan, ushort val)
{
	if (ad5384_verbose > 1)
		printf("ad5384_setoffset(dev %d, chan %d = 0x%x)...\n", dev, chan, val);
	if (chan > 39) {
		printf("channel out of range\n");
		return(-1);
	}
	return(ad5384_write(dev, (AREG | AD5384_WRITE | chan),
		OFFSET | ((val & 0x3f00)>>8),val & 0xff));
}

int
ad5384_getoffset(int dev, int chan, ushort *val)
{
	int rc;
	unsigned char buf[3];

	rc = ad5384_write(dev, (AREG | READ | chan),OFFSET | 0,0);
	if (rc < 0)
		return(rc);
	rc = ad5384_read(dev, buf);
	*val = ((buf[1] << 8) | buf[0]);
	return(rc);
}

static void
ad5384_qinit(int dev)
{
	SPI_ENABLE();

	*pSPI_BAUD = (SYSCLKFREQ/(ad5384_baud*2));
	__builtin_bfin_ssync();

	/* Enable SPI interface as...
	 *  Master, MSB first, 
	 */
	*pSPI_CTL = (SPE | MSTR | CPHA);
	__builtin_bfin_ssync();

	SPI_DISABLE();
}

int
ad5384_decrement_vout(int dev, int chan)
{
	unsigned short vout;

	printf("Decrement channel %d by 0x%x...",
		chan,ad5384_incrTbl[chan]);

	if (ad5384_getvout(dev,chan, &vout) < 0) {
		printf("failed\n");
		return(-1);
	}

	vout -= ad5384_incrTbl[chan];

	if (ad5384_setvout(dev,chan, vout) < 0) {
		printf("failed\n");
		return(-1);
	}

	printf("succeeded\n");
	return(0);
}

/* ad5384_init():
 * The AD5384 uses SPI_SEL2 on the BF537.
 * It is assumed here that the pin configuration has already been set
 * up during basic board init.
 * The hardware is configured to use SPI and standalone mode.
 */
int
ad5384_init(int dev)
{
	int i;
	char *env;

	if (ad5384_verbose > 1)
		printf("ad5384_init(%d)...\n",dev);

	SPI_ENABLE();

	/* The AD5384 is on SPI CS2 line, so enable that line.
	 */
	*pSPI_FLG |= FLS2;
	__builtin_bfin_ssync();

	if ((env = getenv("DACBAUD")) != (char *)0)
		ad5384_baud = strtol(env,0,0);
	else
		ad5384_baud = DACBAUD;

	ad5384_qinit(dev);

	*pSPI_STAT = (MODF | TXE | RBSY | TXCOL);
	__builtin_bfin_ssync();

	/* Init to internal reference:
	 */
	ad5384_ctrlwrite(dev, 0x0400);

	SPI_DISABLE();

	/* Initialize the default increments for each channel to be
	 * 1/100th of the full swing voltage...
 	 * This is changed by the use of the ramp subcommand.
	 */
	for(i=0;i<40;i++)
		ad5384_incrTbl[i] = MAXVOLTS_HEX/100;

	ad5384_initialized = 1;
	return(0);
}

int
getAD5384Chan(char *string, int *chan)
{
	int tmp;

	tmp = atoi(string);
	if ((tmp > 39) || (tmp < 0)) {
		printf("Valid channel range: 0-39\n");
		return(-1);
	}
	*chan = tmp;
	return(0);
}

int
getBits14(char *string, unsigned short *val)
{
	unsigned short tmp;

	tmp = strtol(string,0,0);
	if (tmp > 0x3fff) {
		printf("value out of 14-bit range\n");
		return(-1);
	}
	*val = tmp;
	return(0);
}


char *ad5384Help[] = {
	"Access the AD5384 digital-to-analog converter.",
	"[-d:v] {cmd} [cmd-specific args]",
	" Options:",
	" -d#        specify device number (default=0)",
	" -v         verbosity level (default=0)",
	"",
	" Commands:",
	" init         initialze",
	" pdn          powerdown",
	" pup          powerup",
	" sreset       soft reset",
	" nop          no-operation",
	" ctrl [val]   read/write control register",
	" clr {volts}  set 'clear' voltage value",
	" clr          soft clear",
	" ramp {chan} {start_volts} {stop_volts} {step_volts} {msec}",
	"              ramp up/down through voltage range",
	" vout {chan} [volts]    read/write data reg (per channel)",
	" offset {chan} [hexval] read/write offset reg (per channel)",
	" gain {chan} [hexval]   read/write gain reg (per channel)",
	0,
};

int
ad5384Cmd(int argc,char *argv[])
{
	int devno, opt, rc = 0;
	char *cmd;

	devno = 0;
	ad5384_verbose = 0;
	while((opt=getopt(argc,argv,"d:v")) != -1) {
		switch(opt) {
		case 'd':
			devno = atoi(optarg);
			break;
		case 'v':
			ad5384_verbose++;
			break;
		default:
			return(CMD_PARAM_ERROR);
		}
	}
	
	if (devno >= AD5384_DEVTOT) {
		printf("Devno %d too big\n",devno);
		return(CMD_FAILURE);
	}

	if (argc <= optind)
		return(CMD_PARAM_ERROR);

	cmd = argv[optind];

	if ((argc == optind+1) && (strcmp(cmd,"init") == 0)) {
		ad5384_init(devno);
		return(CMD_SUCCESS);
	}

	if (ad5384_initialized == 0) {
		printf("Run 'ad5384 init' first\n");
		return(CMD_FAILURE);
	}

	ad5384_qinit(devno);

	if ((argc == optind+1) && (strcmp(cmd,"pdn") == 0))
		rc = ad5384_softpowerdown(devno);
	else if ((argc == optind+1) && (strcmp(cmd,"pup") == 0))
		rc = ad5384_softpowerup(devno);
	else if ((argc == optind+1) && (strcmp(cmd,"sreset") == 0))
		rc = ad5384_softreset(devno);
	else if ((argc == optind+1) && (strcmp(cmd,"nop") == 0))
		rc = ad5384_nop(devno);
	else if (strcmp(cmd,"ctrl") == 0) {
		unsigned char buf[3];
		if (argc == 3) {
			rc = ad5384_ctrlwrite(devno,(ushort)strtol(argv[optind+1],0,0));
		}
		else {
			if ((rc = ad5384_ctrlread(devno,buf)) == 0)
				printf("CTRL: %02x %02x %02x\n",buf[0], buf[1],buf[2]);
		}
	}
	else if (strcmp(cmd,"clr") == 0) {
		if (argc == optind+2) {
			ushort val;

			if (ad5384_voltsTohex((float)strtod(argv[optind+1],0), &val) == -1)
				return(CMD_PARAM_ERROR);
			rc = ad5384_writeclrcode(devno,val);
		}
		else if (argc == optind+1)
			rc = ad5384_softclear(devno);
		else
			return(CMD_PARAM_ERROR);
	}
	else if (strcmp(cmd,"vout") == 0) {
		int chan;
		char *range, *env, varname[16];
		ushort vout;

		range = argv[optind+1];
		if (argc == optind+2) {
			char buf[32];
			float fpval;
			int tot = 0;

			// Retrieve vout setting for requested channel range:
			for(chan=0;chan<40;chan++) {
				if (inRange(range,chan)) {
					if (tot == 0)
						printf("VOUT [%s]: ",range);

					// If variable is set, then return fake value...
					sprintf(varname,"VOUT%d",chan);
					if ((env = getenv(varname)) != (char *)0) {
						printf("%d=%s,",chan,env);
					}
					else {
						rc = ad5384_getvout(devno,chan, &vout);
						ad5834_hexTovolts(vout & 0x3fff,&fpval);
						printf("%d=%s,",
							chan,sprnfloat(buf,sizeof(buf),fpval,3));
					}
					tot++;
				}
			}
			if (tot > 0)
				printf("\n");
		}
		else if (argc == optind+3) {
			if (ad5384_voltsTohex((float)strtod(argv[optind+2],0), &vout) == -1)
				return(CMD_PARAM_ERROR);
			for(chan=0;chan<40;chan++) {
				if (inRange(range,chan))
					rc = ad5384_setvout(devno,chan, vout);
			}
		}
		else
			return(CMD_PARAM_ERROR);
	}
	else if (strcmp(cmd,"gain") == 0) {
		int chan;
		ushort gain;
		if (argc == optind+2) {
			// Retrieve the current gain setting for the requested channel
			if (getAD5384Chan(argv[optind+1],&chan) == -1)
				return(CMD_FAILURE);
		
			rc = ad5384_getgain(devno,chan, &gain);
			printf("VOUT %d: 0x%04x\n",chan,gain);
		}
		else if (argc == optind+3) {
			// Retrieve gain for requested channel
			if (getAD5384Chan(argv[optind+1],&chan) == -1)
				return(CMD_FAILURE);
			gain = (short)strtol(argv[optind+2],0,0); 
			rc = ad5384_setgain(devno,chan, gain);
		}
		else
			return(CMD_PARAM_ERROR);
	}
	else if (strcmp(cmd,"offset") == 0) {
		int chan;
		ushort offset;
		if (argc == optind+2) {
			// Retrieve the current offset setting for the requested channel
			if (getAD5384Chan(argv[optind+1],&chan) == -1)
				return(CMD_FAILURE);
			rc = ad5384_getoffset(devno,chan, &offset);
			printf("VOUT %d: 0x%04x\n",chan,offset);
		}
		else if (argc == optind+3) {
			// Retrieve offset for requested channel
			if (getAD5384Chan(argv[optind+1],&chan) == -1)
				return(CMD_FAILURE);
			offset = (short)strtol(argv[optind+2],0,0); 
			rc = ad5384_setoffset(devno,chan, offset);
		}
		else
			return(CMD_PARAM_ERROR);
	}
	else if (strcmp(cmd,"ramp") == 0) {
		int chan, vinchan, delay;
		ushort vout, start, stop, step;

		if (argc != optind+6)
			return(CMD_PARAM_ERROR);
		if (getAD5384Chan(argv[optind+1],&chan) == -1)
			return(CMD_FAILURE);
#if 0
		if (getBits14(argv[optind+2],&start) == -1)
			return(CMD_FAILURE);
		if (getBits14(argv[optind+3],&stop) == -1)
			return(CMD_FAILURE);
		if (getBits14(argv[optind+4],&step) == -1)
			return(CMD_FAILURE);
#else
		if (ad5384_voltsTohex((float)strtod(argv[optind+2],0), &start) == -1)
			return(CMD_PARAM_ERROR);
		if (ad5384_voltsTohex((float)strtod(argv[optind+3],0), &stop) == -1)
			return(CMD_PARAM_ERROR);
		if (ad5384_voltsTohex((float)strtod(argv[optind+4],0), &step) == -1)
			return(CMD_PARAM_ERROR);
#endif
		delay = strtol(argv[optind+5],0,0);
		vout = start;

		ad5384_incrTbl[chan] = step;
		vinchan = ad5384vout_to_ad7490vin(chan);

		while(1) {
			char fpbuf1[32], fpbuf2[32];
			float fpval1 = 0;
			float fpval2 = 0;

			ad5834_hexTovolts(vout,&fpval1);
			sprnfloat(fpbuf1,sizeof(fpbuf1),fpval1,3);

			if (ad5384_setvout(devno,chan, vout) < 0) {
				printf("setvout failed\n");
				break;
			}
			if (vout == stop)
				break;

			if (gotachar()) {
				printf("-abort: vout=%sv (0x%x)\n",fpbuf1,vout);
				break;
			}
			if (start > stop) {		// Are we decrementing or incrementing?
				if ((ushort)(vout - step) > vout) {
					vout = stop;
				}
				else {
					vout -= step;
				}
			}
			else {
				vout += step;
				if (vout > stop)
					vout = stop;
			}
			monDelay(delay);

			ad7490_qinit(devno);
			if (ad7490_scan(devno) > 0)
				break;
			ad5384_qinit(devno);

			if (ad5384_verbose == 1)  {
				ad7490_hexTovolts(vinTbl[devno][vinchan],&fpval2);
				sprnfloat(fpbuf2,sizeof(fpbuf2),fpval2,3);
				printf("vout[%d][%s]: vin[%d][%s]\n",
					chan,fpbuf1,vinchan,fpbuf2);
			}
		}
		if (ad5384_verbose) 
			putchar('\n');
	}
	else
		return(CMD_PARAM_ERROR);

	if (rc != 0)
		printf("RC: %d\n",rc);

	return(CMD_SUCCESS);
}

