#include <stdint.h>
#include "config.h"
#include "stddefs.h"
#include "genlib.h"
#include "cache.h"
#include "warmstart.h"
#include "cpu.h"
#include "cpuio.h"
#include "timer.h"
#include "cdefBF537.h"
#include "date.h"
#include "i2c.h"
#include "cli.h"
#include "spif.h"
#include "spifmap.h"
#include "ad5314.h"
#include "ad5384.h"
#include "ad7490.h"

#ifdef DMA_UART_RCV
#define UBUFSIZ CACHE_LNSZ
uint8_t ubuf[CACHE_LNSZ*3];
uint8_t *ubuf0, *ubuf1;
uint8_t *uptr0, *uptr1;
uint8_t *uend0, *uend1;
#endif


/* devInit():
 * Initialize console...
 */
void static
uartInit(ulong base, int baud)
{
	ushort tmp;

	*(vushort *)(base + GCTL) = 1;		/* Enable UART clocks */
	*(vushort *)(base + MCR) = 0;
	*(vushort *)(base + LCR) = 0x03;	/* no parity, 8-bit char, 1 stopbit */
	tmp = *(vushort *)(base + LSR);		/* clear LSR */
	*(vushort *)(base + IER) = 0;		/* no ints */
	*(vushort *)(base + SCR) = 0;		/* no ints */
	ConsoleBaudSet(baud);

#ifdef DMA_UART_RCV
	if (base == UART0_BASE) {
		/* Init DMA8 (UART0 RCV):
		 */
		*pUART0_IER = 0;
		*pDMA8_CONFIG = 0;
		*pDMA8_START_ADDR = ubuf0;
		*pDMA8_CURR_ADDR = ubuf0;
		*pDMA8_X_COUNT = UBUFSIZ;
		*pDMA8_X_MODIFY = 1;
		*pDMA8_CURR_X_COUNT = 0;
		*pDMA8_CONFIG = DMAEN | WNR | WDSIZE_8 | FLOW_AUTO;
		*pUART0_IER = ERBFI;
		uptr0 = ubuf0;
		uend0 = uptr0+UBUFSIZ;
	}
	if (base == UART1_BASE) {
		/* Init DMA10 (UART1 RCV):
		 */
		*pUART1_IER = 0;
		*pDMA10_CONFIG = 0;
		*pDMA10_START_ADDR = ubuf1;
		*pDMA10_CURR_ADDR = ubuf1;
		*pDMA10_X_COUNT = UBUFSIZ;
		*pDMA10_X_MODIFY = 1;
		*pDMA10_CURR_X_COUNT = 0;
		*pDMA10_CONFIG = DMAEN | WNR | WDSIZE_8 | FLOW_AUTO;
		*pUART1_IER = ERBFI;
		uptr1 = ubuf1;
		uend1 = uptr1+UBUFSIZ;
	}
#endif
}

int static
uartBaudSet(ulong base, int baud)
{
	unsigned short	lctl, hi, lo;

	lo = SYSCLKFREQ/(baud*16);
	hi = (lo & 0xff00) >> 8;
	lo &= 0x00ff;

	lctl = *(vushort *)(base + LCR);	/* Save linectl reg */
	*(vushort *)(base + LCR) = DLAB;
	*(vushort *)(base + DLL) = lo;		/* Set baud divisor */
	*(vushort *)(base + DLH) = hi;
	*(vushort *)(base + LCR) = lctl;	/* Restore linectl reg */
	return(0);
}

void
devInit(int baud)
{
#ifdef DMA_UART_RCV
	/* Align the start of each UART's DMA buffer to be on a
	 * cache-line boundary...
	 */
	ubuf0 = ubuf;
	while(((long)ubuf0 & (CACHE_LNSZ-1)) != 0)
		ubuf0++;
	ubuf1 = ubuf0+CACHE_LNSZ;
	//printf("0x%lx 0x%lx 0x%lx\n",(long)ubuf,(long)ubuf0,(long)ubuf1);
#endif
	uartInit(UART0_BASE,baud);
//	uartInit(UART1_BASE,baud);
}

/* ConsoleBaudSet():
 * Change baud rate to that of 'baud'.
 */
int
ConsoleBaudSet(int baud)
{
	uartBaudSet(UART0_BASE,baud);
//	uartBaudSet(UART1_BASE,baud);
	return(0);
}

/* target_console_empty():
 * Returns 1 if there are no characters waiting to
 * be put out on the UART; else return 0 indicating that the UART
 * is still busy outputting characters from its FIFO.
 */
int
target_console_empty(void)
{
	if (*(ushort *)CONSOLE(LSR) & TEMT)
		return(1);
	return(0);
}

/* target_putchar():
 * When buffer has space available, load the incoming character
 * into the UART.
 */
int
target_putchar(char c)
{
    /* Wait for transmit ready bit */
	while(!(*(vushort *)CONSOLE(LSR) & THRE));

	*(vushort *)CONSOLE(THR) = (ushort)c;

    return((int)c);
}

/* target_gotachar():
 * Return 0 if no char is avaialable at UART rcv fifo; else 1.
 * Do NOT pull character out of fifo, just return status. 
 */
int
target_gotachar(void)
{
#ifdef DMA_UART_RCV
	if (*pDMA8_CURR_ADDR != uptr0)
		return(1);
#else
	if (*(vushort *)CONSOLE(LSR) & DR)
		return(1);
#endif
	if (getenv("SCAN_ENABLE"))
		overcurrent_scan(0);
	return(0);
} 

/* target_getchar():
 * Assume there is a character in the UART's input buffer and just
 * pull it out and return it.
 */
int 
target_getchar(void)
{
#ifdef DMA_UART_RCV
	if (uptr0 == uend0)
		uptr0 = (uint8_t *)ubuf0;

	if (getenv("FLUSHINV"))
		FlushInv((char *)ubuf0);

	return(*uptr0++);
#else
    return((int)*(vushort *)CONSOLE(RBR));
#endif
}

int
iInvalidateRange(char *start, int size)
{
	long addr, end, imemctrl;


	/* If start and size are zero, then we are to invalidate the entire
	 * i-Cache.  That's done by clearing the IMC bit then re-setting it.
	 */
	if ((start == 0) && (size == 0)) {
		imemctrl = *pIMEM_CONTROL;

		asm("ssync;");
		*pIMEM_CONTROL &= ~IMC;
		asm("csync;");
		*pIMEM_CONTROL = imemctrl;
		asm("csync;");
	}
	else {
		addr = (long)start;
		addr &= ~0x1f;
		end = (long)(start + size);

		for(;addr < end; addr += 32)
			IFlush((char *)addr);
	}
	return(0);
}

int
dFlushRange(char *start, int size)
{
	long addr, end;

	/* TODO:
	 * This needs to support the "start=0,size=0" input to flush
	 * the entire d-cache.
 	 */
	addr = (long)start;
	addr &= ~0x1f;
	end = (long)(start + size);

	for(;addr < end; addr += 32)
		Flush((char *)addr);

	return(0);
}

/* cacheInitForTarget():
 * Establish target specific function pointers and
 * enable i-cache...
 * Refer to $core/cache.c for a description of the function pointers.
 * NOTE:
 * If cache (either I or D or both) is enabled, then it is important
 * that the appropriate cacheflush/invalidate function be established.
 * This is very important because programs (i.e. cpu instructions) are
 * transferred to memory using data memory accesses and could
 * potentially result in cache coherency problems.
 */
void
cacheInitForTarget(void)
{
	extern int icache_off(void);
	extern int dcache_off(void);

	icacheDisable = icache_off;
	dcacheDisable = dcache_off;
	dcacheFlush = dFlushRange;
	icacheInvalidate = iInvalidateRange;
}

/* target_reset():
 * From BF537 programmer's manual...
 * The reset vector is determined by the processor system. It points to the
 * start of the on-chip boot ROM (0xef000000), or to the start of
 * external asynchronous memory (0x20000000), depending on the state of
 * the BMODE pins.  uMon is built to run in Mode 0, which means the CPU
 * starts fetching from location 0x20000000 immediately out of reset.
 */
void
target_reset(void)
{
	flushDcache(0,0);
	disableDcache();
	invalidateIcache(0,0);
	disableIcache();

	/* Use Watchdog to generate reset...
	 */
	*pWDOG_CTL = 0x0ad0;	/* Disable watchdog */
	*pWDOG_CNT = 0x0004;	/* Set a low count value */
	*pWDOG_CTL = 0x0000;	/* Enable watchdog */
	while(1);				/* Wait for the hit */
}

/* If any CPU IO wasn't initialized in reset.S, do it here...
 * This just provides a "C-level" IO init opportunity. 
 */
void
initCPUio(void)
{
	int i;

#if PLATFORM_TOSA
	for(i=0;i<AD5314_DEVTOT;i++)
		ad5314_init(i);
#endif

	for(i=0;i<AD5384_DEVTOT;i++)
		ad5384_init(i);

	for(i=0;i<AD7490_DEVTOT;i++)
		ad7490_init(i);
}

char *spiinfo[] = {
	"Addr      Description",
	"  1       AT25DF321 Serial flash",
	0,
};

int
bfin_spishow(int interface)
{
	int i;

	for(i=0; spiinfo[i]; i++)
		printf("%s\n",spiinfo[i]);
	return(0);
}

void
showBlackfinInfo(int center)
{
	uchar *sno;
	int	(*pfunc)(char *, ...);
	uchar *getSerialNumber(void);

	if (center)
		pfunc = cprintf;
	else
		pfunc = printf;
	pfunc("CPU Version %d\n",*pDSPID & 0x0000ffff);
	sno = getSerialNumber();
	if ((sno) && (sno[0] != 0xff))
		pfunc("SNO: %s\n",sno);
}

uchar *
getSerialNumber(void)
{
	static unsigned char snostr[64];

	if (spifReadBlock(SPIF_SNO_ADDRS_BASE,(char *)snostr,sizeof(snostr)) < 0) {
		printf("getSerialNumber: spifReadBlock error\n");
		return(0);
	}

	if (snostr[0] != 0xff)
		setenv("SERIALNO",(char *)snostr);

	return(snostr);
}

void
putSerialNumber(char *sno)
{
	int rc;
	char buf[64];

	if (spifReadBlock(SPIF_SNO_ADDRS_BASE,buf,sizeof(buf)) < 0) {
		printf("putSerialNumber: spifReadBlock error\n");
		return;
	}
	if (buf[0] != (char)0xff) {
		printf("SNO store abort: SPI block (0x%lx bytes at 0x%lx) not empty\n",
			SPIF_SNO_ADDRS_SIZE, SPIF_SNO_ADDRS_BASE);
		return;
	}

	spifWriteEnable();	
	spifGlobalUnprotect();	
	spifWriteEnable();	
	rc = spifWriteBlock(SPIF_SNO_ADDRS_BASE,sno,strlen(sno)+1);
	spifWriteEnable();	
	spifGlobalProtect();
	spifWriteDisable();	
	if (rc < 0) {
		printf("putSerialNumber: spifWriteBlock error\n");
		return;
	}

	printf("SNO burned into SPI flash\n");
}

char *SnoHelp[] = {
	"Display or modify VTC serial number.",
	"[new serial number]",
	0,
};

int
SnoCmd(int argc,char *argv[])
{
	int 	i, rc;
	uchar	*sno;

	if (argc == 1) {
		sno = getSerialNumber();
		if (sno) {
			printf("Current serial number: ");
			for(i=0;i<64;i++) {
				if ((sno[i] == 0) || (sno[i] == 0xff))
					break;
				putchar(sno[i]);
			}
			printf("\n");
		}
	}
	else if (argc == 2) {
		if (strcmp(argv[1],"erase") == 0) {
			spifWriteEnable();	
			spifGlobalUnprotect();	
			spifWriteEnable();	
			rc = spifBlockErase(SPIF_SNO_ADDRS_SIZE,SPIF_SNO_ADDRS_BASE);
			spifWriteEnable();	
			spifGlobalProtect();
			spifWriteDisable();	
			if (rc < 0)
				return(CMD_FAILURE);
		}
		else {
			putSerialNumber(argv[1]);
		}
	}
	else
		return(CMD_PARAM_ERROR);
	return(CMD_SUCCESS);
}

void
rosa_startup(void)
{
	// Set TEC temperature to 25C:
	docommand("ad5384 vout 39 2.046259",0);
}
