/* template_etherdev.c:
 * This is a "starter" file for the ethernet driver interface used by
 * the monitor.  The functions are described, but empty.
 * At a minimum, code must be added in all places marked ADD_CODE_HERE.
 * Additional OPT_ADD_CODE_HERE tags indicate locations that can be
 * omitted, but would be nice to have additional facilities.
 *
 * General notice:
 * This code is part of a boot-monitor package developed as a generic base
 * platform for embedded system designs.  As such, it is likely to be
 * distributed to various projects beyond the control of the original
 * author.  Please notify the author of any enhancements made or bugs found
 * so that all may benefit from the changes.  In addition, notification back
 * to the author will allow the new user to pick up changes that may have
 * been made by other users after this version of the code was distributed.
 *
 * Author:	Ed Sutter
 * email:	esutter@lucent.com
 * phone:	908-582-2351
 */

#include "config.h"
#include "genlib.h"
#include "stddefs.h"
#include "ether.h"
#include "phy.h"
#include "xemac.h"
#include "xemac_l.h"
#include "xparameters.h"
#include "xpacket_fifo_l_v2_00_a.h"

#if INCLUDE_ETHERNET

XEmac	emac;

#define EBASE XPAR_OPB_ETHERNET_0_BASEADDR

ulong tx_buff[400];

void
enreset(void)
{
}

int
eninit(void)
{
	XStatus xresult;

	if ((xresult = XEmac_Initialize(&emac,0)) != XST_SUCCESS)
		printf("XEmac_Initialize() failed %d\n",xresult);

	XEmac_SetOptions(&emac,XEmac_GetOptions(&emac) | XEM_POLLED_OPTION);
	XEmac_SetMacAddress(&emac,BinEnetAddr);
	XEmac_Start(&emac);

	XEmac_PhyReset(&emac);
	phy_autoneg(0);
	return (0);
}

int
EtherdevStartup(int verbose)
{
	enreset();
	return(eninit());
}

/* disablePromiscuousReception():
 * Provide the code that disables promiscuous reception.
 */
void
disablePromiscuousReception(void)
{
	XEmac_SetOptions(&emac,XEmac_GetOptions(&emac) & ~XEM_PROMISC_OPTION);
}

/* enablePromiscuousReception():
 * Provide the code that enables promiscuous reception.
 */
void
enablePromiscuousReception(void)
{
	XEmac_SetOptions(&emac,XEmac_GetOptions(&emac) | XEM_PROMISC_OPTION);
}

/* disableBroadcastReception():
 * Provide the code that disables broadcast reception.
 */
void
disableBroadcastReception(void)
{
	XEmac_SetOptions(&emac,XEmac_GetOptions(&emac) & ~XEM_BROADCAST_OPTION);
}

/* enableBroadcastReception():
 * Provide the code that enables broadcast reception.
 */
void
enableBroadcastReception(void)
{
	XEmac_SetOptions(&emac,XEmac_GetOptions(&emac) | XEM_BROADCAST_OPTION);
}

/* 
 * enselftest():
 *	Run a self test of the ethernet device(s).  This can be stubbed
 *	with a return(1).
 *	Return 1 if success; else -1 if failure.
 */
int
enselftest(int verbose)
{
	XStatus x;

	x = XEmac_SelfTest(&emac);
	if (x == XST_SUCCESS) {
		printf("Passed\n");
		return(1);
	}
	else {
		printf("Failed (xstat=%d)\n",(int)x);
		return(-1);
	}
}

/* ShowEtherdevStats():
 * This function is used to display device-specific stats (error counts
 * usually).
 */
void
ShowEtherdevStats(void)
{
	phy_stat(0,1);
	printf("XrecvFrames: %d\n",emac.Stats.RecvFrames);
	printf("XrecvBytes : %d\n",emac.Stats.RecvBytes);
	printf("XxmitFrames: %d\n",emac.Stats.XmitFrames);
	printf("XxmitBytes : %d\n",emac.Stats.XmitBytes);
}

/* getXmitBuffer():
 * Return a pointer to the buffer that is to be used for transmission of
 * the next packet.  Since the monitor's driver is EXTREMELY basic,
 * there will only be one packet ever being transmitted.  No need to queue
 * up transmit packets.
 */
uchar *
getXmitBuffer(void)
{
	return((uchar *)tx_buff);
}

/* sendBuffer():
 * Send out the packet assumed to be built in the buffer returned by the
 * previous call to getXmitBuffer() above.
 */
int
sendBuffer(int length)
{
#if INCLUDE_ETHERVERBOSE
	if (EtherVerbose &  SHOW_OUTGOING)
		printPkt((struct ether_header *)tx_buff,length,ETHER_OUTGOING);
#endif

	XEmac_PollSend(&emac,(uchar *)tx_buff,length);
	EtherXFRAMECnt++;
	return(0);
}

/* DisableEtherdev():
 */
void
DisableEtherdev(void)
{
	XEmac_Stop(&emac);
}

/* extGetIpAdd():
 * If there was some external mechanism (other than just using the
 * IPADD shell variable established in the monrc file) for retrieval of
 * the board's IP address, then do it here...
 */
char *
extGetIpAdd(void)
{
	return((char *)0);
}


/* extGetEtherAdd():
 * If u-boot has populated the inventory area with a valid MAC address
 * then use it; else, return 0.
 */
#define INVENTORY_BASE	0x207c0000

char *
extGetEtherAdd(void)
{
	return(0);
}

/*
 * polletherdev():
 * Called continuously by the monitor (ethernet.c) to determine if there
 * is any incoming ethernet packets.
 *
 * NOTES:
 * 1. This function must be reentrant, because there are a few cases in
 *    processPACKET() where pollethernet() may be called.
 * 2. It should only process one packet per call.  This is important
 *    because if allowed to stay here to flush all available packets,
 *    it may starve the rest of the system (especially in cases of heavy
 *    network traffic).
 * 3. There are cases in the monitor's execution that may cause the
 *    polling polletherdev() to cease for several seconds.  Depending on
 *    network traffic, this may cause the input buffering mechanism on
 *    the ethernet device to overflow.  A robust polletherdev() function
 *    should support this gracefully (i.e. when the error is detected,
 *    attempt to pass all queued packets to processPACKET(), then do what
 *    is necessary to clear the error).
 *
 * This is a hack of the Xilinx library function XEmac_RecvFrame().
 */
int
polletherdev(void)
{
	static uchar pktbuf[1600];
	long pktlen;

	pktlen = sizeof(pktbuf);
	if (XEmac_PollRecv(&emac,pktbuf,&pktlen) == XST_SUCCESS) {
		processPACKET((struct ether_header *)pktbuf, pktlen);
		return(1);
	}

	return(0);
}

int
phy_read(int phynum, int reg, unsigned short *value)
{
	XStatus x;

	x = XEmac_PhyRead(&emac,(Xuint32)0,(Xuint32)reg,(Xuint16 *)value);
	if (x == XST_SUCCESS)
		return(0);
	return(-1);
}

int
phy_write(int phynum, int reg, unsigned short value)
{
	XStatus x;

	x = XEmac_PhyWrite(&emac,(Xuint32)0,(Xuint32)reg,(Xuint16)value);
	if (x == XST_SUCCESS)
		return(0);
	return(-1);
}

#endif
