/* lboot.c:
 * This lboot is tuned to the Xilinx Linux port.
 */
#include "config.h"
#include "stddefs.h"
#include "cpuio.h"
#include "genlib.h"
#include "ether.h"
#include "cli.h"
#include "warmstart.h"
#include "xgpio.h"
#include "xparameters.h"

#define CONGRUENCE_CLASSES			256
#define LINE_SIZE					32
#define XPAR_DDR_0_SIZE				0x08000000
#define	XPAR_PLB_CLOCK_FREQ_HZ		100000000
#define XPAR_PCI_0_CLOCK_FREQ_HZ	100000000

struct board_info {
	unsigned long   bi_memstart;		/* start of DRAM memory */
	unsigned long   bi_memsize;			/* size  of DRAM memory in bytes */
	unsigned long   bi_flashstart;		/* start of FLASH memory */
	unsigned long   bi_flashsize;		/* size  of FLASH memory */
	unsigned long   bi_flashoffset;		/* reserved area for startup monitor */
	unsigned long   bi_sramstart;		/* start of SRAM memory */
	unsigned long   bi_sramsize;		/* size  of SRAM memory */
	unsigned long   bi_bootflags;		/* boot / reboot flag (for LynxOS) */
	unsigned long   bi_ip_addr;			/* IP Address */
	unsigned char   bi_enetaddr[6];		/* Ethernet adress */
	unsigned short  bi_ethspeed;		/* Ethernet speed in Mbps */
	unsigned long   bi_intfreq;			/* Internal Freq, in MHz */
	unsigned long   bi_busfreq;			/* Bus Freq, in MHz */
	unsigned long   bi_baudrate;		/* Console Baudrate */
	unsigned char   bi_s_version[4];	/* Version of this structure */
	unsigned char   bi_r_version[32];	/* Version of the ROM (IBM) */
	unsigned int    bi_procfreq;		/* CPU (Internal) Freq, in Hz */
	unsigned int    bi_plb_busfreq;		/* PLB Bus speed, in Hz */
	unsigned int    bi_pci_busfreq;		/* PCI Bus speed, in Hz */
	unsigned char   bi_pci_enetaddr[6];	/* PCI Ethernet MAC address */
};

typedef struct orig_board_info {
	unsigned int bi_memsize;		/* DRAM installed, in bytes */
	unsigned char bi_enetaddr[6];	/* Local Ethernet MAC address */
	unsigned int bi_intfreq;		/* Processor speed, in Hz */
	unsigned int bi_busfreq;		/* Bus speed, in Hz */
	unsigned int bi_pci_busfreq;	/* PCI Bus speed, in Hz */
} orig_bd_t;

char *lbootHelp[] = {
	"Boot linux",
	"[-b:ov] [entrypoint]",
#if INCLUDE_VERBOSEHELP
	" -b###   set address of boardinfo+cmdline",
	" -o      use old (from ml403) board info structure",
	" -v      verbose",
	"Variables used to build cmdline: ",
	"  CMDLINE00, CMDLINE01, CMDLINE02...",
	"Variables used to override bdinfo: ",
	"  MEMSTART,   MEMSIZE,     FLASHSTART,  FLASHSIZE, FLASHOFFSET",
	"  SRAMSTART,  SRAMSIZE,    BOOTFLAGS,   ETHSPEED, INTFREQ",
	"  BUSFREQ,    BIS_VERSION, BIR_VERSION, PROCFREQ",
	"  PLBBUSFREQ, PCIBUSFREQ,  PCI_ENETADDR",
#endif
	0
};

#define DEFAULT_BINFO_BASE		0x300000

int
lbootCmd(int argc, char *argv[])
{
	extern void virtex4_errata_fix(void);
	extern void flush_instruction_cache(void);

	struct board_info *bip;
	unsigned long dccr, addr;
	char	*env, *cmd, vname[16], buf[3];
	void	(*kernel)(long,long,long,long,long);
	int		cmdlen, opt, verbose, i, use_old_bdt;

	cmd = 0;
	use_old_bdt = cmdlen = verbose = 0;
	bip = (struct board_info *)DEFAULT_BINFO_BASE;
	kernel = (void(*)(long,long,long,long,long))0;

	while ((opt=getopt(argc,argv,"b:ov")) != -1) {
		switch(opt) {
			case 'b':
				bip = (struct board_info *)strtoul(optarg,0,0);
				break;
			case 'o':
				use_old_bdt = 1;
				break;
			case 'v':
				verbose++;
				break;
		}
	}

	if (argc == optind+1)
		kernel = (void(*)(long,long,long,long,long))strtoul(argv[optind],0,0);

	virtex4_errata_fix();
	flush_instruction_cache();

	/* Clear the board info structure... */
	memset((char *)bip,0,sizeof(struct board_info));

	__asm__("mfdccr %0": "=r" (dccr));
	if (dccr == 0) {
		for (addr=0;addr<(CONGRUENCE_CLASSES*LINE_SIZE); addr+=LINE_SIZE)
            __asm__("dccci 0,%0": :"b"(addr));
    }

	/* Look for CMDLINEDD shell variables (where 00 <= DD <= 31),
	 * and if found, concatenate them to build one long command
	 * line to be passed to the kernel at startup...
	 */ 
	for(i=0;i<32;i++) {
		sprintf(vname,"CMDLINE%02d",i);
		if ((env = getenv(vname)) != 0) {
			if (i == 0) {
				cmd = (char *)(bip+1);
				*cmd = 0;
			}
			strcat(cmd,env);
		}
	}
	if (cmd)
		cmdlen = strlen(cmd);

	/* For each entry in the boardinfo structure, allow a 
	 * shell variable to override the default...
	 */
	if ((env = getenv("MEMSTART")) != 0)
    	bip->bi_memstart = strtoul(env,0,0);
	else
    	bip->bi_memstart = 0;

	if ((env = getenv("MEMSIZE")) != 0)
    	bip->bi_memsize = strtoul(env,0,0);
	else
    	bip->bi_memsize = XPAR_DDR_0_SIZE;

	if ((env = getenv("FLASHSTART")) != 0)	
    	bip->bi_flashstart = strtoul(env,0,0);
	else
    	bip->bi_flashstart = 0;

	if ((env = getenv("FLASHSIZE")) != 0)
    	bip->bi_flashsize = strtoul(env,0,0);
	else
    	bip->bi_flashsize = 0;

	if ((env = getenv("FLASHOFFSET")) != 0)
    	bip->bi_flashoffset = strtoul(env,0,0);
	else
    	bip->bi_flashoffset = 0;

	if ((env = getenv("SRAMSTART")) != 0)
    	bip->bi_sramstart = strtoul(env,0,0);
	else
    	bip->bi_sramstart = 0;

	if ((env = getenv("SRAMSIZE")) != 0)
    	bip->bi_sramsize = strtoul(env,0,0);
	else
    	bip->bi_sramsize = 0;

	if ((env = getenv("BOOTFLAGS")) != 0)
    	bip->bi_bootflags = strtoul(env,0,0);
	else
    	bip->bi_bootflags = 0;

#if INCLUDE_ETHERNET
    memcpy((char *)&bip->bi_ip_addr,BinIpAddr,4);
    memcpy(bip->bi_enetaddr,BinEnetAddr,6);
#else
    memset(bip->bi_enetaddr,0,6);
	bip->bi_ip_addr = 0;
#endif

	if ((env = getenv("ETHSPEED")) != 0)
    	bip->bi_ethspeed = (unsigned short)strtoul(env,0,0);
	else
    	bip->bi_ethspeed = 0;

	if ((env = getenv("INTFREQ")) != 0)
    	bip->bi_intfreq = strtoul(env,0,0);
	else
    	bip->bi_intfreq = XPAR_CPU_PPC405_CORE_CLOCK_FREQ_HZ;

	if ((env = getenv("BUSFREQ")) != 0)
    	bip->bi_busfreq = strtoul(env,0,0);
	else
	    bip->bi_busfreq = XPAR_PLB_CLOCK_FREQ_HZ;

	bip->bi_baudrate = ConsoleBaudRate;


	memset((char *)bip->bi_s_version,0,4);
	if ((env = getenv("BIS_VERSION")) != 0)	{
		buf[2] = 0;
    	for(i=0;i<4;i++) {
			memcpy(buf,env+(2*i),2);
			bip->bi_s_version[i] = (uchar )strtol(buf,0,16);
		}
	}


	memset((char *)bip->bi_r_version,0,32);
	if ((env = getenv("BIR_VERSION")) != 0)	{
		int len = strlen(env) / 2;
		if (len > 32) len = 32;
		buf[2] = 0;
    	for(i=0;i<len;i++) {
			memcpy(buf,env+(2*i),2);
			bip->bi_r_version[i] = (uchar )strtol(buf,0,16);
		}
	}

	if ((env = getenv("PROCFREQ")) != 0)
    	bip->bi_procfreq = strtoul(env,0,0);
	else
    	bip->bi_procfreq = 0;

	if ((env = getenv("PLBBUSFREQ")) != 0)
    	bip->bi_plb_busfreq = strtoul(env,0,0);
	else
    	bip->bi_plb_busfreq = 0;

	if ((env = getenv("PCIBUSFREQ")) != 0)
    	bip->bi_pci_busfreq = strtoul(env,0,0);
	else
    	bip->bi_pci_busfreq = XPAR_PCI_0_CLOCK_FREQ_HZ;

	if ((env = getenv("PCI_ENETADDR")) != 0) {
		if (EtherToBin(env,bip->bi_pci_enetaddr) < 0)
			return(CMD_FAILURE);
	}
	else
		memset((char *)bip->bi_pci_enetaddr,0,6);

	/* -vv on the command line will dump board info...
	 */
	if (verbose > 1) {
		printf("brdinfo @ 0x%08lx:\n",(long)bip);
		printf("  memstart:     0x%08x\n",bip->bi_memstart);
		printf("  enetaddr:     %M\n",bip->bi_enetaddr);
		printf("  intfreq:      %d Mhz\n",bip->bi_intfreq);
		printf("  busfreq:      %d Mhz\n",bip->bi_busfreq);
		printf("  pci_busfreq:  %d Hz\n",bip->bi_pci_busfreq);
		if (use_old_bdt == 0) {
			printf("  memsize:      0x%08x\n",bip->bi_memsize);
			printf("  flashstart:   0x%08x\n",bip->bi_flashstart);
			printf("  flashsize:    0x%08x\n",bip->bi_flashsize);
			printf("  flashoffset:  0x%08x\n",bip->bi_flashoffset);
			printf("  sramstart:    0x%08x\n",bip->bi_sramstart);
			printf("  sramsize:     0x%08x\n",bip->bi_sramsize);
			printf("  bootflags:    0x%08x\n",bip->bi_bootflags);
			printf("  ipaddr:       %I\n",bip->bi_ip_addr);
			printf("  ethspeed:     %d Mbps\n",bip->bi_ethspeed);
			printf("  baudrate:     %d Bps\n",bip->bi_baudrate);
			printf("  s_version:    ");
			for(i=0;i<4;i++)
				printf("%02x",bip->bi_s_version[i]);
			printf("\n  r_version:    ");
			for(i=0;i<16;i++)
				printf("%02x",bip->bi_r_version[i]);
			printf("\n                ");
			for(i=16;i<32;i++)
				printf("%02x",bip->bi_r_version[i]);
			printf("\n  procfreq      %d Hz\n",bip->bi_procfreq);
			printf("  plb_busfreq:  %d Hz\n",bip->bi_plb_busfreq);
			printf("  pci_busfreq:  %d Hz\n",bip->bi_pci_busfreq);
			printf("  pci_enetaddr: %M\n\n",bip->bi_pci_enetaddr);
		}
	}

	/* -v (or -vv) on the command line will kernel entry and cmdline...
	 */
	if (verbose) {
		printf("Entrypoint: 0x%08lx\n",(long)kernel);
		if (cmd)
			printf("Cmdline  @  0x%08lx:  <%s>\n\n\n",cmd,cmd);
	}

	if (use_old_bdt) {
		orig_bd_t *obip = (orig_bd_t *)bip;

		obip->bi_memsize = bip->bi_memsize;
		memcpy(obip->bi_enetaddr,bip->bi_enetaddr,6);
		obip->bi_intfreq = bip->bi_intfreq;
		obip->bi_busfreq = bip->bi_busfreq;
		obip->bi_pci_busfreq = bip->bi_pci_busfreq;
	}

	/* -vvv (or more) will not jump into the kernel, just print
	 * the structure.
	 */
	if (verbose < 3)
		kernel((long)bip,0,0,(long)cmd,(long)cmdlen);

	return(CMD_SUCCESS);
}
