/* cpuio.c:
 * This file contains the generic serial port drivers and other
 * miscellaneous target specific functions.
 *
 * General notice:
 * This code is part of a boot-monitor package developed as a generic base
 * platform for embedded system designs.  As such, it is likely to be
 * distributed to various projects beyond the control of the original
 * author.  Please notify the author of any enhancements made or bugs found
 * so that all may benefit from the changes.  In addition, notification back
 * to the author will allow the new user to pick up changes that may have
 * been made by other users after this version of the code was distributed.
 *
 * Author:	Ed Sutter
 * email:	esutter@lucent.com
 * phone:	908-582-2351
 */
#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "ether.h"
#include "stddefs.h"
#include "devices.h"
#include "devtbl.h"
#include "ppcuic.h"
#include "flash.h"

int ConsoleDevice;

/* InitUART():
 * This is the top-level UART initialization for the monitor. 
 */
int
InitUART(int baud)
{
	if (baud == 0)
		baud = ConsoleBaudRate;

	/* Keep track of the last baud rate, so that it can be used if the
	 * incoming baudrate is NULL.
	 */
	ConsoleBaudRate = baud;

	/* Null out the remote put/getchar functions. */
	InitRemoteIO();

	ConsoleDevice = TTY0;	
	
	return(0);
}

int
ConsoleBaudSet(int baud)
{
	if (baud != ConsoleBaudRate) {
		ioctl(ConsoleDevice,SETBAUD,baud,0);
		ConsoleBaudEnvSet();
	}
	return(0);
}

int
target_putchar(char c)
{
	write(ConsoleDevice,&c,1);
    return((int)c);
}

int
target_gotachar()
{
	return(ioctl(ConsoleDevice,GOTACHAR,0,0));
}

int
target_console_empty(void)
{
	return(1);
}

int	
target_getchar()
{
	char	c;

	read(ConsoleDevice,&c,1);
	return((int)c);
}

void
initCPUio()
{
	/*
	 * Interrupt controller setup for the Walnut board.
	 * Note: IRQ 0-15  405GP internally generated; active high; level sensitive
	 *       IRQ 16    405GP internally generated; active low; level sensitive
	 *       IRQ 17-24 RESERVED
	 *       IRQ 25 (EXT IRQ 0) FPGA; active high; level sensitive
	 *       IRQ 26 (EXT IRQ 1) SMI; active high; level sensitive
	 *       IRQ 27 (EXT IRQ 2) Not Used                          
	 *       IRQ 28 (EXT IRQ 3) PCI SLOT 3; active low; level sensitive
	 *       IRQ 29 (EXT IRQ 4) PCI SLOT 2; active low; level sensitive
	 *       IRQ 30 (EXT IRQ 5) PCI SLOT 1; active low; level sensitive
	 *       IRQ 31 (EXT IRQ 6) PCI SLOT 0; active low; level sensitive
	 * Note for Walnut board:
	 *       An interrupt taken for the FPGA (IRQ 25) indicates that either   
	 *       the Mouse, Keyboard, IRDA, or External Expansion caused the 
	 *       interrupt. The FPGA must be read to determine which device
	 *       caused the interrupt. The default setting of the FPGA clears
	 *          
	 */
	ppcMtuicsr(0xFFFFFFFF);          /* clear all ints */
	ppcMtuicer(0x00000000);          /* disable all ints */
	ppcMtuiccr(0x00000020);          /* set all but FPGA SMI to non-critical*/
	ppcMtuicpr(0xFFFFFFE0);          /* set int polarities */
	ppcMtuictr(0x10000000);          /* set int trigger levels */
	ppcMtuicvcr(0x00000001);         /* set vect base=0,INT0 highest priority*/
	ppcMtuicsr(0xFFFFFFFF);          /* clear all ints */
}

